/*---------------------------------------------------------------------------

                     DDDDDD    IIIIII   SSSSSS    CCCCCC
                     DD    DD    II    SS     ~  CC    CC
                     DD    DD    II     SSSSSS   CC
                     DD    DD    II          SS  CC
                     DD    DD    II    SS    SS  CC    CC
                     DDDDDD    IIIIII   SSSSSS    CCCCCC

  ---------------------------------------------------------------------------

  disc.c

  This program creates and triangulates a two-dimensional set of data points
  and then assigns the triangles various values representing the physical
  quantities corresponding to an astrophysical gas disk.  The data are saved
  to a disk file, from which they may be manipulated ("evolved") by a separ-
  ate program, hydro.c.

  The program uses the triangulation algorithm presented in the article,
  "Computing Dirichlet tesselations" by A. Bowyer (_The_Computer_Journal_,
  volume 24, number 2 [1981]).  The algorithm may be generalized to three
  dimensions with few changes, but the physical evolution thereof (including
  conservation of the appropriate physical quantities) is likely to be diffi-
  cult, at best.  See _The Free Lagrange Method_, Springer-Verlag Lecture
  Notes in Physics #238.

  ---------------------------------------------------------------------------

     Version:  Y-MP/PC, PostScript/EGA-graphic       Greg Roelofs
     Source:   d323.c on eagle and oddjob            25 May 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:   make d323
     on PC:     cl -FPi87 -AS -G2 [-DDIAG | -DEGA] [-DNOTRI] d323.c

     DIAG:   print diagnostics to standard output (see global variable Eflag)
     EGA:    plot tesselation as it is created, in EGA graphics
     NOTRI:  tesselate only; skip triangulation (mainly for use with EGA)

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     p###.c:        stand-alone program to create and plot points only

  -> d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotpts.c:     MS-DOS graphics routines for points load
     plotload.c:    MS-DOS graphics routines for tesselation

     f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      most grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

  To run:  [time] d323

     Main dataset will be written to d###f000t000.0; associated information
     is written to d###.out.

  ---------------------------------------------------------------------------

  Version history:

     d250.c:  changed disk density profile to exponential (was lorentzian?);
               changed global viscosity from kinetic to dynamic (sigh) and
               added disk scale height for conversion to two-dimensional
               viscosity; switched to Tom's units; pulled convex hull in and
               spaced it more closely
     d251.c:  reduced viscosity by 1000; changed output logging to file
               rather than stdout; added print_phys_vars() routine
     d252.c:  conditionally included PC stuff again; changed vertex load to
               allow hole in middle
     d253.c:  tripled size of hole
     d254.c:  changed vertex load to follow mass density (i.e., triangles
               have approx. same mass, so smaller near center); added a few
               more ifdef's for better PC/Y-MP compatibility; fixed nasty
               mass-loading bug in physical_variables()
     d255.c:  changed scale height to constant value (actually done via a
               hack to h252.c)
     d260.c:  changed data structures to include an attribute field for the
               vertices (== row on which vertex was loaded); set temperature,
               viscosity to zero
     r261.c:  reduced linear disk size by factor of two; reduced inclination
               to 15 degrees
     r262.c:  attempted to make initial velocities closer to circular
     d263.c:  fixed construct_points() so first point in each ring is close
               to positive x-axis (relative to two-dimensional picture);
               attempted again to make initial velocities closer to circular
     d270.c:  moved tesselation routines (except for tesselate() itself) to
               separate file (can be compiled independently); changed tilted
               disk to warped one; changed potential parameters and loaded
               velocities consistent with Plummer potential instead of Tom's
               (oblate) Schwarzschild potential; added write_edges() routine
               from hydro.c
     d271.c:  fixed initial velocities to work properly with leapfrog
     r272.c:  changed version number and time step (--> 1/3) of d271 load,
               and initial velocities correspondingly
     d273.c:  purely circular initial velocities; time step = 1/3
     d280.c:  defines/includes/etc. moved into h280.h; tesselate() moved
               into tesselate.c; write_edges() moved into prphy280.c (shared
               with hydro.c)
     d285.c:  "new" load:  Schwarzschild potential, but much larger core
               radius than d250-261 (to test precession rate); also inner
               disk aligned with (oblate) potential, outer part warping to
               40 degrees; moved write_edges() into sh285.c; using GRAV() to
               calculate velocities and swap_edges() to compensate for warp;
               rearranged rarr[]; storing rings info in karr/rarr
     d286.c:  reduced temp/pressure by factor of 9 (v_rms by x3)
     d287.c:  also reduced viscosity by factor of 10
     d290.c:  added new boundary-conditions initialization code; renamed 
               from "delaunay" to "disc"
     d291.c:  modified write_tess() output from VAX VMS to Cray format for
               new tplot.c [never evolved]
     d292.c:  modified write_tri() output from VAX VMS to Cray format; 
               changed inclination function from linear to half-cosine
     d293.c:  changed inclination function from half-cosine to quadratic;
               scaled down Rholm by 20%
     d294.c:  changed inclination function from quadratic to x*sin^2(x)
     d300.c:  changed data structures (added pressure field to T[] struct)
     d301.c:  fixed folded-triangle criterion; changed inclination function
               back to quadratic to test
     d310.c:  moved contruct_points() code to points.c; modified write_tess()
               to use macro
     d311.c:  fixed quadratic inclination function (phantom inclination too
               big by factor of two)
     d312.c:  changed inclination function again, back to x*sin^2(x) (curva-
               ture too sharp at edge along line of nodes)
     d313.c:  changed viscosity and pressure back to old values (x 3)
     d320.c:  reversed inclination (inner part inclined to potential; last
               ring lies in plane)
     d321.c:  potential changed to harmonic oscillator (new grav.f)
     d322.c:  "small warp" to test crinkliness
     d323.c:  changed mu_phantom to mu_phantom_orig; added jacobi_time

     d???.c:  new load:  large, spherical halo + exponential disk potential

  ---------------------------------------------------------------------------*/





/**********************************/
/*  Local Includes, Defines, etc. */
/**********************************/

#define LOADER
#include "fd.h"          /* includes, defines, typedefs, prototypes */

static void read_points( void );

#ifdef EGA
#  include <conio.h>     /* kbhit(), getch() */
#  include <graph.h>     /* _setvideomode(), _clearscreen(), etc. */
#  define XMIN           WinDat[0]
#  define XMAX           WinDat[1]
#  define YMIN           WinDat[2]
#  define YMAX           WinDat[3]
#  define x_convert      WinDat[4]
#  define y_convert      WinDat[5]
#endif

/*
 * Physical scales used [notes 1991.2.26]:
 *   b (disk exp scale length) = 2.3 kpc = 1.23
 *   Rholm (size of optical disk) = 9.5 kpc = 5.1 (Holmberg radius or "R_max")
 *     [notes 1992.1.18:  now (d293+) Rholm = 7.5 kpc = 4.0]
 *   Rdisk (size of whole disk) = 20 kpc = 10.7
 *   Rcore (core radius of halo) = 19.6 kpc = 10.5 (isothermal potential)
 */
#define RDISK            10.7               /* target disk radius:  20 kpc */
#define MDISK            5.4                /* target disk mass:  see mu0, b */
#define MU(r)            (mu0*exp(-(r)/bB)) /* mass density function */





/**********************/
/*  Global Variables  */
/**********************/

char             filename[80], *pointsfn;

int              errorlevel=0, num_fake=0, heap_avail, num_swaps;
int              num_DelDirV, max_DelDirV=0, num_ContigPt, max_ContigPt=0;
int              numDirV, numPt, num_compactions=0, Eflag=0;
int              *const ring_last=pRing_last;

static double    vis_bulk=0., read_time, tri_time, physvar_time;
static double    grav_time, total1, total_time;
double           tessel_time, jacobi_time=0., swapE_time=0., write_time=0.;
double           *const LinMom=pLinMom, *const AngMom=pAngMom;
double           *const ring_last_n=pRing_last_n, *const ring_R=pRing_R;
double           *ring_incl, *ring_v, *ring_azim;    /* all three vary */
double           Rdisk_phantom;

vector           *cm_phantom, s[3], cm_subT[MAX_subT][3];
four_vector      *accel_phantom, accel_potent[MAX_subT][3];

struct POINT     Pt[MAX_V], *Heap[HEAP_SIZE], *pNewPt;
struct POINT     *pContigPt[NUM_DIAM], **pBottomHeap;
struct DVERTEX   DirV[MAX_DirV], *pDelDirV[NUM_DIAM];
struct DVERTEX   *pNewDirV[NUM_DIAM], *pPenumDirV[NUM_DIAM];

/* dbgfile not used, but required for swap_edges() (shared with FLOPPY) */
FILE             *outfile, *dbgfile;

int      karr[256] = {
             323,                    /*   d_version                          */
             0,                      /*   iteration                          */
             0,                      /*   numE (triangulate)                 */
             0,                      /*   numT (triangulate)                 */
             0,                      /*   numV (points.c)                    */
             0,                      /*   f_version (reserved)               */
             3,                      /*   En                                 */
             0,                      /*   numRings (points.c)                */
             0,                      /*   num_phantom (points.c)             */
             0,                      /*   p_version (points.c)               */
             0,                      /*   file_type                          */
             0,                      /*   karr2size                          */
             0                       /*   rarr2size                          */
         };

double   rarr[256] = {
             1.,  /* 1./3. ?? */     /*   delta_t                            */
             0.,                     /*   Time                               */
             0.,                     /*   KinE (reserved)                    */
             0.,                     /*   GravE (physical_variables)         */
             0.,                     /*   IntE (reserved [not used])         */
             0.,                     /*   TotE (reserved)                    */
             0., 0., 0.,             /*   LinMom[3] (reserved)               */
             0.,                     /*   TotLinMom (reserved)               */
             0., 0., 0.,             /*   AngMom[3] (reserved)               */
             0.,                     /*   TotAngMom (reserved)               */
   /* UNITS: */
             1870.,                  /*   pc_length                          */
             1.0e10,                 /*   Msol_mass                          */
             3.4,                    /*   Myr_time                           */
             3.09e18,                /*   cm_pc (3.09e16 km/kpc)             */
             1.99e33,                /*   g_Msol                             */
             86400 * 365.2422 * 1e6, /*   s_Myr                              */
             6.67e-8,                /*   G_cgs                              */
             0.,                     /*   G_pMM (physical_variables)         */
             0.,                     /*   G_LMT (physical_variables)         */
   /* POTENTIAL: */
             10.5,                   /*   Rcore                              */
             .00343,  /* 91.3.22 */  /*   rho0                               */
             -.3764,                 /*   phi0 [notes 1991.2.27]             */
             0.,                     /*   c1half (physical_variables)        */
             0.,                     /*   c2 (physical_variables)            */
   /* DISK: */
             0.,                     /*   Rdisk (points.c)                   */
             0.,                     /*   Rholm (points.c)                   */
             0.,                     /*   bB (points.c)                      */
             0.,                     /*   hole (points.c)                    */
             0.,                     /*   Mdisk (physical_variables)         */
             0.,                     /*   mu0 (points.c)                     */
             .0186,   /* .0062, */   /*   v_rms [v_disp]                     */
             0.,                     /*   kT_m (physical_variables)          */
             5.6e-9,  /* 5.6e-10, */ /*   vis_dyn                            */
             0.,                     /*   gamma_dyn (physical_variables)     */
             40.,                    /*   Inclination [degrees]              */
             0., 0., 0.,             /*   ring_last_n[3] (physical_vars)     */
   /* PHANTOM RING:  notes 1991.6.24-27, 1991.7.11, 1991.8.11: */
             0.,                     /*   r_cm_phantom (physical_variables)  */
             0.,                     /*   mu_phantom_orig (physical_vars)    */
             0.,                     /*   mass_phantom (physical_variables)  */
             0.,                     /*   dvdr_phantom ()                    */
             0.                      /*   scaleht2_phantom (physical_vars)   */
         };

#ifdef EGA
   double   WinDat[6] = {-RDISK*1.33, RDISK*1.33, -RDISK, RDISK, 0., 0.};
#endif             /* = { XMIN, XMAX, YMIN, YMAX, x_convert, y_convert }; */

#ifdef TRI
   struct edge       E[MAX_E], *Heap2[HEAP_SIZE];
   struct triangle   T[MAX_T];
   struct vertex     V[MAX_V];
#endif





/******************/
/*  Main program  */
/******************/

int main(int argc, char *argv[])
  {
    total1 = SECOND();

    if (argc < 2) {
        printf("\nProgram:  disc.c\n  usage:  d%03d filename\n", d_version);
        exit(0);
    }
    pointsfn = argv[1];   /* assume only one argument for now */

    read_points();
    tesselate();

#ifdef TRI
    triangulate();
    physical_variables();
    print_phys_vars();
    write_edges();     /* d###f000v000.0 warped triangulation file for tplot */
    write_tri();       /* d###f000s000.0 "surface" file for render */
    write_data();      /* d###f000t000.0 full checkpoint file for f### */
#endif

#ifndef MSDOS          /* d###{tes,tri}#### initial (two-dimensional) */
    write_tess();      /*  tesselation/triangulation files for tplot */
#elif defined(EGA)
    initialize_plot();
    plot_tri();
#endif

    fprintf(outfile, "\n\nCPU timing information (totals, in seconds):\n");
    fprintf(outfile, "  read_points()                %11.6f\n", read_time);
    fprintf(outfile, "  tesselate()                  %11.6f\n", tessel_time);
#ifdef TRI
    fprintf(outfile, "  triangulate()                %11.6f\n", tri_time);
    fprintf(outfile, "  physical_variables()         %11.6f\n", physvar_time);
    fprintf(outfile, "     swap_edges()              %11.6f\n", swapE_time);
    fprintf(outfile, "     GRAV()                    %11.6f\n", grav_time);
    fprintf(outfile, "  write_edges/tri/data/tess()  %11.6f\n", write_time);
#endif

    total_time = SECOND() - total1;
    fprintf(outfile, "\nTotal program CPU time:        %11.6f\n", total_time);
    fprintf(outfile, "------------------------------------------\n");
    fclose(outfile);

    printf("\nHave a somewhat pleasant day.\n");
  }





/****************************/
/*  Function read_points()  */
/****************************/

static void read_points( void )
{
    char         outfn[80];
    static int   karr2[256];
    int          i, *q=pRing_last, *r=karr2+(pRing_last-karr);
    int          total_bytes;
    double       rarr2[256], t1, phi, offset, *p, *coord;
/*  double       R_hole, R_hole_sqr;   not presently used */
    FILE         *ptsfile;


/*--------------------------------------------------------------------------
    Open "standard" output file first, so have a place to print progress.
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    sprintf(outfn, "d%03d.out", d_version);
    if ((outfile = fopen(outfn,"a")) == NULL) {
        printf("\nerror:  couldn't open output file [ %s ]\n", outfn);
        exit(1);
    }
    printf("Program:  d%03d.c\n", d_version);
    fprintf(outfile, "Program:  d%03d.c\n", d_version);

/*---------------------------------------------------------------------------
    Create convex hull (triangle):  three fake points at pretty big radius.
  ---------------------------------------------------------------------------*/

    /* tesselate() still assumes this is always 3: */
#   define NUM_BIGHULL  3

    for (i = 0;  i < NUM_BIGHULL;  ++i) {
        /* 7.63 = degeneracy-breaking offset: */
        phi = (i*(360./NUM_BIGHULL) + 7.63)*PI180;
        Pt[i].x = PRETTY_BIG * cos(phi);
        Pt[i].y = PRETTY_BIG * sin(phi);
        ++num_fake;
#ifdef DIAG
        if (Eflag)
            printf("Pt[%d] position:    %e,    %e\n", i, Pt[i].x,
              Pt[i].y);
#endif /* DIAG */
    }

    if (hole > 0.)   /* not fully supported */
        ++num_fake;

/*---------------------------------------------------------------------------
    Open input file (points), read karr[]/rarr[] arrays, and copy necessary
    data into our karr[]/rarr[] arrays.
  ---------------------------------------------------------------------------*/

    if ((ptsfile = fopen(pointsfn,"rb")) == NULL) {
        printf("error:  couldn't open file [ %s ]\n", pointsfn);
        exit(1);
    }
    total_bytes = fread(karr2, sizeof(int), 256, ptsfile) * sizeof(int);
    total_bytes += fread(rarr2, sizeof(double), 256, ptsfile) * sizeof(double);

    numV = karr2[&numV-karr];
    numRings = karr2[&numRings-karr];
    num_phantom = karr2[&num_phantom-karr];
    p_version = karr2[&p_version-karr];
    for (i = 0;  i <= numRings;  ++i)   /* copy ring_last[] */
        *q++ = *r++;

    Rdisk = rarr2[&Rdisk-rarr];
    Rholm = rarr2[&Rholm-rarr];
    bB = rarr2[&bB-rarr];
    hole = rarr2[&hole-rarr];
    mu0 = rarr2[&mu0-rarr];

    numPt = numV + NUM_BIGHULL;  /* includes fake hull pts, poss. hole point */
    Rdisk_phantom = rarr2[255];  /* was temporary holder for this */

/*---------------------------------------------------------------------------
    Allocate temporary space for data, read them, and copy them to Pt[].
  ---------------------------------------------------------------------------*/

    coord = (double *) malloc((unsigned) (2*numV)*sizeof(double));
    if (!coord) {
        printf("error:  couldn't allocate memory to read [ %s ]\n",
          pointsfn);
        return;
    }

    total_bytes += fread(coord, sizeof(double), 2*numV, ptsfile) *
                                sizeof(double);

    /* copy points into Pt[] */
    for (p = coord, i = num_fake;  i < numPt;  ++i)
        Pt[i].x = *p++;
    for (i = num_fake;  i < numPt;  ++i)
        Pt[i].y = *p++;

    free(coord);
    fclose(ptsfile);

/*---------------------------------------------------------------------------
    Print count of points and exit.
  ---------------------------------------------------------------------------*/

    fprintf(outfile, "\nLoading of points complete, you betcha.  %d "
      "points (%d bytes) loaded%s", numV, total_bytes, (hole>0.)?
      ",\n including 1 temporary central point to be removed in tesselate().\n"
      : ".\n");

    read_time = SECOND() - t1;

} /* end function read_points() */





#ifdef TRI /*................................................................*/

/****************************/
/*  Function triangulate()  */
/****************************/

void triangulate( void )
{
    int            i, j, j2, k, m, eindex, tindex, cur_attrib;
    double         t1;
    struct POINT   **pHeap;
    struct edge    *pE;


/*--------------------------------------------------------------------------
    Initialize triangulation arrays (only those elements that are not guar-
    anteed to be filled, i.e., those which may be NULL by default:  Heap2
    [initially all NULL] and the second of an edge's two triangles [edge may
    be part of the convex hull, thus only the inside triangle exists]).
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    for (k = 0;  k < HEAP_SIZE;  ++k)
        Heap2[k] = NULL;

    for (k = 0;  k < MAX_E;  ++k) {
        E[k].pT[1] = NULL;          /* assume everybody on hull until */
        E[k].status = E_ON_HULL;    /*  told otherwise... */
    }

/*--------------------------------------------------------------------------
    Copy/convert tesselation information to (dual) triangulation.  (Ha ha!
    Sounds easy, doesn't it?)
  --------------------------------------------------------------------------*/

    /* loop through tesselation points, copying to triangulation vertices */
    for (k = 0;  k < numV;  ++k) {
        V[k].r[0] = Pt[k+num_fake].x;
        V[k].r[1] = Pt[k+num_fake].y;
        V[k].r[2] = 0.;   /* this is assumed in physical_variables() */
    }

    /* assign vertex "attributes" (ring numbers) */
    cur_attrib = (hole > 0.)? 1 : 0;
    for (k = 0;  k < numV;) {
        while (k <= ring_last[cur_attrib])
            V[k++].attrib = cur_attrib;
        ++cur_attrib;
    }

    /* loop over Dirichlet vertices to set up triangles next */
    for (i = 0, eindex = 0, tindex = 0;  i < numDirV;  ++i) {
        for (j = 0;  j < 3;  ++j)
            if ((m = DirV[i].pFormPt[j] - &Pt[num_fake]) < 0)
                goto Next_i;   /* exclude fake triangles */
            else
                /* associate triangle's vertices with corresponding DirV's
                 * forming points (will be 1-1 correspondence between Pt[]
                 * and V[] arrays)--offset triangle's vertex indices to align
                 * with opposite edges, when edges defined later */
                T[tindex].pV[(j+1)%3] = &V[m];    /* OK:  V[m], not DirV[m] */

        /* must have valid triangle now:  set up links */
        for (j = 0;  j < 3;  ++j) {
            /* look at pairs of forming points pFormPt[j], pFormPt[(j+1)%3] */
            pHeap = DirV[i].pFormPt[j]->pHeap;
            j2 = (j+1)%3;
            k = 0;

            /* search through contiguity list of pFormPt[j], looking for
             * pointer to other forming point (j2) */
            while (*++pHeap != DirV[i].pFormPt[j2])
                if (++k >= DirV[i].pFormPt[j]->num_contig) {
                    printf("error:  didn't find contig point in list #1\n");
                    exit(5);
                }

        /*------------------------------------------------------------------
            Have found match (better have:  otherwise contiguity list cor-
            rupted).  Check "parallel heap" to see if still contiguous.  (As
            edges are created, contiguities are "marked out" in Heap2 by way
            of bookkeeping; a "removed contiguity" is marked by placing a
            [non-NULL] pointer to the newly-formed edge in Heap2 in the same
            two locations as those occupied by the original contiguity in
            Heap).
          ------------------------------------------------------------------*/

            if (pE = Heap2[pHeap - Heap]) {
                /* non-zero:  edge already exists (so pT[0] filled) */
                T[tindex].pE[j] = pE;
                pE->pT[1] = &T[tindex];
                pE->status = 0;    /* not on hull after all */
            } else {
                /* NULL pointer:  "create" edge now */
                T[tindex].pE[j] = &E[eindex];
                E[eindex].pT[0] = &T[tindex];
                E[eindex].pV[0] = T[tindex].pV[j2];  /* DirV[i].pFormPt[j]  */
                E[eindex].pV[1] = T[tindex].pV[(j+2)%3];   /*  .pFormPt[j2] */

                /* mark out contiguity for both points (in Heap2) */
                Heap2[ pHeap-Heap ] = &E[eindex];

                pHeap = DirV[i].pFormPt[j2]->pHeap;
                k = 0;
                while (*++pHeap != DirV[i].pFormPt[j])
                    if (++k >= DirV[i].pFormPt[j2]->num_contig) {
                        printf("error:  didn't find contig point in list #2\n");
                        exit(6);
                    }
                Heap2[ pHeap-Heap ] = &E[eindex++];

            }  /* end if/else (edge already exists or not) */
        }  /* end j-loop (forming points/triangulation vertices) */

        ++tindex;
Next_i:
        ;
    }  /* end i-loop (Dirichlet vertices/Delaunay triangles) */

    /* numV already set by points.c */
    numE = eindex;
    numT = tindex;

    cm_phantom = cm_subT[numT];    /* initialize physical_variables pointers */
    accel_phantom = accel_potent[numT];

    printf("\nTriangulation complete, you betcha.\n");
    printf("  Number of vertices loaded:    %7d\n", numV);
    printf("  Number of triangles created:  %7d\n", numT);
    printf("  Number of edges created:      %7d\n", numE);
    fprintf(outfile, "\nTriangulation complete, you betcha.\n");
    fprintf(outfile, "  Number of vertices loaded:    %7d\n", numV);
    fprintf(outfile, "  Number of triangles created:  %7d\n", numT);
    fprintf(outfile, "  Number of edges created:      %7d\n", numE);

    tri_time = SECOND() - t1;

} /* end function triangulate() */





/***********************************/
/*  Function physical_variables()  */
/***********************************/

void physical_variables( void )
  {
    int      i, j, j1, j2, k, cur_attrib, numpts;
    double   cur_incl, sin_cur, cos_cur;
    double   t1, t2, density, area, r, rsqr, v_avg;
    double   p, d1, d2, Rmid_phantom, x1, x2, ex1, ex2, mass_integral;
    double   inclin_phantom, x, theta, r_cm_infinitesimal_phantom;
    vector   cm;

    static double   min_mass=9.e9, max_mass=0.;
    static double   coeff1=11./18., coeff2=7./36.;
    struct Ring {
        double R, inclination, sin_incl, cos_incl;
    } ring[50];



/*--------------------------------------------------------------------------
    Initialize some basic stuff first:  loop over vertices, making sure for-
    ces and effective vertex-masses all equal zero.

    Also calculate potential-shape parameters on basis of elliptical type
    (e.g., E3 or E7:  En value); see deZeeuw and Merritt, Ap.J. v267:p571
    (1983), or David et al, Ap.J. v295:p65 (1985).
  --------------------------------------------------------------------------*/

#ifndef MSDOS
    t1 = SECOND();
#endif

    p = 10. / (10-En);   /* axis ratios, rel. to short axis   */
    d1 = (SQR(p)-1.) / (7.*(SQR(p)+2.));
    d2 = (1. - 3./(2.*powri(p,3) + 1.)) / 6.;
    c1half = pow(d1,.6) * pow(d2,.4) / 2.;
    c2 = pow(d1/d2, .4);

    kT_m = SQR(v_rms);
    gamma_dyn = vis_bulk - (2./3.)*vis_dyn;

    /* gravitational constant in parsecs, M_sol, Myr units: */
    G_pMM = G_cgs * g_Msol * SQR(s_Myr) / powri(cm_pc, 3);
    /* ...in program Length, Mass, Time units: */
    G_LMT = G_pMM * Msol_mass * SQR(Myr_time) / powri(pc_length, 3);

    for (k = 0;  k < numV;  ++k) {
        V[k].mass = 0.;
        V[k].F[0] = 0.;
        V[k].F[1] = 0.;
        V[k].F[2] = 0.;
    }

/*---------------------------------------------------------------------------
    Warp disk out of xy-plane, starting at cylindrical R = Rholm and ending
    with last ring, at R = Rdisk.  Will use swap_edges() to fix any distor-
    tion of the triangulation.

    Linear inclination function discarded because of bending at transition
    region (d290f291-2 run); replaced with half-cosine (d292), quadratic
    (d293, disastrous) and then linear*half-cosine (d294) function [notes
    1992.1.18-21].  INCLIN_PHANT is a linear extrapolation of INCLIN at
    x = (r-Rholm)/(Rdisk-Rholm) = 1.  Replaced with inverse (rotated) incli-
    nation function (d320):  inclined in center, declining to zero at edge.
  ---------------------------------------------------------------------------*/

#  define INCLIN(r)  (atan((Rdisk/(20.*(r)))*sin(TWOPI*(r)/Rdisk))/PI180)
#  define INCLIN_PHANT(r)  INCLIN(r)
#  define INCLIN_INNER  0.
#if 0
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (Inclination*((r)-Rholm)/(Rdisk-Rholm))       /* linear */
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (Inclination*.5*(1.-cos(PI*((r)-Rholm)/(Rdisk-Rholm))))
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (Inclination*SQR((r)-Rholm)/SQR(Rdisk-Rholm)) /* quadr. */
#  define INCLIN_PHANT(r)  (Inclination*(2.*((r)-Rholm)/(Rdisk-Rholm)-1.))
#  define INCLIN_INNER  0.
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (Inclination*(((r)-Rholm)/(Rdisk-Rholm))\
                      *.5*(1.-cos(PI*((r)-Rholm)/(Rdisk-Rholm))))  /* x sin^2 */
#  define INCLIN_PHANT(r)  (Inclination*((r)-Rholm)/(Rdisk-Rholm)) /* x sin^2 */
#  define INCLIN_INNER  0.
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (Inclination - Inclination*(((r)-Rholm)/(Rdisk-Rholm))\
                    *.5*(1.-cos(PI*((r)-Rholm)/(Rdisk-Rholm))))   /* x sin^2 */
#  define INCLIN_PHANT(r)  (Inclination - Inclination*((r)-Rholm)/(Rdisk-Rholm))
#  define INCLIN_INNER  Inclination                      /* inverted x sin^2 */
/*---------------------------------------------------------------------------*/
#  define INCLIN(r)  (atan((Rdisk/(20.*(r)))*sin(TWOPI*(r)/Rdisk))/PI180)
#  define INCLIN_PHANT(r)  INCLIN(r)
#  define INCLIN_INNER  0.                                   /* "small warp" */
/*---------------------------------------------------------------------------*/
#endif /* 0 */

    /* GRR:  this assumes no hole */
    cur_attrib = 0;
    cur_incl = INCLIN_INNER;
    sin_cur = sin(PI180 * cur_incl);
    cos_cur = cos(PI180 * cur_incl);

    ring[cur_attrib].R = 0.;
    ring[cur_attrib].inclination = cur_incl;
    ring[cur_attrib].sin_incl = sin_cur;
    ring[cur_attrib].cos_incl = cos_cur;

#if 0 /* 322 */
    /* find place where warp is to begin */
    for (k = 1;  k < numV;  ++k)
        if (V[k].attrib == cur_attrib) {
            V[k].r[2] = -V[k].r[0] * sin_cur;
            V[k].r[0] =  V[k].r[0] * cos_cur;
            continue;
        } else if ((r = sqrt(SQR(V[k].r[0]) + SQR(V[k].r[1]))) > Rholm)
            break;
        else {
            cur_attrib = V[k].attrib;
            ring[cur_attrib].R = r;
            ring[cur_attrib].inclination = cur_incl;
            ring[cur_attrib].sin_incl = sin_cur;
            ring[cur_attrib].cos_incl = cos_cur;
            V[k].r[2] = -V[k].r[0] * sin_cur;
            V[k].r[0] =  V[k].r[0] * cos_cur;
        }

    if (k >= numV) {
        fprintf(stderr, "error:  didn't find warp radius\n");
        exit(2);
    }
#else /* 1 */ /* 322 */
    k = 1;
#endif /* 0 */ /* 322 */

    while (k < numV) {        /* ready to roll:  start tilting rings */
        r = sqrt(SQR(V[k].r[0]) + SQR(V[k].r[1]));
        cur_incl = INCLIN(r);

        cur_attrib = V[k].attrib;
        ring[cur_attrib].R = r;
        ring[cur_attrib].inclination = cur_incl;
        ring[cur_attrib].sin_incl = sin(PI180 * cur_incl);
        ring[cur_attrib].cos_incl = cos(PI180 * cur_incl);

        while (k <= ring_last[cur_attrib]) {
            V[k].r[2] = -V[k].r[0] * ring[cur_attrib].sin_incl;
            V[k].r[0] =  V[k].r[0] * ring[cur_attrib].cos_incl;
            ++k;
        }
    }

/*---------------------------------------------------------------------------
    Save rings information into karr/rarr [notes 1991.3.25].
  ---------------------------------------------------------------------------*/

    if (numRings != cur_attrib)
        printf("\nerror:  numRings (%d) != cur_attrib (%d)\n", numRings,
          cur_attrib);

    /* GRR: just malloc these instead? (write after karr/rarr, before rest) */
    ring_incl = ring_R +   (numRings+1);   /* pointers to sub-array */
    ring_v    = ring_R + 2*(numRings+1);   /* ditto */
    ring_azim = ring_R + 3*(numRings+1);   /* ditto */
    ring_v[0] = 0.;

    for (j = 0;  j <= numRings;  ++j) {
        ring_R[j] = ring[j].R;
        ring_incl[j] = ring[j].inclination;
    }

    ring_last_n[0] = ring[numRings].sin_incl;
    ring_last_n[1] = 0.;
    ring_last_n[2] = ring[numRings].cos_incl;

/*---------------------------------------------------------------------------
    Triangulation now distorted, so calculate edge-lengths, edge-normals and
    pointers to edge-normals, and then do swap_edges() to correct.
  ---------------------------------------------------------------------------*/

    for (k = 0;  k < numE;  ++k)
        E[k].len = DIST( E[k].pV[0]->r, E[k].pV[1]->r );

    for (k = 0;  k < numT;  ++k) {
        for (j = 0;  j < 3;  ++j)        /* three edges/vertices  */
            for (i = 0;  i < 3;  ++i)    /* cartesian coordinates */
                s[j][i] = T[k].pV[j]->r[i] - T[k].pV[(j+1)%3]->r[i];

        for (j = 0;  j < 3;  ++j) {      /* three vertices again  */
            double   norm;

            j1 = (j+1)%3;

            /* loop over cartesian indices:  calc normal to opposite edge */
            for (i = 0;  i < 3;  ++i) {
                T[k].n[j][i] = s[j][i] - DOT(s[j],s[j1]) * s[j1][i] /
                                            SQR(T[k].pE[j]->len);
                if (T[k].pE[j]->pT[0] == &T[k])
                    T[k].pE[j]->n[0] = T[k].n[j];
                else
                    T[k].pE[j]->n[1] = T[k].n[j];
            }

            /* loop over cartesian indices again and make edge normals unit */
            norm = NORM(T[k].n[j]);
            for (i = 0;  i < 3;  ++i)
                T[k].n[j][i] /= norm;

        } /* end j-loop (vertices) */
    } /* end k-loop (triangles) */

    /* don't have masses yet, but shouldn't matter:  no division */
    i = 0;
    while (swap_edges(i++)) ;

/*--------------------------------------------------------------------------
    Loop over triangles, calculating center-of-mass position, density, area,
    and, finally, mass of each.
  --------------------------------------------------------------------------*/

    for (Mdisk = 0., k = 0;  k < numT;  ++k) {
        /* loop over cartesian coordinates:  calculate center-of-mass
         * position of current triangle (average of vertices) */
        for (i = 0;  i < 3;  ++i) {
            cm[i] = (T[k].pV[0]->r[i] + T[k].pV[1]->r[i] + T[k].pV[2]->r[i])
                                              / 3.;
            /* loop over 3 sides */
            for (j = 0;  j < 3;  ++j)
                s[j][i] = T[k].pV[j]->r[i] - T[k].pV[(j+1)%3]->r[i];
        }

        rsqr = DOT(cm, cm);
        r = sqrt(rsqr);

        /* exponential disk; scale length = b, central surface density = mu0 */
        density = MU(r);
        area = .5 * sqrt(  SQR( XCROSS(s[1],s[2]) ) + SQR( YCROSS(s[1],s[2]) )
                         + SQR( ZCROSS(s[1],s[2]) )  );

        T[k].mass = density * area;
        Mdisk += T[k].mass;
        min_mass = MIN( min_mass, T[k].mass );
        max_mass = MAX( max_mass, T[k].mass );

/*      T[k].mass = MDISK / ((double) numT);  [for non-regular point load]  */
/*      T[k].pressure = density * kT_m;       [for polytropic eqn of state] */

        /* loop through vertices, adding in this triangle's contribution to
         * its vertices' masses and forces [notes 1990.8.1, p.2] */
        for (j = 0;  j < 3;  ++j)
            T[k].pV[j]->mass += T[k].mass / 3.;


        /* need gravitational curvature in equatorial plane at current radius:
         * scale height = sqrt( (kT/m) / g1 )  [see RHM notes]  */
/*      T[k].scaleht2 = 2. * v_rms / sqrt(g1);    */

        /* variable scale height doesn't work well; use constant one for now */
        T[k].scaleht2 = .6;

    } /* end k-loop (triangles) */

    printf("\n  Min triangle mass: %10.3e\n", min_mass);
    printf("  Max triangle mass: %10.3e\n", max_mass);
    printf("  Min vertex mass:   %10.3e (approx)\n", V[numV-1].mass);
    printf("  Max vertex mass:   %10.3e (approx)\n", V[0].mass);
    fprintf(outfile, "\n  Min triangle mass: %10.3e\n", min_mass);
    fprintf(outfile, "  Max triangle mass: %10.3e\n", max_mass);
    fprintf(outfile, "  Min vertex mass:   %10.3e (approx)\n", V[numV-1].mass);
    fprintf(outfile, "  Max vertex mass:   %10.3e (approx)\n\n", V[0].mass);

/*---------------------------------------------------------------------------
    Calculate physical quantities of phantom ring at boundary of disk.
    [Notes 1991.6.24-27, 1991.7.11,18,22,29, 1991.8.5-7, 1991.12.11-14]
  ---------------------------------------------------------------------------*/

    scaleht2_phantom = .6;   /* same as other scale heights (just above) */

    Rmid_phantom = .5*(Rdisk_phantom + Rdisk);
    x1 = Rdisk / bB;
    x2 = Rmid_phantom / bB;  /* phantom wedgies = inner half of phantom ring */
    ex1 = exp(-x1);
    ex2 = exp(-x2);
    mass_integral = (1.+x1)*ex1 - (1.+x2)*ex2;   /* polar integral of MU(r) */

    r_cm_infinitesimal_phantom = bB * (2. + (SQR(x1)*ex1 - SQR(x2)*ex2) /
                                                   mass_integral);
    r_cm_phantom = r_cm_infinitesimal_phantom * sin(PI/num_phantom) /
                                                 (PI/num_phantom);

    mass_phantom = PI * mu0 * SQR(bB) * mass_integral;
    mu_phantom_orig = MU(r_cm_infinitesimal_phantom);

    /* create phantom ring of points, with consistent inclination, at boundary:
     *    x = r cos theta
     *    y = r sin theta
     *    z = 0
     *    x' =  x cos i
     *    y' =  y
     *    z' = -x sin i
     * put results directly in cm_phantom, since not needed anywhere else.
     * [will use for calculating dvdr_phantom:  notes 1991.12.15]
     */
    inclin_phantom = PI180 * INCLIN_PHANT(Rmid_phantom);
    for (j = num_phantom;  j < 2*num_phantom;  ++j) {
        theta = (j-num_phantom) * TWOPI / num_phantom;
        x = Rmid_phantom * cos(theta);
        cm_phantom[j][0] = x * cos(inclin_phantom);
        cm_phantom[j][1] = Rmid_phantom * sin(theta);
        cm_phantom[j][2] = -x * sin(inclin_phantom);
    }

/*---------------------------------------------------------------------------
    Calculate center-of-mass position of each triangle-third, phantom wedgie,
    and outer phantom ring; then send array off to have accelerations and
    potential-values calculated.  Finally, calculate gravitational force on 
    each vertex.  [Notes 1991.3.7-13,24, 1991.12.15]
  ---------------------------------------------------------------------------*/

    for (k = 0;  k < numT;  ++k)        /* triangles */
        for (j = 0;  j < 3;  ++j)       /* sub-triangles */
            for (i = 0;  i < 3;  ++i)   /* cartesian coordinates */
                cm_subT[k][j][i] = coeff1 * T[k].pV[j]->r[i]  +
                  coeff2 * (T[k].pV[(j+1)%3]->r[i] + T[k].pV[(j+2)%3]->r[i]);

    numpts = 3*numT + 2*num_phantom;

    /* loop over "phantom" external wedgies */
    for (k = ring_last[numRings-1]+1, j = 0;  j < num_phantom;  ++j, ++k) {
        for (i = 0;  i < 3;  ++i)
            cm_phantom[j][i] = (r_cm_phantom / Rdisk) * V[k].r[i];
        V[k].mass += mass_phantom;   /* add mass contribs to outer points */
    }

    t2 = SECOND();
    GRAV(&numpts);
    grav_time = SECOND() - t2;

    for (k = 0;  k < numT;  ++k) {           /* triangles */
        double   mass_over_3=T[k].mass/3.;   /* mass of each sub-triangle */

        for (j = 0;  j < 3;  ++j) {          /* sub-triangles */
            for (i = 0;  i < 3;  ++i)        /* cartesian coordinates */
                T[k].pV[j]->F[i] += mass_over_3 * grav_accel[k][j][i];
            GravE += mass_over_3 * grav_potent[k][j][3];
        }
    }

    /* add in "phantom" gravitational contribs to vertices on outermost ring */
    for (k = ring_last[numRings-1]+1, j = 0;  j < num_phantom;  ++j, ++k) {
        for (i = 0;  i < 3;  ++i)
            V[k].F[i] += mass_phantom * accel_phantom[j][i];
        GravE += mass_phantom * potent_phantom[j][3];
    }

/*---------------------------------------------------------------------------
    Time-average circular velocities over an orbit (i.e., assume circular or-
    bits and average over each ring the velocities of vertices, weighted by
    the relative time (~ inverse velocity) spent at that velocity).  Then
    assign (vector) velocity to each vertex, rotated appropriately to follow
    the warp.  [Notes 1991.3.17-18]
  ---------------------------------------------------------------------------*/

    for (v_avg = 0., j = 1;  j <= numRings;  ++j) {   /* skip center vertex */
#       define v_avg_R   v_avg

        /* sum inverse velocities */
        for (k = ring_last[j-1] + 1;  k <= ring_last[j];  ++k)
            v_avg += 1. / sqrt(DOT(V[k].F, V[k].r) / -V[k].mass);

        /* calc average velocity */
        ring_v[j] = (double) (ring_last[j] - ring_last[j-1]) / v_avg;
        v_avg_R = ring_v[j] / ring[j].R;   /* do div. outside of loop below */

        fprintf(outfile, "  r[%2d] = %7.4f,  v = %10.4e,  inclin = %5.2f\n", 
          j, ring_R[j], ring_v[j], ring_incl[j]);

        /* assign averaged, tilted velocity vectors and clear force array */
        for (k = ring_last[j-1] + 1;  k <= ring_last[j];  ++k) {
            V[k].v[0] = v_avg_R * (-V[k].r[1] * ring[j].cos_incl);
            V[k].v[1] = v_avg_R * (V[k].r[0] * ring[j].cos_incl -
                                   V[k].r[2] * ring[j].sin_incl);
            V[k].v[2] = v_avg_R * (V[k].r[1] * ring[j].sin_incl);
            for (i = 0;  i < 3;  ++i)
                V[k].F[i] = 0.;
        }
    }

/*---------------------------------------------------------------------------
    Time-average circular velocity of (massless) phantom ring points (radius
    Rmid_phantom), and then calculate shear derivative for phantom wedgies.
    [Notes 1991.12.15]
  ---------------------------------------------------------------------------*/

    for (v_avg = 0., j = num_phantom;  j < 2*num_phantom;  ++j)
        v_avg += 1. / sqrt(-DOT(accel_phantom[j], cm_phantom[j]));

    v_avg = num_phantom / v_avg;
    dvdr_phantom = (v_avg - ring_v[numRings]) / (Rmid_phantom - Rdisk);

    fprintf(outfile,
      "  r[%2d] = %7.4f,  v = %10.4e,  inclin = %5.2f   [middle of phantom]\n", 
      numRings+1, Rmid_phantom, v_avg, inclin_phantom/PI180);
    fprintf(outfile, "  dvdr_phantom = %10.4e\n", dvdr_phantom);

/*---------------------------------------------------------------------------
    Double-check that edge-normals really are normal.
  ---------------------------------------------------------------------------*/

#if FALSE
    for (k = 0;  k < numE;  ++k) {
        for (i = 0;  i < 3;  ++i)
            s[0][i] = E[k].pV[0]->r[i] - E[k].pV[1]->r[i];
        for (j2 = 0;  j2 < 2;  ++j2)
            if (E[k].pT[j2]) {
                x = DOT(E[k].n[j2], s[0]);
                if (fabs(x) > 1.e-10)
                    fprintf(outfile, "    ERROR:  edge normal not normal:  "
                      "%13.4e\n", x);
                }
    }
#endif

#ifndef MSDOS
    physvar_time = SECOND() - t1;
#endif

  } /* end function physical_variables() */

#endif /* TRI ...............................................................*/





#ifndef MSDOS  /*............................................................*/

/*************************/
/* Function write_tess() */
/*************************/

void write_tess( void )
{
    int            i, j, total_bytes, numDirE=0;
    double         t1, *coord, *endbuf, *p;
    struct POINT   **pHeap;
    FILE           *tessfile;


/*--------------------------------------------------------------------------
    Allocate buffer in which to copy coordinates, open first output file,
    write karr[] and rarr[], copy coordinates, write buffer [all in macro],
    and close output file and free buffer.
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    sprintf(filename, "d%03dtes%04d", d_version, numV);
    file_type = TESS;

    MALLOC_OPEN_WRITEPTS(13*numV+4)

/*--------------------------------------------------------------------------
    Write coordinates of pairs of vertices; only write connection if neigh-
    bor vertex has higher index than current index (avoids duplicates).
    Then close first (tesselation) output file.
  --------------------------------------------------------------------------*/

    endbuf = coord + 13*numV - 1;   /* ptr to last "official" spot in buffer */
    p = coord;

    for (i = 0;  i < numDirV-1;  ++i)   /* loop over Dirichlet vertices */
        for (j = 0;  j < 3;  ++j)       /* loop over neighboring vertices */
            /* compute index of neighbor; if greater than index (i) of current
             * vertex, connection hasn't yet been written, so do it now (if
             * neighbor is NULL, "index" will be less than 0, so never written)
             */
            if ((DirV[i].pNbrDirV[j] - DirV) > i) {
                *p++ = DirV[i].x;
                *p++ = DirV[i].y;
                *p++ = DirV[i].pNbrDirV[j]->x;
                *p++ = DirV[i].pNbrDirV[j]->y;
                ++numDirE;
                if (p > endbuf) {   /* full:  write block */
                    total_bytes += fwrite(coord, sizeof(double), p-coord,
                      tessfile) * sizeof(double);
                    p = coord;
                }
            }

    if ((p - coord) > 0)   /* write last, partial block */
        total_bytes += fwrite(coord, sizeof(double), p-coord, tessfile) * 
                                     sizeof(double);
    fclose(tessfile);
    printf("Total bytes written to [ %s ]:  %ld\n", filename, total_bytes);
    fprintf(outfile, "Total bytes written to [ %s ]:  %ld\n", filename,
      total_bytes);

/*--------------------------------------------------------------------------
    Open second (triangulation) output file and again write karr/rarr[].
    Loop through coordinates of pairs of points:  only write connection if
    contiguous point has higher index than current point (avoids duplicates).
    Then close second file, free allocated space, and exit.
  --------------------------------------------------------------------------*/

    sprintf(filename, "d%03dtri%04d", d_version, numV);
    file_type = TRI2D;

    if ((tessfile = fopen(filename,"wb")) == NULL) {
        printf("error:  couldn't open file [ %s ]\n", filename);
        free(coord);
        return;
    }

    total_bytes = fwrite(karr, sizeof(int), 256, tessfile) * sizeof(int);
    total_bytes += fwrite(rarr, sizeof(double), 256, tessfile) * sizeof(double);

    /* loop through points; skip over fake ones */
    for (p=coord, numDirE=0, i = num_fake;  i < numPt-1;  ++i) {
        pHeap = Pt[i].pHeap + 1;
        /* loop over contiguous points */
        for (j = 0;  j < Pt[i].num_contig;  ++j, ++pHeap)
            /* compute index of contiguous point; if greater than index of
             * current point, connection hasn't yet been written, so do it now
             */
            if ((*pHeap - Pt) > i) {
                *p++ = Pt[i].x;
                *p++ = Pt[i].y;
                *p++ = (*pHeap)->x;
                *p++ = (*pHeap)->y;
                ++numDirE;  /* well, really numE */
                if (p > endbuf) {   /* full:  write block */
                    total_bytes += fwrite(coord, sizeof(double), p-coord,
                      tessfile) * sizeof(double);
                    p = coord;
                }
            }
    }

    if ((p - coord) > 0)   /* write last, partial block */
        total_bytes += fwrite(coord, sizeof(double), p-coord, tessfile) * 
                                     sizeof(double);
    fclose(tessfile);
    free(coord);
    printf("Total bytes written to [ %s ]:  %ld\n", filename, total_bytes);
    fprintf(outfile, "Total bytes written to [ %s ]:  %ld\n", filename,
      total_bytes);

    if (numDirE != numE)
        printf("error:  saved numE (%d) should match real numE (%d)\n",
          numDirE, numE);

    write_time += SECOND() - t1;

} /* end function write_tess() */

#endif /* !MSDOS ............................................................*/





#if defined(EGA)
#  include "plotload.c"   /* was called "plot_del.c" */
#endif
