/*---------------------------------------------------------------------------

            FFFFFFFF LL        OOOO    PPPPPP    PPPPPP   YY    YY
            FF       LL      OO    OO  PP    PP  PP    PP  YY  YY
            FFFFFF   LL      OO    OO  PP    PP  PP    PP   YYYY
            FF       LL      OO    OO  PPPPPP    PPPPPP      YY
            FF       LL      OO    OO  PP        PP          YY
            FF       LLLLLLL   OOOO    PP        PP          YY

  ---------------------------------------------------------------------------

  floppy.c

  This program reads in a Delaunay triangulation from disk (representing
  a sheet of gas) and evolves it hydrodynamically, updating edges by swap-
  ping as needed.  The evolution is based on a fixed Plummer gravitational
  potential, hydrodynamic pressure (isothermal gas), and viscosity.  Hydro
  optionally saves a binary-format data file to disk at a user-selected
  interval; it may be plotted with DISSPLA, using program TPLOT on galileo.

  Note that "vertices" in this program are different entities from those
  used in disc.c, which are here denoted "Dirichlet vertices."  The exact 
  correspondences are as follows:

     floppy.c               disc.c
     -------                ----------
     triangles, T[]         Dirichlet vertices, DirV[]
     vertices, V[]          points, Pt[]
     edges, E[]             contiguities between points (in Heap[])

  ---------------------------------------------------------------------------

     Version:  Y-MP, PostScript-graphic              Greg Roelofs
     Source:   f332.c on eagle                       25 May 1992

  ---------------------------------------------------------------------------

  To compile:

     scc -c [-DUDEBUG] [-DNOPLOT] f332.c
     segldr f332.o update332.o grav.o jacobi.o repair.o sh310.o -o f332

     UDEBUG:  compile with debugging version of update_points() (updated.c)
     NOPLOT:  disable writing of triangulation data to disk for plots

  ---------------------------------------------------------------------------

  To run:  f332 [-n#] [-t#] [-v[#]] [filename]

     -n#        (optional) Run for # time steps (int).  Default is 32.
     -t#        (optional) Change size of time step to # (double).  This
                  becomes the new time step for subsequent runs as well.
                  Default is set in the loader (d###.c).
     -v[#]      (optional) Save an edges plot-file to disk every # steps
                  (int).  If # is not present, file will be saved every
                  32 steps.  Default is not to write edges files.
     filename   (optional) Name of dataset to be evolved.  Default name
                  is "d332f000t000.0".  Output dataset will be similarly
                  named, but with the floppy version-number and time fields
                  changed appropriately.

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotload.c:    contains MS-DOS graphics routines

     points.c:      stand-alone program to create and plot points only

  -> f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

  Version history:

     h250.c:  corrected viscosity code (yet again, sigh:  dynamic viscosity,
               isothermal scale height); changed to use Tom's units, exponen-
               tial disk, and oblate, logarithmic (Schwarzschild) potential
     h251.c:  corrected gravity bug (missing mass); added print_phys_vars()
     h252.c:  added write_movie_data(); changed "time" to "Time" to allow
               use of system date/time routines in new subroutine; put in
               hack to simulate d255.c load from d254.c loadfile (constant
               scale height)
     h260.c:  changed data structures (added attribute field to vertex
               struct); changed write_movie_data() to write attributes as
               well; changed name of edges file to include loader version
     h261.c:  corrected gravitational energy calculation
     h270.c:  reverted to spherical Plummer potential
     h271.c:  minor tweak to movie.out; moved edg2vax conversion code into
               write_edges() (must now link with vaxwr.o); changed declara-
               tion of init() to static to avoid conflicts with init in
               libio.a
     h272.c:  [debug/test version] added new diagnostics to check for grid
               tangling; removed write_std_data() (inline in update_points());
               added repair_grid() routine to swap overrunning points
     h280.c:  completely reorganized grid-repair routines, swap_edges() in-
               cluded; moved grid-repair routines into separately-compilable
               file (repair.c); made update280.c separately compilable; split
               a bunch of defines, includes, prototypes, etc., into separate
               header file (h280.h)
     h281.c:  added write_tri() routine to save triangles for 3D rendering;
               modified write_movie_data() to indicate steps written; fixed
               vertex-mass bug (central masses not swapped) and edge-normal
               bug (normals not recalculated [==> rewrite], edge-lengths not
               recalculated) in swap_points()
     h282.c:  fixed folded-triangle mass bug (areas no longer preserved) in
               swap_edges(); changed header filename to hydro.h
     h285.c:  changed "VAX" to "PLOT"; accessing shared routines in sh285.c
               (separately compilable); rarr[] rearranged; forces calculated
               in FORTRAN routine GRAV(); disk-write times combined; moved
               save_data() to write_data() in sh285.c
     h286.c:  fixed viscosity bug in update_points() and dbgfile bug
     h287.c:  unrolled loops in update_points()
     f290.c:  added "phantom" ring for better boundary-condition handling;
               changed name from "hydro" to "floppy"; changed header file-
               name to fd.h
     f291.c:  changed definition of normal at boundary for pressure, viscous
               force calculations
     f292.c:  modified for use with new output routines (for Cray tplot)
     f300.c:  fixed pressure calculation (removed 2nd-order effect); changed
               data structures (added pressure field to triangle struct);
               fixed folded-edge criterion; changed default plot interval to
               32 steps
     f310.c:  changed numRings usage to conform to d310; other modifications
               for compiling with v3.10 code
     f320.c:  changed potential to harmonic oscillator; modified command-line
               switch use
     f330.c:  added local gravitational restoring force to update_points()
     f331.c:  changed restoring force from that of infinitely thin sheet to
               that of isothermal disk (sech^2 z)
     f332.c:  reduced restoring-force scale height to 120 pc (viscosity not
               changed); added jacobi timing info

     f???.c:  [wanted to...] renamed output and debug files to form "d###f###.*"

  ---------------------------------------------------------------------------*/





/********************************/
/*  Local Defines and Includes  */
/********************************/

#ifndef NOPLOT           /* enable writing of edges data to disk */
#  define PLOT           /*   for plotting with DISSPLA          */
#elif defined(PLOT)      /* disable plotting:  makes for quicker */
#  undef PLOT            /*   compilation and smaller executable */
#endif

#include "fd.h"          /* includes, defines, typedefs, prototypes */

#include <time.h>        /* time(), localtime(), strftime() */
#include <sys/types.h>   /* rename() */
#include <sys/stat.h>    /* rename() */

#define FLOPPY_VERSION   332
#define NUM_STEPS        32
#define NUM_EDGEPLOT     32

static void   init       ( int, char** );




/********************************/
/*  Global and Local Variables  */
/********************************/

char     filename[80];

int      errorlevel=0, num_swaps, *const ring_last=pRing_last;
int      karr[256];

double   rarr[256];
double   update_time=0., grav_time=0., jacobi_time=0., repair_time=0.;
double   swappts_time=0., swapE_time=0., write_time=0.;
double   *const LinMom=pLinMom, *const AngMom=pAngMom;
double   *const ring_last_n=pRing_last_n, *const ring_R=pRing_R;
double   *ring_incl, *ring_v, *ring_azim;    /* all three vary */

vector            *cm_phantom, s[3], cm_subT[MAX_subT][3];
four_vector       *accel_phantom, accel_potent[MAX_subT][3];

struct edge       E[ MAX_E ];
struct triangle   T[ MAX_T ];
struct vertex     V[ MAX_V ];

FILE              *infile, *outfile, *dbgfile;
FILE              *moviefile, *karrfile, *moviejunk;

static int        edgeplot=1, num_edgeplot=NUM_EDGEPLOT, num_steps=NUM_STEPS;
static int        save_config=FALSE;
static double     temp_dt=0., init_time, read_time, evolve_time;
static double     total1, total_time;





/******************/
/*  Main program  */
/******************/

main( int argc, char *argv[] )
  {
    total1 = SECOND();   /* begin timing entire program execution */

    init(argc, argv);
    read_data();
    evolve();
    write_data();

    fprintf(outfile, "\n\nCPU timing information (totals, in seconds):\n");
    fprintf(outfile, "  init()                        %11.6f\n", init_time);
    fprintf(outfile, "  read_data()                   %11.6f\n", read_time);
    fprintf(outfile, "  evolve()                      %11.6f\n", evolve_time);
    fprintf(outfile, "     update_points()            %11.6f\n", update_time);
    fprintf(outfile, "        GRAV()                  %11.6f\n", grav_time);
    fprintf(outfile, "        grr_jacobi()            %11.6f\n", jacobi_time);
    fprintf(outfile, "     repair_grid()              %11.6f\n", repair_time);
    fprintf(outfile, "        swap_points()           %11.6f\n", swappts_time);
    fprintf(outfile, "        swap_edges()            %11.6f\n", swapE_time);
    fprintf(outfile, "  write_edges/tri/movie/data()  %11.6f\n", write_time);

    total_time = SECOND() - total1;
    fprintf(outfile, "\nTotal program CPU time:         %11.6f\n", total_time);
    fprintf(outfile, "-------------------------------------------\n\n\n\n");
    fclose(outfile);

    printf("\nProgram completed, possibly successfully.  Have a nice day.\n");
  }





/*********************/
/*  Function init()  */
/*********************/

static void init( int argc, char *argv[] )
  {
    static char   *pf = filename;
    static int    error = 0;
    int           c, c1, n;
    double        t1;



/*---------------------------------------------------------------------------
    Open "standard" output file first.  (Movie file is opened in its own
    subroutine, since its name depends on data which has not yet been read.)
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

#ifdef UDEBUG
    sprintf(filename, "f%03dd.out", FLOPPY_VERSION);
#else
    sprintf(filename, "f%03d.out", FLOPPY_VERSION);
#endif

    if ((outfile = fopen(filename,"a")) == NULL) {
        printf("\nerror:  couldn't open output file [ %s ]\n", filename);
        exit(1);
    }

#ifdef UDEBUG
    sprintf(filename, "f%03dd.dbg", FLOPPY_VERSION);
#else /* !UDEBUG */
    sprintf(filename, "f%03d.dbg", FLOPPY_VERSION);
#endif /* ?UDEBUG */

    if ((dbgfile = fopen(filename,"a")) == NULL) {
        printf("\nerror:  couldn't open debug file [ %s ]\n", filename);
        exit(1);
    }

/*---------------------------------------------------------------------------
    Now parse command line for '-' options.
  ---------------------------------------------------------------------------*/

    if ( argc <= 1 )
      --argc;
    else
      while ( --argc > 0  &&  (*++argv)[0] == '-')
        while ( c = *++(argv[0]) )   /* argv was inc => argv[0] not prognam */
          switch (c) {
              case 'v':      /* edge-plot (standard:  x, y, z views) */
                edgeplot = 1;
                /* get char after 'v' without changing pointer */
                c1 = *(argv[0] + 1);
                if ((c1 >= '0') && (c1 <= '9')) {
                    n = atoi( argv[0]+1 );   /* pointer to digit string */
                    *(argv[0]+1) = '\0';     /* overwrite first digit   */
                    if (n > 0)
                        num_edgeplot = n;
                    /* GRR:  make (n == 0) ==> last-frame plot only */
                } else if (c1 == '-')
                    edgeplot = 0;
                break;

              case '0':
              case '1':
              case '2':
              case '3':
              case '4':
              case '5':
              case '6':
              case '7':
              case '8':
              case '9':
                n = atoi( argv[0] );
                *(argv[0]+1) = '\0';   /* overwrite second digit */
                if (n > 0)
                    num_steps = n;
                break;

              case 'n':
                /* get char after 'n' without changing pointer */
                c1 = *(argv[0] + 1);
                if ((c1 >= '0') && (c1 <= '9')) {   /* GRR:  use isdigit() */
                    n = atoi( argv[0]+1 );   /* pointer to digit string */
                    *(argv[0]+1) = '\0';     /* overwrite first digit   */
                    if (n > 0)
                        num_steps = n;
                } else {
                    printf("f%03d.c:  option [ -n ] requires numeric argument"
                      " (type int)\n", FLOPPY_VERSION);
                    ++error;
                }
                break;

	      case 't':
                /* get char after 't' without changing pointer */
		c1 = *(argv[0] + 1);
		if (((c1 >= '0') && (c1 <= '9'))  ||  c1 == '.') {
		    temp_dt = atof( argv[0]+1 );  /* pointer to digit string */
		    *(argv[0]+1) = '\0';          /* overwrite first digit   */
                    /* set delta_t = temp_dt after rarr[] read in read_data() */
		} else {
                    printf("f%03d.c:  option [ -t ] requires numeric argument"
                      " (type double)\n", FLOPPY_VERSION);
                    ++error;
		}
		break;

              default:
                printf("f%03d.c:  invalid option [ -%c ]\n", FLOPPY_VERSION, c);
                ++error;
                *(argv[0]+1) = '\0';   /* terminate this argument; try next */
                break;
          }  /* end switch */

/*---------------------------------------------------------------------------
    If any invalid options were encountered, print syntax synopsis.
  ---------------------------------------------------------------------------*/

    if (error) {
        printf("Usage:  f%03d [-[n]#] [-t#] [-v[#]] [datafile]\n", 
          FLOPPY_VERSION);
        printf("  -n#        run for # steps (int:  default = %d steps)\n",
          NUM_STEPS);
        printf("  -t#        size of timestep (double:  default set in loader"
               " program)\n             note:  # becomes new default for "
               "subsequent runs\n");
        printf("  -v-        don't make plot files\n");
        printf("  -v#        save plot file to disk every # steps (int:  "
               "default = %d steps)\n", NUM_EDGEPLOT);
        printf("  datafile   read data from file [ datafile ] (default file "
               "is [ d%03df000t000.0 ])\n\n", FLOPPY_VERSION);
    }

/*---------------------------------------------------------------------------
    Any argument remaining (i.e., not preceded by '-') must be the name of
    the data file to be opened, so do that now.  Otherwise, if no arguments
    left, open the default data file (of form "d###f000t000.0", where ### is
    the version number of this program).
  ---------------------------------------------------------------------------*/

    if (argc > 0)
        while (*pf++ = *argv[0]++)   /* copy argument into filename[] */
            ;
    else
        sprintf(filename, "d%03df000t000.0", FLOPPY_VERSION);

    printf("\nOpening data file [ %s ] for input.\n", filename);
    fprintf(outfile, "Opening data file [ %s ] for input.\n", filename);
    if ((infile = fopen(filename,"rb")) == NULL) {
        printf("\nerror:  couldn't open file [ %s ]\n", filename);
        exit(1);
    }

    printf("  length of run:  %d steps\n", num_steps);
    fprintf(outfile, "  length of run:  %d steps\n", num_steps);
    if (edgeplot) {
        printf("  edge plot every %d steps\n", num_edgeplot);
        fprintf(outfile, "  edge plot every %d steps\n", num_edgeplot);
    } else {
        printf("  no edge plots\n");
        fprintf(outfile, "  no edge plots\n");
    }

    init_time = SECOND() - t1;

  } /* end function init() */





/**************************/
/*  Function read_data()  */
/**************************/

void read_data( void )
{
    int               num_rd, total_bytes, i, j, index;
    double            t1;
    struct edge       *pE;
    struct triangle   *pT;
    struct vertex     *pV;



/*---------------------------------------------------------------------------
    Read triangulation data.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    total_bytes = fread( karr, sizeof(int), 256, infile ) * sizeof(int);
    f_version = FLOPPY_VERSION;

    if ( d_version/10 != f_version/10 )
        printf("\nwarning:  attempting to read version %d.%.2d load data\n"
          "        (this is version %d.%.2d of floppy.c)\n\n",
          d_version/100, d_version%100, f_version/100, f_version%100);

    total_bytes += fread( rarr, sizeof(double), 256, infile ) *
                                sizeof(double);
    cm_phantom = cm_subT[numT];             /* initialize update.c pointers */
    accel_phantom = accel_potent[numT];

    /* GRR: just malloc these instead? (write after karr/rarr, before rest) */
    ring_incl = ring_R +   (numRings+1);   /* pointers to sub-array */
    ring_v    = ring_R + 2*(numRings+1);   /* ditto */
    ring_azim = ring_R + 3*(numRings+1);   /* ditto */

    if (temp_dt > 0.)
        delta_t = temp_dt;
    printf("  delta t:  %8.3f units\n\n", delta_t);
    fprintf(outfile, "  delta t:  %8.3f units\n\n", delta_t);
    fprintf(outfile, "Points version %d.%02d   Loader version %d.%02d   "
      "Evolver version %d.%02d\n", p_version/100, p_version%100,
      d_version/100, d_version%100, f_version/100, f_version%100);

    total_bytes += fread( &pE, sizeof(struct edge *), 1, infile ) *
                               sizeof( struct edge * );
    total_bytes += fread( &pT, sizeof(struct triangle *), 1, infile ) *
                               sizeof( struct triangle * );
    total_bytes += fread( &pV, sizeof(struct vertex *), 1, infile) *
                               sizeof( struct vertex * );

    num_rd = fread( E, sizeof(struct edge), numE, infile);
    fprintf(outfile, "  number edges read:       E[ %d ]\t(expect %d)\n",
      num_rd, numE);
    total_bytes += num_rd * sizeof(struct edge);

    num_rd = fread( T, sizeof(struct triangle), numT, infile);
    fprintf(outfile, "  number triangles read:   T[ %d ]\t(expect %d)\n",
      num_rd, numT);
    total_bytes += num_rd * sizeof(struct triangle);

    num_rd = fread( V, sizeof(struct vertex), numV, infile);
    fprintf(outfile, "  number vertices read:    V[ %d ]\t(expect %d)\n",
      num_rd, numV);
    total_bytes += num_rd * sizeof(struct vertex);

    fprintf(outfile, "Total bytes read from [ %s ]:  %d\n", filename,
      total_bytes);

    fclose(infile);

/*  print out all initialized physical values, but only at start of expmt,  */
/*  not for a continuation run                                              */
    if (!iteration)
        print_phys_vars();

/*---------------------------------------------------------------------------
    Convert pointer references from loader addresses (disc.c) to ones
    relative to this program (floppy.c).
  ---------------------------------------------------------------------------*/

    for ( i = 0  ;  i < numE  ;  i++ ) {
        E[i].n[0] = (double *) ((int) E[i].n[0] + (int) T - (int) pT);
        E[i].pT[0] = (struct triangle *) ((int) E[i].pT[0] + (int) T
                                                                 - (int) pT);
        if (E[i].pT[1]) {   /*  may be NULL: only adjust if not  */
            E[i].n[1] = (double *) ((int) E[i].n[1] + (int) T - (int) pT);
            E[i].pT[1] = (struct triangle *) ((int) E[i].pT[1] + (int) T
                                                                 - (int) pT);
        }
        E[i].pV[0] = (struct vertex *) ((int) E[i].pV[0] + (int) V - (int) pV);
        E[i].pV[1] = (struct vertex *) ((int) E[i].pV[1] + (int) V - (int) pV);
    }

    for ( i = 0  ;  i < numT  ;  i++ ) {
        T[i].pV[0] = (struct vertex *) ((int) T[i].pV[0] + (int) V - (int) pV);
        T[i].pV[1] = (struct vertex *) ((int) T[i].pV[1] + (int) V - (int) pV);
        T[i].pV[2] = (struct vertex *) ((int) T[i].pV[2] + (int) V - (int) pV);
        T[i].pE[0] = (struct edge *) ((int) T[i].pE[0] + (int) E - (int) pE);
        T[i].pE[1] = (struct edge *) ((int) T[i].pE[1] + (int) E - (int) pE);
        T[i].pE[2] = (struct edge *) ((int) T[i].pE[2] + (int) E - (int) pE);
    }

/*---------------------------------------------------------------------------
    Check pointer references for bad load (exclude NULLs from E[].pT[]
    check, since boundary edges only point to one triangle).
  ---------------------------------------------------------------------------*/

    for ( i = 0  ;  i < numE  ;  i++ ) {
        for ( j = 0  ;  j < 2  ;  j++ ) {
            index = E[i].pT[j] - T;
            if (E[i].pT[j]  &&  ((index < 0) || (index >= numT)))
                printf("E[%d].pT[%d] -> T[%d]\n", i, j, index);
        }
        for ( j = 0  ;  j < 2  ;  j++ ) {
            index = E[i].pV[j] - V;
            if ((index < 0) || (index >= numV))
                printf("E[%d].pV[%d] -> V[%d]\n", i, j, index);
        }
    }

    for ( i = 0  ;  i < numT  ;  i++ ) {
        for ( j = 0  ;  j < 3  ;  j++ ) {
            index = T[i].pV[j] - V;
            if ((index < 0) || (index >= numV))
                printf("T[%d].pV[%d] -> V[%d]\n", i, j, index);
        }
        for ( j = 0  ;  j < 3  ;  j++ ) {
            index = T[i].pE[j] - E;
            if ((index < 0) || (index >= numE))
                printf("T[%d].pE[%d] -> E[%d]\n", i, j, index);
        }
    }

    read_time = SECOND() - t1;

} /* end function read_data() */





/**********************/
/*  Function evolve() */
/**********************/

void evolve( void )
  {
    int      i;
    double   t1;



/*---------------------------------------------------------------------------
    Open movie files (one for actual binary data, one for info).  If this
    is a new run (iteration = 0), write karr, rarr, and initial vertex
    positions to movie file.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    if (!iteration)   /* first step:  open files and write karr, rarr, verts */
        write_movie_data(1);
    else              /* any step but first:  open files only */
        write_movie_data(2);

    fprintf(outfile, "\nStep%6d     Time%9.3f\n", iteration, Time);

/*---------------------------------------------------------------------------
    MAIN LOOP:  update positions and velocities of vertices, swap edges to
    maintain triangulation, and write vertex positions to movie file.  Also
    write configuration to edges file every num_edgeplot steps, if edge
    plotting is enabled.
  ---------------------------------------------------------------------------*/

    for (i = 0;  i < num_steps;  i++) {
/*      if (!((i+1) % (num_steps/4)))  */
        if (!((i+1) % 16))
            printf("  [step %d]\n", iteration+1);
        update_points();
        repair_grid();
        write_movie_data(0);

#ifdef PLOT /*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
        if (edgeplot && !(iteration%num_edgeplot)) {
            write_edges();
            write_tri();
        }
#endif /*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

        fflush(outfile);
        fflush(dbgfile);
    }

/*---------------------------------------------------------------------------
    All done:  close movie files and return for cleanup.
  ---------------------------------------------------------------------------*/

    write_movie_data(3);
    evolve_time = SECOND() - t1;

  } /* end function evolve() */





/*********************************/
/*  Function write_movie_data()  */
/*********************************/

void write_movie_data( int open_close )
  {
    char          now[40];
    int           k, *q;
    double        t1, *coords, *p;
    time_t        CurTim, *pCurTim=&CurTim;
    static char   moviefn[80], karrfn[80];
    static int    nsteps=0, total_bytes=0, first_iter;



/*--------------------------------------------------------------------------
    Ordinarily, just write another block of vertex data (binary format) to
    movie file and karr/rarr to karrmovie file (open_close = 0).
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    if (!open_close) {
        coords = (double *) malloc((unsigned) (3*numV)*sizeof(double));
        if (!coords) {
            printf("error:  couldn't allocate memory to write [ %s ]\n",
              moviefn);
            exit(1);
        }
        for (p = coords, ++nsteps, k = 0;  k < numV;  ++k) {
            *p++ = V[k].r[0];
            *p++ = V[k].r[1];
            *p++ = V[k].r[2];
        }
        total_bytes += fwrite(coords, sizeof(double), 3*numV, moviefile) *
                                      sizeof(double);
        free(coords);

        file_type = KARRMOVIE;
        fwrite(karr, sizeof(int), 256, karrfile) * sizeof(int);
        fwrite(rarr, sizeof(double), 256, karrfile) * sizeof(double);
    }

/*--------------------------------------------------------------------------
    Close movie file, make notation in movie information file about number
    of bytes written and date/time at which this was done, and then close
    that file as well.
  --------------------------------------------------------------------------*/

    else if (open_close == 3) {
        fclose(moviefile);
        fclose(karrfile);

        time(pCurTim);
        strftime(now, 40, "%a %Y.%m.%d %H:%M:%S", localtime(pCurTim));
        fprintf(moviejunk, "%s: %3d steps, %7d bytes: %4d - %4d\t%s\n",
          moviefn, nsteps, total_bytes, first_iter, iteration, now);
        fclose(moviejunk);
    }

/*--------------------------------------------------------------------------
    Open both the normal binary movie file and the ASCII movie information
    file for appending, and remember the current step number.  If binary
    file is being created, also write karr, rarr, attributes, and initial
    vertex positions.
  --------------------------------------------------------------------------*/

    else {
        sprintf(moviefn, "d%03df%03d.movie", d_version, f_version);
        if ((moviefile = fopen(moviefn,"ab")) == NULL) {
            printf("error:  couldn't open file [ %s ]\n", moviefn);
            exit(1);
        }
        sprintf(karrfn, "d%03df%03d.karr", d_version, f_version);
        if ((karrfile = fopen(karrfn,"ab")) == NULL) {
            printf("error:  couldn't open file [ %s ]\n", karrfn);
            exit(1);
        }
        if ((moviejunk = fopen("movie.out","a")) == NULL) {
            printf("error:  couldn't open file [ movie.out ]\n");
            exit(1);
        }
        if (open_close == 1) {   /* have just created movie file */
            time( pCurTim );
            strftime(now, 40, "%a %Y.%m.%d %H:%M:%S", localtime(pCurTim));
            fprintf(moviejunk, "%s:  created\t\t\t\t%s\n", moviefn, now);

            file_type = MOVIE;
            total_bytes += fwrite(karr, sizeof(int), 256, moviefile) *
                                        sizeof(int);
            total_bytes += fwrite(rarr, sizeof(double), 256, moviefile) *
                                        sizeof(double);

            coords = (double *) malloc((unsigned) (3*numV)*sizeof(double));
            if (!coords) {
                printf("error:  couldn't allocate memory to write [ %s ]\n",
                  moviefn);
                exit(1);
            }
            for (q = (int *)coords, k = 0;  k < numV;  ++k)
                *q++ = V[k].attrib;
            total_bytes += fwrite((int *)coords, sizeof(int), numV, moviefile)
                                               * sizeof(int);
            for (p = coords, ++nsteps, k = 0;  k < numV;  ++k) {
                *p++ = V[k].r[0];
                *p++ = V[k].r[1];
                *p++ = V[k].r[2];
            }
            total_bytes += fwrite(coords, sizeof(double), 3*numV, moviefile) *
                                          sizeof(double);
            free(coords);

            file_type = KARRMOVIE;
            fwrite(karr, sizeof(int), 256, karrfile);
            fwrite(rarr, sizeof(double), 256, karrfile);
            first_iter = iteration;
        } else
            first_iter = iteration + 1;
    } /* end if (open_close) */

    write_time += SECOND() - t1;

  } /* end function write_movie_data() */
