C----------------------------------------------------------------------------
C
      subroutine GRAV( NUMPTS )
C
C     This routine calculates the gravitational potential and gravitational
C     acceleration at a set of positions given by the CMSUBT common area.
C     CMSUBT is a globally declared array in the calling C program, as is
C     ACCPOT, the return data.  NUMPTS is the number of positions to compute.
C     Not compatible with hydro versions before version 2.90.
C
C----------------------------------------------------------------------------
C
C        Version:  Y-MP/PC                              Greg Roelofs
C        Source:   grav.f on eagle                      19 March 1992
C
C----------------------------------------------------------------------------
C
C     To compile:
C
C        on Y-MP:  cft77 -i64 grav.f
C        on PC:    fl -c -AS -G2 -FPi87 -Ox -4Nt grav.f   (.for?)
C
C----------------------------------------------------------------------------
C 
C     Associated files:
C
C        fd.h:          main header file:  common macros, prototypes, etc.
C        sh###.c:       routines shared by both loader and evolver
C     -> grav.f:        gravity code (used by update###.c, d###.c)
C        [vaxwr.f:      routine for conversion of edge data to VMS format]
C
C        d###.c:        main tesselation/triangulation code; loader
C        tesselate.c:   tesselation routines called by d###.c
C        plotload.c:    contains MS-DOS graphics routines
C
C        points.c:      stand-alone program to create and plot points only
C
C        f###.c:        main evolution code (evolver)
C        update###.c:   physics (update_points, used only by f###.c)
C        jacobi.c:      ring-fitting code (used only by update###.c)
C        repair.c:      grid-maintenance routines (used only by f###.c)
C
C        update###d.c, repaird.o, f###d:  debugging versions
C
C----------------------------------------------------------------------------
C
C      Version history:
C
C        h285.c:  gravitational potential/acceleration code first split 
C                  out into separately compiled FORTRAN source file (based
C                  on SCHWAR.FOR, VAX plotting program from v2.50--2.62)
C        f290.c:  added NUMPTS argument to allow variable number of posi-
C                  tions to be calculated
C        f310.c:  changed NUM_POINTS to 4800
C        f311.c:  small optimizations (division -> multiplication)
C        f???.c:  modified to be a driver for RHM potential-solver code
C
C----------------------------------------------------------------------------



      implicit  none
      integer   NUM_POINTS, MAX_subT
      parameter (NUM_POINTS=4800)
      parameter (MAX_subT=2*(NUM_POINTS+20)+NUM_POINTS/10)

      integer   i, k, numpts
c     integer   karr(0:255)
      real      rarr(0:255), Rcore, c1half, c2, phi0
      real      cm_subT(3, 3*MAX_subT), accel_potent(4, 3*MAX_subT)
      real      r(3), x, y, z, g(3)
      real      zsqr, rsqr, rabs, root, logg1, logg, numer, denom

c     common /KARR/ karr
      common /RARR/ rarr
      common /CMSUBT/ cm_subT
      common /ACCPOT/ accel_potent

      equivalence (r(1), x)
      equivalence (r(2), y)
      equivalence (r(3), z)
      equivalence (rarr(23), Rcore)
      equivalence (rarr(25), phi0)
      equivalence (rarr(26), c1half)
      equivalence (rarr(27), c2)

C----------------------------------------------------------------------------
C     Loop through positions and calculate Schwarzschild potential unless
C     too close to origin.  (See physical_variables() routine of d285.c for
C     reference to paper by David et al.)  All working variables except phi0
C     and Rcore are dimensionless.
C----------------------------------------------------------------------------

      g(1) = 0.
      g(2) = 0.
      Rcoreinv = 1. / Rcore
      phi0Rcore = phi0 * Rcoreinv

      do 100 k = 1, numpts
          do 10 i = 1, 3
   10         r(i) = cm_subT(i, k) * Rcoreinv
          zsqr = z**2
          rsqr = x**2 + y**2 + zsqr
          rsqrinv = 1. / rsqr
          rabs = sqrt(rsqr)
          rabsinv = 1. / rabs

          if (rabs .gt. 1.e-8) then
              root = sqrt(1. + rsqr)
              logg1 = alog(rabs+root) * rabsinv
              logg = logg1 - 1./root
              numer = -2. + c2*(rsqr-9.*zsqr)
              denom = 1. + c2*rsqr
              g(3) = 6.*denom
c             v_circ(k) = sqrt(logg)

C           -----------------------------------------------------------------
C             Calculate value of the potential at the given position, then
C             loop and do accelerations [see notes 1991.3.24].
C           -----------------------------------------------------------------

              accel_potent(4, k) = phi0 * (logg1 -
     +          (3.*zsqr*rsqrinv - 1.)*c1half*rsqr * denom**(-1.5))
C             now calculate the acceleration [notes 1991.3.24]:
              do 50 i = 1, 3
   50             accel_potent(i, k) = r(i) * phi0Rcore * (logg*rsqrinv
     +              +  c1half*(numer+g(i)) * denom**(-2.5))
          else
              accel_potent(4, k) = phi0
              do 60 i = 1, 3
   60             accel_potent(i, k) = 0.
          endif

  100 continue

      return
      end
