#include <stdio.h>
#include <math.h>
#include <malloc.h>

#ifdef M_PI
#  define PI180      (M_PI/180.)
#else
#  define PI180      (3.14159265358979323846264/180.)
#endif

#ifdef CRAY
#  define SQR(x)   powri((x),2)
#else
#  define SQR(x)   ((x)*(x))
#endif

#define ROTATE(a,i,j,k,l)   {g=a[n*i+j];  h=a[n*k+l];  a[n*i+j]=g-s*(h+g*tau);\
                             a[n*k+l]=h+s*(g-h*tau);}

extern double    jacobi_time;

fortran double   SECOND    (void);
void             nrerror   (char *);

void grr_jacobi(double **real_sym, int n, double *e_val,
                double **e_vec, int *nrot)
/*
 * Computes all eigenvalues and eigenvectors of a real, symmetric, n x n
 * matrix real_sym[][].  On output, elements of real_sym above the diagonal
 * are destroyed.  e_val[] returns the n eigenvalues of real_sym.  e_vec[][]
 * is a matrix whose *rows* contain, on output, the normalized eigenvectors
 * of real_sym; that is, e_vec[] is an array of eigenvectors.  nrot returns
 * the number of Jacobi rotations that were required.
 *
 * This routine is from Numerical Recipes in C, section 11.1, pp. 364-366.
 * It has been modified in the following ways [GRR, 1991.07.24]:
 *    (1) all data items are doubles, not floats
 *    (2) all n-dimensional vectors, etc., run from 0 to n-1, instead of
 *         the original, pseudo-FORTRAN fashion of NRiC
 *    (3) the output eigenvector matrix has been transposed to conform to
 *         normal C usage instead of pseudo-FORTRAN usage; see note about
 *         e_vec[] above [I profoundly disagree with their decision on this
 *         point]
 *    (4) the inefficient routines {d}vector() and free_{d}vector() have 
 *         been replaced with straight calls to malloc() and free() [the 
 *         vector() routines were potentially dangerous as well, since they
 *         pulled some sneaky stuff with pointers and allocation sizes]
 *    (5) various minor touch-ups, including more-helpful variable names
 * Additional modification [GRR, 1992.05.25]:
 *    (6) CPU timing information added
 */
{
    int      j, iq, ip, i;
    double   threshold, theta, tau, t, sm, s, h, g, c, *b, *z, *evec, *rsym;
    double   t1;


    t1 = SECOND();

    b = (double *) malloc((unsigned) n*sizeof(double));
    if (!b) nrerror("malloc returns NULL in jacobi (b)");

    z = (double *) malloc((unsigned) n*sizeof(double));
    if (!z) nrerror("malloc returns NULL in jacobi (z)");

    /* initialize to identity matrix */
    for (evec = (double *)e_vec, ip = 0;  ip < SQR(n);  ++ip)
        *evec++ = 0.;
    for (evec = (double *)e_vec, ip = 0;  ip < SQR(n);  ip += n+1)
        evec[ip] = 1.;

    /* initialize work pointers for last time (won't change from now on) */
    evec = (double *) e_vec;
    rsym = (double *) real_sym;

    for (ip = 0;  ip < n;  ++ip) {
        b[ip] = e_val[ip] = rsym[(n+1)*ip];
        z[ip] = 0.;
    }
    *nrot = 0;

    for (i = 0;  i < 50;  ++i) {
        sm = 0.;
        for (ip = 0;  ip < n-1;  ++ip) {
            for (iq = ip+1;  iq < n;  ++iq)
                sm += fabs(rsym[n*ip+iq]);
        }
        if (sm == 0.) {     /* the normal return, which relies on */
            free(z);        /*  quadratic convergence to machine */
            free(b);        /*  underflow */
            jacobi_time += SECOND() - t1;
            return;
        }
        if (i < 3)
            threshold = .2 * sm / SQR(n);   /* ...on first 3 sweeps */
        else
            threshold = 0.;                 /* ...thereafter */
        for (ip = 0;  ip < n-1;  ++ip) {
            for (iq = ip+1;  iq < n;  ++iq) {
                g = 100. * fabs(rsym[n*ip+iq]);
                if (i > 3 && fabs(e_val[ip])+g == fabs(e_val[ip])
                    && fabs(e_val[iq])+g == fabs(e_val[iq]))
                    rsym[n*ip+iq] = 0.;
                else if (fabs(rsym[n*ip+iq]) > threshold) {
                    h = e_val[iq]-e_val[ip];
                    if (fabs(h)+g == fabs(h))
                        t = rsym[n*ip+iq] / h;
                    else {
                        theta = .5 * h / (rsym[n*ip+iq]);
                        t = 1. / (fabs(theta) + sqrt(1.+SQR(theta)));
                        if (theta < 0.)
                            t = -t;
                    }
                    c = 1. / sqrt(1+SQR(t));
                    s = t * c;
                    tau = s/(1.+c);
                    h = t * rsym[n*ip+iq];
                    z[ip] -= h;
                    z[iq] += h;
                    e_val[ip] -= h;
                    e_val[iq] += h;
                    rsym[n*ip+iq] = 0.;
                    for (j = 0;  j < ip;  ++j)
                        ROTATE(rsym,j,ip,j,iq)
                    for (j = ip+1;  j < iq;  ++j)
                        ROTATE(rsym,ip,j,j,iq)
                    for (j = iq+1;  j < n;  ++j)
                        ROTATE(rsym,ip,j,iq,j)
                    for (j = 0;  j < n;  ++j)
/*                      ROTATE(evec,j,ip,j,iq)     <-- WORKS, but transpose */
                        ROTATE(evec,ip,j,iq,j)     /* <-- GRR:  C likes it */
                    ++(*nrot);
                }
            } /* end for-loop (iq) */
        } /* end for-loop (ip) */

        for (ip = 0;  ip < n;  ++ip) {
            b[ip] += z[ip];
            e_val[ip] = b[ip];
            z[ip] = 0.;
        }
    } /* end for-loop (i:  iterations) */

    nrerror("Too many iterations in routine JACOBI");
}

#undef ROTATE




void nrerror(char *error_text)
{
    fprintf(stderr,"Numerical Recipes run-time error...\n");
    fprintf(stderr,"%s\n",error_text);
    fprintf(stderr,"...now exiting to system...\n");
    exit(1);
}
