/*---------------------------------------------------------------------------

            PPPPPP      OOOO    IIIIII  NN    NN  TTTTTTTT  SSSSSS 
            PP    PP  OO    OO    II    NNN   NN     TT    SS     ~
            PP    PP  OO    OO    II    NN N  NN     TT     SSSSSS 
            PPPPPP    OO    OO    II    NN  N NN     TT          SS
            PP        OO    OO    II    NN   NNN     TT    SS    SS
            PP          OOOO    IIIIII  NN    NN     TT     SSSSSS 

  ---------------------------------------------------------------------------

  points.c

  This program creates a two-dimensional set of data points and writes them
  to a file.  The points are then used by DISC to create a triangulation.

  ---------------------------------------------------------------------------

     Version:  Y-MP/PC, EGA-graphic                  Greg Roelofs
     Source:   p313.c on eagle and oddjob            28 February 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:   make p313
     on PC:     cl $(FP) -AS -G2 [-DDIAG | -DEGA] p313.c

     DIAG:   print diagnostics to standard output (see global variable Eflag)
     EGA:    plot tesselation as it is created, in EGA graphics

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotpts.c:     MS-DOS graphics routines for points load
     plotload.c:    MS-DOS graphics routines for tesselation

  -> p###.c:        stand-alone program to create and plot points only

     f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      most grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

  To run:  [time] p313

     Points data will be written to p###pts####, where last four digits are
     number of points.

  ---------------------------------------------------------------------------

  Version history:

   < p310.c:  included as part of d###.c; points.c used only for debugging
     p310.c:  moved construct_points() out of d###.c and into p###.c; made
               output format match that of write_tess(); modified for shell
               structure/variable points density; increased NUM_POINTS from
               1000 to 4000; increased Rdisk from 10.7 to 11.5 to approxi-
               mate older loads
     p311.c:  modified smoothed-transition two-shell load to a continuously
               changing average-triangle-mass load (i.e., transition region
               extends to edge of disk now) [number of points half of wanted]
     p312.c:  better approximation of total number of points
     p313.c:  non-linear transition function (faster than linear)

  ---------------------------------------------------------------------------*/





/**********************************/
/*  Local Includes, Defines, etc. */
/**********************************/

#define LOADER
#include "fd.h"        /* includes, defines, typedefs, prototypes */

#define Pt   pts   /* avoid redeclaration of Pt[] (struct POINT in fd.h) */

#ifdef EGA
#  include <conio.h>   /* kbhit(), getch() */
#  include <graph.h>   /* _setvideomode(), _clearscreen(), etc. */
#  define XMIN         WinDat[0]
#  define XMAX         WinDat[1]
#  define YMIN         WinDat[2]
#  define YMAX         WinDat[3]
#  define x_convert    WinDat[4]
#  define y_convert    WinDat[5]
#endif

/*
 * Physical scales used [notes 1991.2.26, 1992.2.22]:
 *   b (disk exp scale length) = 2.3 kpc = 1.23
 *   Rholm (size of optical disk) = 9.5 kpc = 5.1 (Holmberg radius or "R_max")
 *     [notes 1992.1.18:  now (d293+) Rholm = 7.5 kpc = 4.0]
 *   Rdisk (size of whole disk) = 20 kpc = 10.7
 *   Rcore (core radius of halo) = 19.6 kpc = 10.5 (isothermal potential)
 */
/*      RDISK               10.7 */  /* target disk radius:  20 kpc */
#define RDISK               11.5     /* target disk radius:  20 kpc */
#define MDISK               5.4      /* target disk mass:  see mu0, b */
#define PSEUDO_B            3.57     /* used to scale with Rdisk:  Rdisk/3 */
#define PSEUDO_MU0          .0156    /* ditto:  .00795 * SQR(15./Rdisk) */
#define PSEUDO_MU(r)        (PSEUDO_MU0*exp(-(r)/PSEUDO_B))
/*      R_SHELL0            Rhole    (for example)  */
#define NUM_SHELL1          312      /* target number:  from d294.out */
#define R_SHELL1            3.217    /* target radius:  from d294.out */
#define NUM_SHELL2          (NUM_POINTS-numV)  /* target final number */
#define R_SHELL2            RDISK              /* target final radius */
#define MASS_SHELL(r,s) \
 (TWOPI*bsqr*(PSEUDO_MU(r)*(1.+(r)/PSEUDO_B) - PSEUDO_MU(s)*(1.+(s)/PSEUDO_B)))
#define TMASS_AVG2(r)       (Tmass_avg1*Tmass_avg2*(R_SHELL2-R_SHELL1) \
 / ((Tmass_avg1-Tmass_avg2)*(r)+Tmass_avg2*R_SHELL2-Tmass_avg1*R_SHELL1))
/* p313:  smooth 1/r transition from low avg. mass to high [notes 1992.2.28] */

#if 0  /* save old ones for possible future use */
#  define TMASS_AVG2(r)   \
   ((((r)-R_SHELL1)/(R_SHELL2-R_SHELL1))*(Tmass_avg2-Tmass_avg1) + Tmass_avg1)
/* p311-2:  linear transition from low density to high [notes 1992.2.24-27] */
#endif /* 0 */

static void   solve_n_r   ( int *, double * );





/**********************/
/*  Global Variables  */
/**********************/

char     filename[80];
int      errorlevel=0, Eflag=0;
int      karr[256], *const ring_last=pRing_last;
double   rarr[256], *const ring_last_n=pRing_last_n, *const ring_R=pRing_R;
FILE     *outfile;

static int      phantom=FALSE;
static double   Tmass_avg;   /* average mass per triangle */

const double    pi2_1p5 = PI*PI*1.5;

struct points {
    double x;
    double y;
}                Pt[MAX_V];

int      karr[256] = {
             0,                      /*   d_version (reserved)               */
             0,                      /*   step (reserved)                    */
             0,                      /*   numE (reserved)                    */
             0,                      /*   numT (reserved)                    */
             0,                      /*   numV (construct_points)            */
             0,                      /*   f_version (reserved)               */
             0,                      /*   En (reserved)                      */
             0,                      /*   numRings (construct_points)        */
             0,                      /*   num_phantom (construct_points)     */
             313,                    /*   p_version                          */
             0                       /*   file_type (write_points)           */
         };

double   rarr[256] = {
             0.,                     /*   delta_t (reserved)                 */
             0.,                     /*   Time (reserved)                    */
             0.,                     /*   KinE (reserved)                    */
             0.,                     /*   GravE (reserved)                   */
             0.,                     /*   IntE (reserved [not used])         */
             0.,                     /*   TotE (reserved)                    */
             0., 0., 0.,             /*   LinMom[3] (reserved)               */
             0.,                     /*   TotLinMom (reserved)               */
             0., 0., 0.,             /*   AngMom[3] (reserved)               */
             0.,                     /*   TotAngMom (reserved)               */
   /* UNITS: */
             1870.,                  /*   pc_length                          */
             1.0e10,                 /*   Msol_mass                          */
             3.4,                    /*   Myr_time                           */
             3.09e18,                /*   cm_pc (3.09e16 km/kpc)             */
             1.99e33,                /*   g_Msol                             */
             86400 * 365.2422 * 1e6, /*   s_Myr                              */
             6.67e-8,                /*   G_cgs                              */
             0.,                     /*   G_pMM (reserved)                   */
             0.,                     /*   G_LMT (reserved)                   */
   /* POTENTIAL: */
             0.,                     /*   Rcore (reserved)                   */
             0.,                     /*   rho0 (reserved)                    */
             0.,                     /*   phi0 (reserved)                    */
             0.,                     /*   c1half (reserved)                  */
             0.,                     /*   c2 (reserved)                      */
   /* DISK: */
             RDISK,                  /*   Rdisk                              */
             4.0,                    /*   Rholm                              */
             1.23,                   /*   bB                                 */
             0.,                     /*   hole ratio (not fully supported)   */
             0.,                     /*   Mdisk (reserved)                   */
             0.,                     /*   mu0 (construct_points)             */
             0.,                     /*   v_rms [v_disp] (reserved)          */
             0.,                     /*   kT_m (reserved)                    */
             0.,                     /*   vis_dyn (reserved)                 */
             0.,                     /*   gamma_dyn (reserved)               */
             0.,                     /*   Inclination [degrees] (reserved)   */
             0., 0., 0.,             /*   ring_last_n[3] (reserved)          */
   /* PHANTOM RING: */
             0.,                     /*   r_cm_phantom (reserved)            */
             0.,                     /*   mu_phantom (reserved)              */
             0.,                     /*   mass_phantom (reserved)            */
             0.,                     /*   dvdr_phantom (reserved)            */
             0.                      /*   scaleht2_phantom (reserved)        */
         };

#ifdef EGA
   double   WinDat[6] = {-RDISK*1.33, RDISK*1.33, -RDISK, RDISK, 0., 0.};
#endif             /* = { XMIN, XMAX, YMIN, YMAX, x_convert, y_convert }; */





/******************/
/*  Main program  */
/******************/

main()
{
    construct_points();

#ifndef MSDOS
    write_points();     /* p###pts####:  points file for d###, tplot */
#elif defined(EGA)
    initialize_plot();
    plot_points();
#endif

    printf("\nHave a somewhat pleasant day.\n");
}





/*********************************/
/*  Function construct_points()  */
/*********************************/

void construct_points( void )
{
    int      k, n;
    double   r, phi, offset, half_delta_r, Tmass_avg1, Tmass_avg2;
    double   Rdisk_b=Rdisk/bB, bsqr=SQR(PSEUDO_B);
/*  double   R_hole, R_hole_sqr;   not presently used */


/*--------------------------------------------------------------------------
    Open "standard" output file first, so have a place to print progress.
  --------------------------------------------------------------------------*/

    sprintf(filename, "p%03d.out", p_version);
    if ((outfile = fopen(filename,"a")) == NULL) {
        printf("\nerror:  couldn't open output file [ %s ]\n", filename);
        exit(1);
    }
    printf("Program:  p%03d.c\n", p_version);
    fprintf(outfile, "Program:  p%03d.c\n", p_version);

/*---------------------------------------------------------------------------
    Distribute points on rings subject to criteria that:  (1) triangles have
    nearly equal masses; (2) triangles are, on average, close to equilateral;
    and (3) number of points and size of disk are close to NUM_POINTS and
    Rdisk, respectively.  Mass and number-of-points criteria determine the
    average triangle mass, and given that the centermost ring must have six
    triangles, one can immediately find the radius of the innermost ring.
    [In order that the disk may be easily resized, set mu0 here, on the
    basis of the value of Rdisk:  notes 1990.7.25]  [For looking at warp in
    outer region, less interested in following mass in center:  use pseudo
    load (3 e-foldings) with new mass profile (8-9 e-foldings).  1991.3.23]
    [Boost resolution in outer regions (warp) while leaving inner region un-
    changed.  Distribute via "shells," where Tmass_avg determines density of
    points in local region.  First "shell" is just central point and ring 1;
    Second shell extends to some uninteresting radius and has same density of
    points; and third shell extends to edge of computational disk and has a
    relatively higher density of points.  Notes 1992.2.22-23, 1992.2.27]
  ---------------------------------------------------------------------------*/

    mu0 = MDISK / (TWOPI*SQR(bB)*( 1. - (1.+Rdisk_b)*exp(-Rdisk_b) ));

    Pt[numV].x   = 0.;   /* central point (breaks tesselation degeneracy) */
    Pt[numV++].y = 0.;   /* will adjust numV for hole later in tesselate() */
    ring_last[0] = 0;    /* == (numV - 1) */
    numRings = 0;

    /* "shell" is misnomer:  really an annulus, except for first one */
    Tmass_avg1 = MASS_SHELL(0.,R_SHELL1) / (2.*NUM_SHELL1);
    Tmass_avg2 = MASS_SHELL(R_SHELL1,R_SHELL2) / (2.*NUM_SHELL2);     /* 310 */
    fprintf(outfile, "\n    Tmass_avg1 = %9.6f\nold Tmass_avg2 = %9.6f\n",
      Tmass_avg1, Tmass_avg2);
/*  Tmass_avg2 = Tmass_avg1 * exp(-Tmass_avg1/Tmass_avg2);  */        /* 312 */
    Tmass_avg2 = Tmass_avg1*Tmass_avg2 / (2.*Tmass_avg1-Tmass_avg2);  /* 313 */
    fprintf(outfile, "new Tmass_avg2 = %9.6f\n\n", Tmass_avg2);

    n = 6;   /* ring 1 */
    r = sqrt(sqrt(3)*Tmass_avg1 / PSEUDO_MU0);
    for (k = numV;  numV < k + n;  ++numV) {
        phi = (numV-k) * TWOPI/n  +  .03;
        Pt[numV].x = r * cos(phi);          /* distribute points on circle */
        Pt[numV].y = r * sin(phi);
    }
    ring_last[++numRings] = numV - 1;   /* last index of last ring */

    fprintf(outfile, "ring values:   r[1] = %6.3f  n[1] =   6\n\n", r);
    /* end of first shell */

/*--------------------------------------------------------------------------
    One may now iterate outward according to the N/R relations.  Second 
    shell:  variable (dynamic) triangle mass, extending all the way to the
    boundary of the "third shell" (edge of disk).  [See notes 1990.6.20-22,
    1992.2.23, 1992.2.25].
  --------------------------------------------------------------------------*/

    do {
        double  r_last=r;

        Tmass_avg = (r <= R_SHELL1)? Tmass_avg1 : TMASS_AVG2(r);
        solve_n_r(&n, &r);
        half_delta_r = .5 * (r - r_last);

        offset = fmod((rand()/32767.), TWOPI/n);
        for (k = numV;  numV < k + n;  ++numV) {
            phi = (numV-k) * TWOPI/n  +  offset;
            Pt[numV].x = r * cos(phi);
            Pt[numV].y = r * sin(phi);
        }
        ring_last[++numRings] = numV - 1;   /* last index of last ring */

        fprintf(outfile, "ring values:   r[%d] = %6.3f  n[%d] = %3d\n",
          numRings, r, numRings, n);
#ifdef DIAG
        printf("final values:   r[%d] = %6.3f  n[%d] = %3d\n",
          numRings, r, numRings, n);
#endif /* DIAG */

    } while ((r < R_SHELL2-half_delta_r) && (numV+n < MAX_V));  /* end do */
    /* GRR:  change this to NUM_POINTS criterion; let rescaling take care
     *  of Rdisk */

    fprintf(outfile, "end o' second shell:  cumulative points = %d\n\n", numV);
    /* end of second shell */
    /* end of third shell */

/*---------------------------------------------------------------------------
    Check that numRings is not too large, and rescale positions of points now,
    since outermost ring will not lie at precisely r = Rdisk.  Set number of
    "phantom wedgies" (boundary conditions aid) equal to number of points in
    outermost ring and solve N/R relations one more time to get Rdisk_phantom.
    [Notes 1991.6.24]
  ---------------------------------------------------------------------------*/

    if (numRings >= 50)
        printf("\nwarning:  array overrun likely in rarr[] (numRings = %d)\n",
          numRings);

#   define ratio   half_delta_r     /* available temporary variable */
    ratio = Rdisk / r;

    for (k = 0;  k < numV;  ++k) {
        Pt[k].x *= ratio;
        Pt[k].y *= ratio;
    }
    r = Rdisk;   /* now it does... */

    num_phantom = n;
    phantom = TRUE;

    solve_n_r(&n, &r);

    /* [don't save phantom points; only radius] */
    rarr[255] = r;  /* Rdisk_phantom temporary storage (until read by d###) */

    fprintf(outfile, "ring values:   r[%d] = %6.3f  n[%d] = %3d   [phantom]\n",
      numRings+1, r, numRings+1, n);
#ifdef DIAG
    printf("final values:   r[%d] = %6.3f  n[%d] = %3d   [phantom]\n",
      numRings+1, r, numRings+1, n);
#endif /* DIAG */

/*---------------------------------------------------------------------------
    Print count of points and exit.
  ---------------------------------------------------------------------------*/

    fprintf(outfile, "\nConstruction of points complete, you betcha.  %d "
      "points created%s", numV, (hole>0.)? ",\n including 1 temporary central "
      "point to be removed in tesselate().\n" : ".\n");

} /* end function construct_points() */





/**************************/
/*  Function solve_n_r()  */
/**************************/

static void solve_n_r(int *pn, double *pr)
{
    int      cont, n, n_last;
    double   m, nguess, nguess_best, ntest, min_diff=PRETTY_BIG;
    double   bconst, r, r_last, r_best;


/*--------------------------------------------------------------------------
    Solve for the radius (r) and number of points (n) of the current ring,
    given the values of the previous ring (r_last and n_last).  The equa-
    tions relating n and r ("N/R relations") are, in LaTeX format:
    \begin{eqnarray*}
      \R 
       &=& \r \left[ 1 +  {{3\pi^2/2}\over{\N \n}} 
          \left(1 + \sqrt{1+ {{2\N \n}\over{3\pi^2/2}}}\right) \right] \\
      \N 
       &=& -\n + {{\pi\left(\R - \r\right) \left(
          \r \tilde{\mu}\left(\smallfrac{2\r}{3} + \smallfrac{\R}{3}\right) +
          \R \tilde{\mu}\left(\smallfrac{\r}{3} + \smallfrac{2\R}{3}\right)
          \right)} \over {\left<m_\Delta\right>}} ,
   \end{eqnarray*}
   [See notes 1990.6.20-22].
  --------------------------------------------------------------------------*/

    n_last = *pn;
    r_last = *pr;

    for (m=(int)(.8*n_last), cont=TRUE; (m < 3*n_last) && cont;  ++m) {
        nguess = m;
        bconst = pi2_1p5 / (nguess*n_last);
        r = r_last * (1. + bconst*(1. + sqrt(1.+2./bconst)));
        ntest = -n_last + PI*(r-r_last)*( r_last*PSEUDO_MU(2.*r_last/3. + r/3.)
          + r*PSEUDO_MU(r_last/3. + 2.*r/3.) ) / Tmass_avg;

#ifdef DIAG
        if (phantom) {
            fprintf(outfile, "ring %2d:  r = %6.3f  n(guess) = %6.2f  n(test) "
              "= %6.2f   [phantom]\n", numRings+1, r, nguess, ntest);
            printf("ring %2d:  r = %6.3f  n(guess) = %6.2f  n(test) "
              "= %6.2f   [phantom]\n", numRings+1, r, nguess, ntest);
        } else {
            fprintf(outfile, "ring %2d:  r = %6.3f  n(guess) = %6.2f  n(test) "
              "= %6.2f\n", numRings, r, nguess, ntest);
            printf("ring %2d:  r = %6.3f  n(guess) = %6.2f  n(test) "
              "= %6.2f\n", numRings, r, nguess, ntest);
        }
#endif /* DIAG */

        if (fabs(nguess-ntest) < min_diff) {
            min_diff = fabs(nguess-ntest);
            nguess_best = nguess;
            r_best = r;
        } else
            cont = FALSE;
    }

    /* return best solution */
    *pn = (int) (nguess_best + .5);
    *pr = r_best;
}





#ifndef MSDOS

/***************************/
/* Function write_points() */
/***************************/

void write_points( void )
{
    int      i, total_bytes, numPt=numV, num_fake=0;  /* both used in macro */
    double   *coord, *p;
    FILE     *tessfile;


/*--------------------------------------------------------------------------
    Allocate buffer in which to copy coordinates, open output file, write
    karr[] and rarr[], copy coordinates, write buffer [all in macro], and
    close output file and free buffer.
  --------------------------------------------------------------------------*/

    sprintf(filename, "p%03dpts%04d", p_version, numV);  /* includes hole pt */
    file_type = POINTS;

    MALLOC_OPEN_WRITEPTS(2*numV)

    fclose(tessfile);
    free(coord);
    printf("Total bytes written to [ %s ]:  %ld\n", filename, total_bytes);
    fprintf(outfile, "Total bytes written to [ %s ]:  %ld\n", filename,
      total_bytes);

} /* end function write_points() */

#endif /* !MSDOS */





#if defined(EGA)
#  include "plotpts.c"   /* contains initialize_plot(), plot_points() */
#endif
