/*--------------------------------------------------------------------------

  render.c

  This program reads in data from a Cray "surface" file (d###f###s###.#)
  and renders the disk surface using DISSPLA.  The output is a PostScript
  file (s##########.ps or w##########.ps, where the #'s correspond to
  those in the input file's name).

  --------------------------------------------------------------------------

  To compile:    scc -c render.c
                 segldr render.o -l$disspla -o render

  --------------------------------------------------------------------------

     Version:  Y-MP/DISSPLA/PostScript               Greg Roelofs
     Source:   render.c on eagle                     19 December 1991

  --------------------------------------------------------------------------

     v 1.0:   original program (based on radii v1.1 and disksurf.f v2.0)
     v 1.1:   modified to use new header file; minor corrections to comments

  --------------------------------------------------------------------------*/



/*****************************/
/*  Includes, Globals, etc.  */
/*****************************/

#include "fd.h"          /* all sorts of great stuff */
#include "disspla.h"     /* DISSPLA prototypes, fortran.h */
#include <string.h>      /* strlen() */

int      axes=FALSE, shaded=FALSE, remove_hidden=FALSE;
int      karr[256], tri[MAX_T][3];
double   rarr[256], x=30., y=30., z=8.;
double   Vx[MAX_V], Vy[MAX_V], Vz[MAX_V];
FILE     *infile;

static void   usage( void );





/******************/
/*  Main program  */
/******************/

main( int argc, char *argv[] )
  {
    char         filename[80];
    char         c, c1, *pf=filename;
    int          n, error=0, sflag=FALSE, wflag=FALSE;
    int          total_bytes;
    double       t1, dt;

    char            title[80];
    int             iarr[16]={0.,0.,0.}, id, ipos;
    static int      i0=0, i1=1, i2=2, i3=3, i4=4;
    static int      i5=5, i6=6, i7=7, i8=8, i9=9;
    static int      im3=-3, im102=-102, im103=-103, im204=-104, im111=-111;
    static int      i100=100, i500000=500000;
    static double   d0=0., d1=1., d4=4., d8=8., d9=9., d10=10., d11=11.;
    static double   d15=15., d30=30., d45=45., d90=90., d100=100.;
    static double   dp001=.001, dp006=.006, dp2=.2, dp25=.25, dp7=.7;
    static double   d1p75=1.75, d5p5=5.5, d8p5=8.5, d9p625=9.625, d9p875=9.875;
    static double   axmin=-12., axinc=4., axmax=12.;



/*---------------------------------------------------------------------------
    Parse command line for '-' options.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    if ( argc <= 1 )
      --argc;
    else
      while ( --argc > 0  &&  (*++argv)[0] == '-')
        while ( c = *++(argv[0]) )   /* argv was inc => argv[0] not prognam */
          switch (c) {
            case 'a':      /* draw (ugly) Cray axes */
                axes = TRUE;
                break;

            case 's':      /* do shaded plot */
                sflag = TRUE;
		break;

            case 'w':      /* do wireframe plot */
                wflag = TRUE;
                c1 = *(argv[0] + 1);   /* get char after 'w' w/o changing ptr */
                if ((c1 >= '0') && (c1 <= '1')) {   /* allow more in future? */
                    n = atoi( argv[0]+1 );        /* pointer to digit string */
                    *(argv[0]+1) = '\0';          /* overwrite digit */
                    if (n > 0)
                        remove_hidden = TRUE;
                    if (n > 1)
                        ++error;
                }
                break;

            default:
                printf("render.c:  invalid option [ -%c ]\n", c);
                ++error;
                *(argv[0]+1) = '\0';   /* terminate this argument; try next */
                break;
          }  /* end switch */

/*---------------------------------------------------------------------------
    If any invalid options were encountered, print syntax synopsis.  Any 
    argument remaining (i.e., not preceded by '-') must be the name of the 
    data file to be opened, so do that now.  Otherwise, if no arguments 
    left, print synopsis and exit.
  ---------------------------------------------------------------------------*/

    if (argc > 0) {
        while (*pf++ = *argv[0]++);   /* copy argument into filename[] */
        if (error)
            usage();   /* bad switch, but have filename so can continue */
    } else {
        usage();
        exit(1);       /* no filename:  ho ho, gotta go */
    }

    printf("\n\nOpening data file [ %s ] for input.\n\n", filename);
    if ((infile = fopen(filename,"rb")) == NULL) {
        printf("error:  couldn't open file [ %s ]\n", filename);
        exit(1);
    }

/*---------------------------------------------------------------------------
    Compensate for user screwiness and print quick summary of what we are 
    about to do.
  ---------------------------------------------------------------------------*/

    if (sflag || !wflag) {    /* both, neither, or just sflag */
        shaded = TRUE;
        remove_hidden = FALSE;
    }

    if (shaded)
        printf("  shaded disk plot\n");
    else if (remove_hidden)
        printf("  wireframe disk plot, hidden lines removed (be patient)\n");
    else
        printf("  wireframe disk plot, hidden lines NOT removed\n");

/*---------------------------------------------------------------------------
    Read karr[] and rarr[] first, then vertices and triangle "pointers."
  ---------------------------------------------------------------------------*/

    total_bytes  = fread( karr, sizeof(int), 256, infile ) *
                                sizeof(int);
    total_bytes += fread( rarr, sizeof(double), 256, infile ) *
                                sizeof(double);
    if (numV > MAX_V) {
        printf("\nerror:  too many vertices (%d) for arrays (%d)\n",
          numV, MAX_V);
        exit(4);
    }
    if (numT > MAX_T) {
        printf("\nerror:  too many triangles (%d) for arrays (%d)\n",
          numT, MAX_T);
        exit(4);
    }

    total_bytes += fread( Vx, sizeof(double), numV, infile ) *
                              sizeof(double);
    total_bytes += fread( Vy, sizeof(double), numV, infile ) *
                              sizeof(double);
    total_bytes += fread( Vz, sizeof(double), numV, infile ) *
                              sizeof(double);

    total_bytes += fread( tri, sizeof(int), 3*numT, infile ) *
                               sizeof(int);

    fclose(infile);
    printf("\n  total bytes read from file:  %d\n", total_bytes);
    printf("  number of vertices:  %5d\n", numV);
    printf("  number of triangles: %5d\n", numT);
    printf("  step number:         %5d\n", iteration);
    printf("  time:             %8.3f\n\n", Time);

/*---------------------------------------------------------------------------
    Open output file for plot:  use IOMGR to intercept PostScript on way to
    printer, and save in file instead.  This version appends to the end of 
    an existing file of the same name rather than overwriting it (either way,
    the rename() call at the end keeps the files straight).  Also signal use
    of hardware fonts and shading.
  ---------------------------------------------------------------------------*/

    IOMGR( &i5, &im102);
/*  IOMGR( &i2, &im104);         overwrite old file */
    IOMGR( "render.ps", &im103); /* this string works...why? */
    IOMGR( &iarr, &im111);       /* ??? */
    PSCRPT( &d0, &d0, &dp006);   /* default page height/width, .006" lines */
    /* use hardware (PS) characters for all alphabets through Hebrew */
/*  HWCHAR( &d100, &d100, &d90, &i1, _cptofcd("HEBREW",6));  */
    HWCHAR( &d100, &d100, &d90, &i1, "HEBREW");   /* this works:  why?? */
    HWSHD();

/*---------------------------------------------------------------------------
    Initialize DISSPLA workspace, plotting region, etc.  The BCOMON call is
    necessary, but apparently we don't need a "COMMON" set aside (on either
    the VAX or Y-MP).  Amount of memory requested is a wild guess.
  ---------------------------------------------------------------------------*/

    BCOMON( &i500000);
    SYSBUF();   /* dynamically allocate memory */
/*  RESET( "ALL" );   */
    PAGE( &d8p5, &d11);
/*  NOBRDR();  */
    AREA2D( &d8, &d10);
    VOLM3D( &d30, &d30, &d30);   /* from -15 to +15, all three axes */
    VIEW( &x, &y, &z);           /* view from user-defined position */
    if (axes) {
        INTAXS();
        X3NAME( _cptofcd("X",1), &i1);
        Y3NAME( _cptofcd("Y",1), &i1);
        Z3NAME( _cptofcd("Z",1), &i1);
    }
    GRAF3D( &axmin,&axinc,&axmax, &axmin,&axinc,&axmax, &axmin,&axinc,&axmax);

    /* all three of the methods below for passing character strings to DISSPLA
     *  work (for now); under version 11, however, only the last is likely to */
/*  HEADIN( "HYDRO", &five, &htmult1, &nlines );
    nchars = strlen(pTitle2);
    HEADIN( pTitle2, &nchars, &htmult2, &nlines );
    nchars = strlen(pTitle3);
    HEADIN( _cptofcd(pTitle3,nchars), &nchars, &htmult2, &nlines );
 */

/*---------------------------------------------------------------------------
    Set up and render the scene.
  ---------------------------------------------------------------------------*/

    INISCN();   /* initialize the scene */
    /* set object visibility:  FRONT, BACK, or BOTH */
    OBJVIS( _cptofcd("BOTH",4), _cptofcd("NOEDGE",6));
/*  NEWCLR( _cptofcd("CYAN",4));  */
/*  HAND( _cptofcd("LEFT",4));  *//* signal left-handed polygon orientations */

    if (shaded) {   /* if not wireframe, add some light sources */
        NEWOBJ( &id);                   /* add one to the right... */
        LIGHT( &id, &d10, &d45, &d8);   /* (0,45,5) is OK; (10,45,8) better */
        AMBHSI( &d4, &d1, &dp25);       /* make ambient light dim */
        SETHLT( &dp7, &d15);   /* make surface somewhat shiny (.9 = silver) */
    }

    NEWOBJ( &id);   /* begin disk object */
    /* store vertex positions in internal vertex table (IPOS gets returned) */
    VERTEX( Vx, Vy, Vz, &numV, &ipos);
    /* store faces ("pointers" to vertices) in internal face table */
    FACE( &tri[0][0], &im3, &numT, &ipos);   /* (can use ipos or i1) */

    if (remove_hidden)
        HIDE();

    /* draw the scene */
    if (shaded)
        DRWSCN( _cptofcd("SHADE",5));
    else
        DRWSCN( _cptofcd("FRAME",5));

/*--------------------------------------------------------------------------
    Finish up:  print title line, close plot, close DISSPLA.
  --------------------------------------------------------------------------*/

    SHDCHR( &d90, &i1, &dp001, &i1);
    NEWCLR( _cptofcd("FORE",4));

    SCMPLX();   /* Roman font */
    sprintf( title, "Versions:  d%03d, f%03d", d_version, f_version);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d5p5, &d9p875);
    sprintf( title, "Time:  %.2f units", Time);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d5p5, &d9p625);

    SWISSM();   /* Helvetica font */
    HEIGHT( &dp2);
    sprintf( title, "Evolved Galactic Warp, %d Myr", (int)(Time*Myr_time+.5) );
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d1p75, &dp7);

    ENDPL( &i0);
    DONEPL();

/*--------------------------------------------------------------------------
    Rename file according to program version numbers and current time.
    [Renaming prevents appending to an existing file; if exists, get error
    message *after* finished, but render.ps still exists.]
  --------------------------------------------------------------------------*/

    sprintf(filename, "%c%03d%03d%04.0f.ps", (shaded? 's':'w'), d_version,
      f_version, 10.*Time);
    printf("\n  renaming output file to [ %s ]\n", filename);
    if (rename("render.ps",filename)) {
        printf("error:  could not rename [ render.ps ] to [ %s ]\n",
          filename);
    }

/*---------------------------------------------------------------------------
    All done:  calculate CPU time and exit.
  ---------------------------------------------------------------------------*/

    printf("\nProgram completed, possibly successfully.  Have an "
      "excruciatingly nice day.\n");

    dt = SECOND() - t1;
    printf("  program CPU time:  %11.4e\n", dt);

  } /* end main() */





/**********************/
/*  Function usage()  */
/**********************/

static void usage( void )
  {
    printf("\tUsage:  render [-a] [-s | -w[#]] file\n"
      "\t  -a      include axes in plot (highly unappetizing)\n"
      "\t  -s      do shaded surface (default)\n"
      "\t  -w[0]   do wireframe surface (hidden lines not removed)\n"
      "\t  -w1     do wireframe surface (hidden lines removed:  VERY slow)\n"
      "\t  file    data file from which to be read\n");
  }
