/*---------------------------------------------------------------------------

            RRRRRRR   EEEEEEEE  PPPPPP      AA     IIIIII  RRRRRRR   
            RR    RR  EE        PP    PP   AAAA      II    RR    RR  
            RR    RR  EEEEEEE   PP    PP  AA  AA     II    RR    RR  
            RRRRRR    EE        PPPPPP   AA    AA    II    RRRRRR    
            RR   RR   EE        PP       AAAAAAAA    II    RR   RR   
            RR    RR  EEEEEEEE  PP       AA    AA  IIIIII  RR    RR  

  ---------------------------------------------------------------------------

  repair.c

  This file contains two of the three grid-repair routines for the evolver
  (hydro.c):  repair_grid() and swap_points().  swap_edges() is also used by
  the evolver, so it is included within shared.c.  This file compiles into a
  stand-alone object file.  [Notes 1990.12.31]

  ---------------------------------------------------------------------------

     Version:  Y-MP 3.00                             Greg Roelofs
     Source:   repair.c on Eagle                     14 February 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:  scc -c [-DUDEBUG] repair.c
     on PC:    cl -c -AS -G2 -FPi87 [-DUDEBUG] repair.c

     UDEBUG:   include diagnostics in log file and stderr

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     [vaxwr.f:      routine for conversion of edge data to VMS format]

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotload.c:    contains MS-DOS graphics routines

     points.c:      stand-alone program to create and plot points only

     f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     grav.f:        gravity code (used only by update###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
  -> repair.c:      grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

  Version history:

     h280.c:  grid-repair code first split out into separately compiled
               source file; repair routines significantly modified
     h285.c:  swap_edges() moved into sh285.c (now shared by loader)
     f290.c:  evolver filename changed to floppy.c; header filename changed 
               to fd.h
     f300.c:  warnings to dbgfile as well as outfile; modified conditions
               to skip viscosity

  ---------------------------------------------------------------------------*/



#include "fd.h"

double            norm;
struct edge       *pCenterE;                       /* local globals */
struct triangle   *pFirstFoldedT, *pSecondFoldedT;





/****************************/
/*  Function repair_grid()  */
/****************************/

void repair_grid( void )
{
    char          error_msg[256];
    int           i, j, j1, k, m, mask, fold_config;
    double        min_len[2], t1, costhet, angle;
    struct edge   *min_pE[2], *pE;



/*---------------------------------------------------------------------------
    First loop through triangles and calculate new edge-normals.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    for ( k = 0  ;  k < numT  ;  k++ ) {
        for ( j = 0  ;  j < 3  ;  j++ )        /* three edges/vertices  */
            for ( i = 0  ;  i < 3  ;  i++ )    /* cartesian coordinates */
                s[j][i] = T[k].pV[j]->r[i] - T[k].pV[(j+1)%3]->r[i];

        for ( j = 0  ;  j < 3  ;  j++ ) {      /* three vertices again  */
            j1 = (j+1)%3;

            /* loop over cartesian indices:  calc normal to opposite edge */
            for ( i = 0  ;  i < 3  ;  i++ ) {
                T[k].n[j][i] = s[j][i] - DOT(s[j],s[j1]) * s[j1][i] /
                                            SQR(T[k].pE[j]->len);
/*
 * This if-stmt will be unnecessary when swap_edges preserves normals:
 *              if ( T[k].pE[j]->pT[0] == &T[k] )
 *                  T[k].pE[j]->n[0] = T[k].n[j];
 *              else
 *                  T[k].pE[j]->n[1] = T[k].n[j];
 */
            }

            /* loop over cartesian indices again and make edge normals unit */
            norm = NORM(T[k].n[j]);
            for ( i = 0  ;  i < 3  ;  i++ )
                T[k].n[j][i] /= norm;

        } /* end j-loop (vertices) */
    } /* end k-loop (triangles) */

/*---------------------------------------------------------------------------
    Now loop through edges, starting with calculation of folding angles.
  ---------------------------------------------------------------------------*/

    for ( k = 0  ;  k < numE  ;  k++ )
      /* exclude edges on hull, ones already checked */
      if ( !(E[k].status & (E_ON_HULL | E_CHECKED_FOLD))  &&
           (costhet = DOT(E[k].n[0],E[k].n[1])) > -0.5 )  {

          angle = (costhet < 1.)?  acos(costhet) / PI180  :  0.;

#if FALSE /* needs work */
  fprintf(dbgfile, "    E[%d]:  |n[0]| = %6.3f, |n[1]| = %6.3f, "
    "costhet = %17.14f radians\n", k, NORM(E[k].n[0]),
    NORM(E[k].n[1]), costhet);
  x = (E[k].pV[0]->r[0] + E[k].pV[1]->r[0]) / 2.;
  y = (E[k].pV[0]->r[1] + E[k].pV[1]->r[1]) / 2.;
  z = (E[k].pV[0]->r[2] + E[k].pV[1]->r[2]) / 2.;
  fprintf(dbgfile, "    WARNING (t=%5.1f):  large bend (%.1f deg): "
    "E[%d], r = (%6.3f,%6.3f,%6.3f)\n", Time, angle, k, x, y, z);
  fprintf(dbgfile, "    (skipping viscous-force calculation)\n");
  for ( i = 0  ;  i < 3  ;  ++i )
    fprintf(dbgfile, "    force[%d] on pV[0]: %13.4e    on pV[1]:"
      " %13.4e\n", i, E[k].pV[0]->F[i], E[k].pV[1]->F[i] );
#endif

          fprintf(outfile, "    WARNING:  E[%d] large bend (%.1f deg)"
            " -- ", k, angle);
          fprintf(dbgfile, "    WARNING:  E[%d] large bend (%.1f deg)"
            " -- ", k, angle);

      /*---------------------------------------------------------------------
          If fold has occurred (acute angle), check for fix:  first figure
          out which edge is central one, starting from current one.  Look
          for shortest edge of two neighbor triangles, and print angle info
          on all edges of triangles which neighbor that edge, if debugging.
        ---------------------------------------------------------------------*/

          if (costhet > 0.) {

              for ( m = 0  ;  m < 2  ;  ++m ) {
                  min_pE[m] = E[k].pT[m]->pE[0];
                  min_len[m] = min_pE[m]->len;
                  for ( j = 1  ;  j < 3  ;  ++j ) {
                      pE = E[k].pT[m]->pE[j];
                      if ( pE->len < min_len[m] ) {
                          min_pE[m] = pE;
                          min_len[m] = pE->len;
                      }
                  }
              }

              pCenterE = min_pE[ ((min_len[0] < min_len[1])? 0 : 1) ];
              pFirstFoldedT = pCenterE->pT[0];
              pSecondFoldedT = pCenterE->pT[1];

#ifdef UDEBUG
              fprintf(dbgfile, "\n    folded edge:  E[%d]  (%.2f)\n", k, angle);
              fprintf(dbgfile,
                "    shortest sides of two triangles adjacent to fold:\n");
              for (m = 0;  m < 2;  ++m)
                  fprintf(dbgfile, "     E[%d]:  %7.5f\n", min_pE[m]-E,
                    min_len[m] );
              fprintf(dbgfile, "    pCenterE = E[%d]   x = %5.2f, y = %5.2f,"
                " z = %5.2f\n", pCenterE-E,
                .5*(pCenterE->pV[0]->r[0] + pCenterE->pV[1]->r[0]),
                .5*(pCenterE->pV[0]->r[1] + pCenterE->pV[1]->r[1]),
                .5*(pCenterE->pV[0]->r[2] + pCenterE->pV[1]->r[2]) );
              fprintf(dbgfile, "    pCenterE -> V[%d], V[%d]; T[%d], T[%d]\n",
                pCenterE->pV[0]-V, pCenterE->pV[1]-V,
                pCenterE->pT[0]-T, pCenterE->pT[1]-T);

              for ( j = 0  ;  j < 3  ;  ++j ) {
                  pE = pFirstFoldedT->pE[j];
                  costhet = DOT( pE->n[0], pE->n[1] );
                  fprintf(dbgfile, "    E[%d]:  fold angle = %9.5f deg\n",
                    pE-E, ((costhet<1.)? acos(costhet)/PI180 : 0.) );
              }
              for ( j = 0  ;  j < 3  ;  ++j )
                  if ( (pE = pSecondFoldedT->pE[j]) != pCenterE ) {
                      costhet = DOT( pE->n[0], pE->n[1] );
                      fprintf(dbgfile, "    E[%d]:  fold angle = %9.5f deg\n",
                        pE-E, ((costhet<1.)? acos(costhet)/PI180 : 0.) );
                  }
#endif  /* UDEBUG */

          /*-----------------------------------------------------------------
              Scope out configuration of fold:  if both skinny triangles are
              folded, fix by swapping points; otherwise a non-convex edge-
              swap should take care of things.  Bit 0 of fold_config is the
              (short) CenterE; bits 1 and 2 are the remaining edges of First-
              FoldedT; and bits 3 and 4 correspond to SecondFoldedT's edges.
              [Any way to pinpoint edge to be swapped?  Have E_FORCE_SWAP...
              See notes 1990.1.3]
            -----------------------------------------------------------------*/

              for (mask = 2, fold_config = 0, j = 0;  j < 3;  ++j) {
                  if ((pE = pFirstFoldedT->pE[j]) == pCenterE) {
                      if (DOT(pE->n[0],pE->n[1]) > 0.)    /* costhet */
                          fold_config |= 1;
                  } else {
                      if (DOT(pE->n[0],pE->n[1]) > 0.)
                          fold_config |= mask;
                      mask <<= 1;
                  }
                  pE->status |= E_CHECKED_FOLD;
              }
              for (j = 0;  j < 3;  ++j) {
                  if ((pE = pSecondFoldedT->pE[j]) != pCenterE) {
                      if (DOT(pE->n[0],pE->n[1]) > 0.)    /* costhet */
                          fold_config |= mask;
                      mask <<= 1;
                      pE->status |= E_CHECKED_FOLD;  /* already did pCenterE */
                  }
              }

              switch (fold_config) {
                  case 7:    /* bits 0-2:  only one tri. folded:  edge-swap */
                  case 25:   /* bits 0, 3-4 */
                      fprintf(outfile, "saving for swap_edges()\n");
                      fprintf(dbgfile, "saving for swap_edges()\n");
             /*       ???->status |= E_FORCE_SWAP;   */
                      break;
                  case 30:
                      fprintf(outfile, "swapping points on edge E[%d]\n",
                        pCenterE-E);
                      fprintf(dbgfile, "swapping points on edge E[%d]\n",
                        pCenterE-E);
                      swap_points();   /* both folded:  point-swap */
                      break;
                  default:
                      /* skip viscosity on original (acute-bend) edge; print
                       * error message about "center" edge */
                      sprintf(error_msg, "    repair_grid ERROR:  unknown con"
                        "figuration (%d):  E[%d]\n", fold_config, pCenterE-E);
                      fprintf(outfile, "skipping visc (rings %d, %d);\n%s",
                        E[k].pV[0]->attrib, E[k].pV[1]->attrib, error_msg);
                      fprintf(dbgfile, "skipping visc (rings %d, %d);\n%s",
                        E[k].pV[0]->attrib, E[k].pV[1]->attrib, error_msg);
                      fprintf(stderr, error_msg);
                      E[k].status |= E_SKIP_VISC;
                      break;
              }

          } else {  /* costhet <= 0:  E[k] forms large bend, but not acute */

              /* don't skip viscosity unless acute bend (just warn)... */
              fprintf(outfile, "won't skip visc (rings %d, %d)\n",
                E[k].pV[0]->attrib, E[k].pV[1]->attrib);
              fprintf(dbgfile, "won't skip visc (rings %d, %d)\n",
                E[k].pV[0]->attrib, E[k].pV[1]->attrib);
          }

      } /* end if (not on convex hull; not checked; folded), k-loop (edges) */

/*---------------------------------------------------------------------------
    Serious grid-repair finished; now do edge-swaps to take care of normal
    distortion of grid.  [Change this to a single call with arg = num_passes]
  ---------------------------------------------------------------------------*/

      swap_edges(0);
      swap_edges(1);
      swap_edges(2);

      repair_time += SECOND() - t1;

}  /* end function repair_grid() */





/****************************/
/*  Function swap_points()  */
/****************************/

void swap_points( void )
  {
    int               i, j, j1, j_save, j_cur, j_next, m;
    double            t1, dt;
    struct edge       *pCurE;
    struct triangle   *pCurT, *pNextFoldedT;
    struct vertex     *pCurV, *pAltV;
#ifdef UDEBUG
    struct edge       *pE[9]={&E[77],&E[230],&E[232],&E[57],&E[126],
                              &E[125],&E[128],&E[132],&E[56]};
#endif



/*---------------------------------------------------------------------------
    Initialize quantities prior to beginning loop.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    pCurV = pCenterE->pV[0];
    pAltV = pCenterE->pV[1];
    pCurT = pNextFoldedT = pFirstFoldedT;

    /* loop through verts of folded tri to set up for first "current" edge */
    for ( j = 0  ;  j < 3  ;  ++j )
        if ( pCurT->pV[j] == pCurV )
            break;

#if FALSE
    if (j>2) {
        fprintf(dbgfile, "    swap_points ERROR:  folded triangle T[%d] "
          "doesn't point to central vertex V[%d]\n", pCurT-T, pCurV-V);
        fprintf(stderr, "    swap_points ERROR:  folded triangle T[%d] "
          "doesn't point to central vertex V[%d]\n", pCurT-T, pCurV-V);
        exit(5);
    }
#endif

    j_cur = j;
    j1 = (j+1) % 3;
    j_save = (pCurT->pV[j1] == pAltV) ?  j1  :  (j+2) % 3;

/*---------------------------------------------------------------------------
    Begin main loop through triangles/edges surrounding two central vertices.
    Two folded triangles are special cases, so check for them first.
  ---------------------------------------------------------------------------*/

    do {

        if (pCurT == pNextFoldedT) {
#ifdef UDEBUG
            fprintf(dbgfile, "    swap_points:  doing folded triangle T[%d]:"
              "  E[%d], E[%d], E[%d]\n", pCurT-T, pCurT->pE[0]-E,
              pCurT->pE[1]-E, pCurT->pE[2]-E);
/* fprintf(dbgfile, "      mass = %11.3e\n", pCurT->mass);  */
#endif  /* UDEBUG */

            pNextFoldedT = pSecondFoldedT;    /* (doesn't matter 2nd time) */

            /* pCurE was either not set or set incorrectly, so fix now */
            pCurE = pCurT->pE[j_cur];   /* j_cur:  set previously */

            /* fix up current edge & triangle pointers to central vertices */
            j = (pCurE->pV[0] == pAltV) ?  0  :  1;
            pCurE->pV[j]      = pCurV;
            pCurT->pV[j_cur]  = pAltV;
            pCurT->pV[j_save] = pCurV;

            /* other half of the loop:  exchange working labels of two verts */
            pCurV = pAltV;
            pAltV = pCurE->pV[j];      /* old pCurV */

            /* recalculate length of modified edge */
            pCurE->len = DIST(pCurE->pV[0]->r, pCurE->pV[1]->r);

        /*-------------------------------------------------------------------
            Set up pointers for next triangle (can't reset folded-triangle
            edge normals yet because pFirstFoldedT still has a bad edge).
          -------------------------------------------------------------------*/

            pCurT = (pCurE->pT[0] == pCurT) ? pCurE->pT[1]  :  pCurE->pT[0];

            for ( j = 0  ;  j < 3  ;  ++j )
              if ( pCurT->pE[j] == pCurE )
                break;

#if FALSE
            if (j>2) {
                fprintf(dbgfile, "    swap_points ERROR:  triangle T[%d]
                  "doesn't point to edge E[%d]\n", pCurT-T, pCurE-E);
                fprintf(stderr, "    swap_points ERROR:  triangle T[%d] "
                  "doesn't point to edge E[%d]\n", pCurT-T, pCurE-E);
                exit(5);
            }
#endif

            j1 = (j+1) % 3;
            if ( pCurT->pV[j1] == pCurV ) {
                j_cur = j1;                     /* for next triangle */
                pCurE = pCurT->pE[ (j+2)%3 ];
            } else {
                j_cur = (j+2) % 3;              /* next triangle */
                pCurE = pCurT->pE[j1];
            }

        } /* end if (folded) */

    /*-----------------------------------------------------------------------
        Do the next regular-type triangle.
      -----------------------------------------------------------------------*/

#ifdef UDEBUG
        fprintf(dbgfile, "    swap_points:  doing regular triangle T[%d]:"
          "  E[%d], E[%d], E[%d]\n", pCurT-T, pCurT->pE[0]-E, pCurT->pE[1]-E,
          pCurT->pE[2]-E);
/* fprintf(dbgfile, "      mass = %11.3e\n", pCurT->mass);  */
#endif  /* UDEBUG */

        j = (pCurE->pV[0] == pCurV) ?  0  :  1;
        pCurE->pV[j]     = pAltV;
        pCurT->pV[j_cur] = pAltV;    /* j_cur:  from last time */

    /*-----------------------------------------------------------------------
        Recalculate edge length and normals for current (regular) triangle.
      -----------------------------------------------------------------------*/

        pCurE->len = DIST(pCurE->pV[0]->r, pCurE->pV[1]->r);

        for ( j = 0  ;  j < 3  ;  j++ )        /* loop over edges/vertices   */
            for ( i = 0  ;  i < 3  ;  i++ )    /* loop over cartesian coords */
                s[j][i] = pCurT->pV[j]->r[i] - pCurT->pV[(j+1)%3]->r[i];

        for ( j = 0  ;  j < 3  ;  j++ ) {      /* three vertices again  */
            j1 = (j+1)%3;
            /* loop over cartesian indices:  calc normal to opposite edge */
            for ( i = 0  ;  i < 3  ;  i++ )
                pCurT->n[j][i] = s[j][i] - DOT(s[j],s[j1]) * s[j1][i] /
                                             SQR(pCurT->pE[j]->len);
            /* loop over cartesian indices again and make edge normals unit */
            norm = NORM(pCurT->n[j]);
            for ( i = 0  ;  i < 3  ;  i++ )
                pCurT->n[j][i] /= norm;
        } /* end j-loop (vertices) */

#if FALSE      /* double-check that edge-normals really are normal to edges */
    if (pCurT == &T[23]) {
        fprintf(dbgfile, "    ^^^ T[23]:\n");
        fprintf(dbgfile, "        len(s[0]) = %11.3e;  E[230].len = %11.3e;  "
          "pCurT->pE[2]->len = %11.3e\n", NORM(s[0]), E[230].len,
          pCurT->pE[2]->len );
        fprintf(dbgfile, "        len(s[1]) = %11.3e;   E[77].len = %11.3e;  "
          "pCurT->pE[0]->len = %11.3e\n", NORM(s[1]), E[77].len,
          pCurT->pE[0]->len );
        fprintf(dbgfile, "        len(s[2]) = %11.3e;  E[154].len = %11.3e;  "
          "pCurT->pE[1]->len = %11.3e\n", NORM(s[2]), E[154].len,
          pCurT->pE[1]->len );
        fprintf(dbgfile, "        s[0] * n[2] = %11.3e\n",
          DOT(s[0],pCurT->n[2]) );
        fprintf(dbgfile, "        s[1] * n[0] = %11.3e\n",
          DOT(s[1],pCurT->n[0]) );
        fprintf(dbgfile, "        s[2] * n[1] = %11.3e\n",
          DOT(s[2],pCurT->n[1]) );
    }

    for ( j = 0  ;  j < 3  ;  ++j ) {     /* 3 triangle edges */
        for ( i = 0  ;  i < 3  ;  ++i )
            s[0][i] = pCurT->pE[j]->pV[0]->r[i] - pCurT->pE[j]->pV[1]->r[i];
        fprintf(dbgfile, "    <#> T[%d] -> E[%d];  normal = %11.3e;  vtx: %d "
          "%d %d\n", pCurT-T, pCurT->pE[j]-E, DOT(pCurT->n[j],s[0]),
          pCurT->pV[0]-V, pCurT->pV[1]-V, pCurT->pV[2]-V );
    }
#endif

    /*-----------------------------------------------------------------------
        Set up pointers for next triangle.
      -----------------------------------------------------------------------*/

        pCurT = (pCurE->pT[0] == pCurT) ? pCurE->pT[1]  :  pCurE->pT[0];

        for ( j = 0  ;  j < 3  ;  ++j )
          if ( pCurT->pE[j] == pCurE )   /* old pCurE */
            break;

#if FALSE
        if (j>2) {
            fprintf(dbgfile, "    swap_points ERROR:  triangle T[%d] doesn't "
              "point to edge E[%d]\n", pCurT-T, pCurE-E);
            fprintf(stderr, "    swap_points ERROR:  triangle T[%d] doesn't "
              "point to edge E[%d]\n", pCurT-T, pCurE-E);
            exit(5);
        }
#endif

        j_save = j;
        j1 = (j_save+1) % 3;
        if ( pCurT->pV[j1] == pCurV ) {
            j_cur = j1;                         /* for next time */
            pCurE = pCurT->pE[ (j_save+2)%3 ];  /* overwritten for folded T */
        } else {
            j_cur = (j_save+2) % 3;             /* next time */
            pCurE = pCurT->pE[j1];
        }

    } while (pCurT != pFirstFoldedT);  /* end do-while */

/*---------------------------------------------------------------------------
    Recalculate edge normals for two formerly folded triangles.
  ---------------------------------------------------------------------------*/

    for ( m = 0  ;  m < 2  ;  ++m ) {

        pCurT = pCenterE->pT[m];

        for ( j = 0  ;  j < 3  ;  j++ )     /* loop over edges/vertices */
            for ( i = 0  ;  i < 3  ;  i++ )   /* loop over coords */
                s[j][i] = pCurT->pV[j]->r[i] - pCurT->pV[(j+1)%3]->r[i];

        for ( j = 0  ;  j < 3  ;  j++ ) {      /* three vertices again  */
            j1 = (j+1)%3;
            /* loop over cartesian coords:  calc normal to opposite edge */
            for ( i = 0  ;  i < 3  ;  i++ )
                pCurT->n[j][i] = s[j][i] - DOT(s[j],s[j1]) * s[j1][i] /
                                             SQR(pCurT->pE[j]->len);
            /* loop over cartesian coords:  make edge normals unit */
            norm = NORM(pCurT->n[j]);
            for ( i = 0  ;  i < 3  ;  i++ )
                pCurT->n[j][i] /= norm;

        } /* end j-loop (vertices) */
    } /* end m-loop (two triangles) */

/*---------------------------------------------------------------------------
    Switch masses of two central vertices (since surrounding triangles have
    been swapped), and print CPU time.
  ---------------------------------------------------------------------------*/

    dt = pCurV->mass;           /* use dt for temporary storage */
    pCurV->mass = pAltV->mass;
    pAltV->mass = dt;
#if FALSE
    fprintf(dbgfile, "    V[%d] new mass:  %11.3e\n", pCurV-V, pCurV->mass);
    fprintf(dbgfile, "    V[%d] new mass:  %11.3e\n", pAltV-V, pAltV->mass);
#endif

#if FALSE      /* double-check that edge-normals really are normal to edges */
    for ( m = 0  ;  m < 9  ;  ++m ) {     /* 9 "spokes" edges */
        for ( i = 0  ;  i < 3  ;  ++i )
            s[0][i] = pE[m]->pV[0]->r[i] - pE[m]->pV[1]->r[i];  /* temp vec */
        for ( j = 0  ;  j < 2  ;  ++j )
            fprintf(dbgfile, "    ### E[%d] -> T[%d];  normal = %11.3e;  vtx:"
              " %d %d\n", pE[m]-E, pE[m]->pT[j]-T, DOT(pE[m]->n[j],s[0]),
              pE[m]->pV[0]-V, pE[m]->pV[1]-V );
    }
#endif

    dt = SECOND() - t1;
    fprintf(outfile, "  swap_points() time:          %8.6f\n", dt);
    swappts_time += dt;

  } /* end function swap_points() */
