/*---------------------------------------------------------------------------

           SSSSSS   HH    HH     AA     RRRRRRR   EEEEEEEE  DDDDDD    
          SS        HH    HH    AAAA    RR    RR  EE        DD    DD  
           SSSSSS   HH    HH   AA  AA   RR    RR  EEEEEEE   DD    DD  
                SS  HHHHHHHH  AA    AA  RRRRRR    EE        DD    DD  
          SS    SS  HH    HH  AAAAAAAA  RR   RR   EE        DD    DD  
           SSSSSS   HH    HH  AA    AA  RR    RR  EEEEEEEE  DDDDDD    

  ---------------------------------------------------------------------------

  shared.c

  Functions shared by loader and evolver.  Routines include swap_edges(),
  print_phys_vars(), write_edges(), write_tri(), and write_data() (none of 
  which is used by the loader when NOTRI is defined).

  ---------------------------------------------------------------------------

     Version:  Y-MP/PC                               Greg Roelofs
     Source:   sh310.c on eagle                      20 February 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:  scc -c sh310.c
     on PC:    cl -c -AS -G2 -FPi87 sh310.c

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
  -> sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotload.c:    contains MS-DOS graphics routines

     points.c:      stand-alone program to create and plot points only

     f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

  Version history:

     sh285.c:  renamed from prphy281.c; made separately compilable; moved 
                swap_edges() from repair.c (now shared by loader); moved and
                combined write_data()/save_data() from d###.c/h###.c
     sh290.c:  header filename changed to fd.h
     sh291.c:  changed write_edges() output from VAX VMS to Cray format for
                new tplot.c
     sh292.c:  changed write_tri() output from VAX VMS to Cray format
     sh300.c:  fixed bug in folded-triangle criteria (swap_edges()); slight
                 modification to debugging output and comments
     sh310.c:  modified to use file_type when writing data files

  ---------------------------------------------------------------------------*/



#include "fd.h"

#define MAX_PASS   5



/***************************/
/*  Function swap_edges()  */
/***************************/

int swap_edges( int pass )
{
    char            error_msg[256];
    int             i, j, j1, j2, k, m, index1[2], index2[2], index3[2];
    double          t1, dt, new_len, costh[8], cur_max_cos, alt_max_cos;
    double          d0d5, d1d2, d1d3, d1d4, d1d5, d2d4, d3d5, d4d4, d5d5;
    double          Mass, new_area0, fraction[2];
    vector          d[6];
    struct edge     *pE[4];
    struct vertex   *pV[2];



/*---------------------------------------------------------------------------
    Before checking whether a given edge should be reassigned to a new pair
    of vertices, must first make sure that the quadrilateral defined by the
    edge (via its two adjacent triangles) does, in fact, exist; then can
    define the two "other" vertices (those not pointed at by the edge E[k]).
  ---------------------------------------------------------------------------*/

    t1 = SECOND();
    num_swaps = 0;

    /* loop through edges (each defines a possible quadrilateral) */
    for ( k = 0  ;  k < numE  ;  k++ ) {

        /* make sure swap-check is necessary and that quadrilateral exists */
        if (!(E[k].status & (E_ON_HULL | E_UNMODIFIED))) {
            E[k].status |= E_UNMODIFIED;   /* re-set bit */

            /* loop over adjacent triangles:  find the two "other" vertices */
            for ( j = 0  ;  j < 2  ;  j++ ) {
                /* loop over triangle's edges (vertex has same index as E[k]) */
                for ( i = 0 ;  (i < 3) && (E[k].pT[j]->pE[i] != &E[k])  ; i++ )
                    ;

                /* pV points to the vertex of triangle T[j] which is NOT
                 * currently attached to edge E[k] */
                pV[j] = E[k].pT[j]->pV[i];
                index1[j] = i;

            }  /* end j-loop (triangles) */

        /*-------------------------------------------------------------------
            Calculate vector components once for each quadrilateral; these
            are used for the cross-product and dot-product calculations
            below.
          -------------------------------------------------------------------*/

            for ( i = 0  ;  i < 3  ;  i++ ) {
                d[0][i] = E[k].pV[1]->r[i] -      pV[0]->r[i];
                d[1][i] =      pV[1]->r[i] - E[k].pV[0]->r[i];
                d[2][i] =      pV[0]->r[i] - E[k].pV[0]->r[i];
                d[3][i] = E[k].pV[1]->r[i] -      pV[1]->r[i];
                d[4][i] = E[k].pV[1]->r[i] - E[k].pV[0]->r[i];
                d[5][i] =      pV[0]->r[i] -      pV[1]->r[i];
            }

            d0d5 = DOT(d[0],d[5]);
            d1d2 = DOT(d[1],d[2]);
            d1d3 = DOT(d[1],d[3]);  /* used only once */
            d1d4 = DOT(d[1],d[4]);
            d1d5 = DOT(d[1],d[5]);
            d2d4 = DOT(d[2],d[4]);
            d3d5 = DOT(d[3],d[5]);
            d4d4 = DOT(d[4],d[4]);  /* used only once */
            d5d5 = DOT(d[5],d[5]);

            new_len = sqrt(d5d5);

        /*-------------------------------------------------------------------
            Check for integrity of the tentative new configuration:  first
            compute dot-product of two cross-products:  d[3] x d[5]  and
            d[5] x -d[1].  If cross-products point in the same direction
            (i.e., they have a positive scalar product; note sign on d[1]:
            defined to point away from d[4], i.e., *toward* d[5]), then the
            new configuration is valid (i.e., unfolded) and we may continue.
            We will do a similar test below for the validity of the current
            configuration.  [Notes 1992.02.14]
          -------------------------------------------------------------------*/

            if (d1d3*d5d5 > d1d5*d3d5) {

                /* now go ahead and define the four bounding edges, pE[j];
                 * start by looping over two triangles adjacent to E[k] */
                for ( j = 0  ;  j < 2  ;  j++ ) {
                    j1 = (index1[j] + 1) % 3;
                    j2 = (index1[j] + 2) % 3;

                /*-----------------------------------------------------------
                    Find the edge of triangle T[j] which is opposite to ver-
                    tex V[j] of old E[k].  If a swap is necessary, this edge
                    will be traded to the other triangle (and vice versa:
                    save index so know which edge to replace later).  Also
                    save the other edge of T[j] which is NOT E[k], for dot-
                    product angle-calculations later.
                  -----------------------------------------------------------*/

                    if ( E[k].pT[j]->pV[j1] == E[k].pV[j] ) {
                        pE[j]   = E[k].pT[j]->pE[j1];
                        pE[j+2] = E[k].pT[j]->pE[j2];
                        index2[j] = j1;
                        index3[j] = j2;
                    } else {
                        pE[j]   = E[k].pT[j]->pE[j2];
                        pE[j+2] = E[k].pT[j]->pE[j1];
                        index2[j] = j2;
                        index3[j] = j1;
                    }

                }  /* end j-loop (triangles) */

            /*---------------------------------------------------------------
                Now can calculate cosines of all of the "split-angles," i.e.,
                the eight angles on either side of either diagonal (see dia-
                gram below).
              ---------------------------------------------------------------*/

                /* SKIP THIS PART IF E_FORCE_SWAP BIT IS SET */
                costh[0] =       d2d4     / (pE[2]->len * E[k].len);
                costh[1] =       d1d4     / (pE[1]->len * E[k].len);
                costh[2] = DOT(d[3],d[4]) / (pE[3]->len * E[k].len);
                costh[3] = DOT(d[0],d[4]) / (pE[0]->len * E[k].len);

                costh[4] = -     d1d5       / (pE[1]->len * new_len);
                costh[5] =       d3d5       / (pE[3]->len * new_len);
                costh[6] = -     d0d5       / (pE[0]->len * new_len);
                costh[7] =   DOT(d[2],d[5]) / (pE[2]->len * new_len);

                cur_max_cos = MAX( MAX(costh[0], costh[1]),
                                   MAX(costh[2], costh[3]) );
                alt_max_cos = MAX( MAX(costh[4], costh[5]),
                                   MAX(costh[6], costh[7]) );


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                           *
 *               Before:                               After:                *
 *               -------            Note:              ------                *
 *                              (#) <--> d[#]                                *
 *         index3[0] index2[1]                       index2[1]               *
 *              E[k].pV[1]                               _                   *
 *                 /|\                                  / \                  *
 *         pE[0]  / | \  pE[3]                  pE[0]  /   \  pE[3]          *
 *              (0) | (3)                            (0) 1 (3)               *
 *              /  E[k] \                E[k].pV[0]  /       \  E[k].pV[1]   *
 *       pV[0] /   (4)   \ pV[1]              pV[0] /___(5)___\ pV[1]        *
 *   index1[0] \ 0  |  1 / index1[1]      index1[0] \         / index1[1]    *
 *              \   |   /                 index3[1]  \   0   /  index3[0]    *
 *              (2) | (1)                            (2)   (1)               *
 *         pE[2]  \ | /  pE[1]                  pE[2]  \   /  pE[1]          *
 *                 \|/                                  \_/                  *
 *              E[k].pV[0]                                                   *
 *         index2[0] index3[1]                       index2[0]               *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



            /*---------------------------------------------------------------
                If the current configuration has the smallest angle (biggest
                cosine), or if it is folded (dot-product of cross-products
                really IS negative this time), [or if the FORCE_SWAP bit is
                set for some other reason], swap diagonals:  that is, "maxi-
                mize the minimum angle."  (pE[2] and pE[3] are used only in
                the dot-product calculations of the split angles; as edges
                they remain with their respective triangles).  [Revised fold
                criterion:  notes 1992.02.14]
              ---------------------------------------------------------------*/

                if ((cur_max_cos > alt_max_cos) || (d1d2*d4d4 > d1d4*d2d4)) {
                    /* || (E[k].status & E_FORCE_SWAP) */

                    ++num_swaps;
                    /* E[k].status &= ~E_FORCE_SWAP; */

                /*-----------------------------------------------------------
                    Adjust physical variables first, before pointers get
                    screwed up (for now, only masses adjusted: isothermal).
                    Begin by subtracting off old triangle masses from four
                    vertices.
                  -----------------------------------------------------------*/

                    for (j = 0;  j < 2;  j++) {
                        pV[j]->mass      -=   E[k].pT[j]->mass / 3.;
                        E[k].pV[j]->mass -= ( E[k].pT[0]->mass +
                                              E[k].pT[1]->mass ) / 3.;
                    }

                    Mass = E[k].pT[0]->mass + E[k].pT[1]->mass;

                    new_area0 = sqrt(SQR( XCROSS(d[5],d[1]) ) +
                                     SQR( YCROSS(d[5],d[1]) ) +
                                     SQR( ZCROSS(d[5],d[1]) ));  /* times .5 */

                    fraction[0] = new_area0 / ( new_area0 +
                                    sqrt(SQR( XCROSS(d[3],d[5]) ) +
                                         SQR( YCROSS(d[3],d[5]) ) +
                                         SQR( ZCROSS(d[3],d[5]) )) );
                    fraction[1] = 1. - fraction[0];

                /*-----------------------------------------------------------
                    Begin loop over triangles adjacent to E[k] (well, some
                    of the time, at any rate:  m is an index for all sorts
                    of things in this loop).  Start by calculating new tri-
                    angle masses and adding them back to all four vertices.
                  -----------------------------------------------------------*/

                    for (m = 0;  m < 2;  m++) {
                        E[k].pT[m]->mass = fraction[m] * Mass;

                        E[k].pV[m]->mass += E[k].pT[m]->mass / 3.;
                        pV[0]->mass      += E[k].pT[m]->mass / 3.;
                        pV[1]->mass      += E[k].pT[m]->mass / 3.;

                    /*-------------------------------------------------------
                        T -> E :   Loop over triangles adjacent to E[k], up-
                        T -> V :   dating edges and vertices.  One edge of
                                   each triangle remains unchanged (index3),
                        one gets swapped to a different index in the same
                        triangle (E[k]), and one gets swapped to the other
                        triangle.  T[m] gets vertex V[m] (arbitrary choice;
                        index3).  Point is to ensure that edges and their
                        opposite vertices still correspond after the swap.
                        Also reset E[k] normal pointers (JUST the pointers,
                        not the actual vectors!) according to new indices
                        of triangle pointers [notes 1991.1.6].
                      -------------------------------------------------------*/

                        E[k].pT[m]->pE[index2[m]]  =  &E[k];
                        E[k].pT[m]->pE[index1[m]]  =  pE[1-m];
                        E[k].pT[m]->pV[index3[m]]  =  pV[1-m];

                        E[k].n[m] = E[k].pT[m]->n[index2[m]];

                    /*-------------------------------------------------------
                        E -> T :   Of the two triangles pointed at by edge
                                   pE[m], find which one is one of "our"
                        triangles, and replace it with the other of "ours."
                        Also reset pE[m] normal pointer according to new in-
                        dices of triangle pointers (JUST the pointer, not
                        the actual vector--see below) [notes 1991.1.6].
                      -------------------------------------------------------*/

                        if (pE[m]->pT[0] == E[k].pT[m]) {
                            pE[m]->pT[0] = E[k].pT[1-m];
                            pE[m]->n[0]  = E[k].pT[1-m]->n[index1[1-m]];
                        } else {
                            pE[m]->pT[1] = E[k].pT[1-m];
                            pE[m]->n[1]  = E[k].pT[1-m]->n[index1[1-m]];
                        }

                    /*-------------------------------------------------------
                        E -> V :   Point edge E[k] at its new vertices.
                      -------------------------------------------------------*/

                        E[k].pV[m] = pV[m];

                    } /* end m-loop (triangles, edges, or vertices) */

                /*-----------------------------------------------------------
                    Store E[k]'s new length and clear unmodified-flags of
                    surrounding edges.
                  -----------------------------------------------------------*/

                    E[k].len = new_len;
                    for (m = 0;  m < 4;  m++)
                        pE[m]->status &= ~E_UNMODIFIED;

                /*-----------------------------------------------------------
                    And finally, recalculate all edge normals within the new
                    triangles.  Use existing vectors (d[]) instead of s[].
                    Each normal points inward from its edge (or outward from
                    the opposing vertex).
                  -----------------------------------------------------------*/

                    for (i = 0;  i < 3;  ++i) {   /* cartesian indices */
                        E[k].pT[0]->n[index1[0]][i]  =
                           d[2][i] - (d1d2 / SQR(pE[1]->len)) * d[1][i];
                        E[k].pT[0]->n[index2[0]][i]  =
                          -d[1][i] + (d1d5 / SQR(E[k].len)) * d[5][i];
                        E[k].pT[0]->n[index3[0]][i]  =
                           d[1][i] - (d1d2 / SQR(pE[2]->len)) * d[2][i];

                        E[k].pT[1]->n[index1[1]][i]  =
                          -d[5][i] + (d0d5 / SQR(pE[0]->len)) * d[0][i];
                        E[k].pT[1]->n[index2[1]][i]  =
                           d[3][i] - (d3d5 / SQR(E[k].len)) * d[5][i];
                        E[k].pT[1]->n[index3[1]][i]  =
                           d[5][i] - (d3d5 / SQR(pE[3]->len)) * d[3][i];
                    }

                    for ( m = 0  ;  m < 2  ;  ++m )          /* triangles */
                        for ( j = 0  ;  j < 3  ;  ++j ) {    /* edges/normals */
                            double   norm = NORM( E[k].pT[m]->n[j] );
                            for ( i = 0  ;  i < 3  ;  ++i )  /* cart. indices */
                                E[k].pT[m]->n[j][i] /= norm;
                        }

                } /* endif (swap diagonals) */
            } /* endif (new configuration valid) */
        } /* endif (swap-check necessary and quadrilateral exists) */
    } /* end k-loop (edges) */

    if ( num_swaps || !pass )    /* always print on pass 0, or if had swaps */
      fprintf(outfile, "  %3d swap(s)", num_swaps);
    else
      fprintf(outfile, "             ");

    dt = SECOND() - t1;
    fprintf(outfile, " (pass %d); time:  %8.6f\n", pass, dt);
    swapE_time += dt;

    if (pass < MAX_PASS)
        return num_swaps;
    else {
        sprintf(error_msg, "  swap_edges WARNING:  too many passes (%d)\n",
          pass);
        fprintf(outfile, error_msg);
        fprintf(stderr, error_msg);
        if (f_version)
            fprintf(dbgfile, error_msg);
        return 0;
    }

}  /* end function swap_edges() */





/*******************************/
/*  Function print_phys_vars() */
/*******************************/

void print_phys_vars( void )
  {
    double   tmp;
    double   cm_length = cm_pc * pc_length;
    double   s_time = s_Myr * Myr_time;
    double   g_mass = g_Msol * Msol_mass;



/*--------------------------------------------------------------------------
    Log summary of physical values (in various units) and model parameters 
    to regular output file.
  --------------------------------------------------------------------------*/

    fprintf(outfile, "\n\n   ---------------------------------------------"
      "----------------------------\n");
    fprintf(outfile, "   Units scale:  mass unit = %10.4e M_sol\n", Msol_mass);
    fprintf(outfile, "                           = %10.4e g\n",
      g_Msol*Msol_mass);
    fprintf(outfile, "                 length unit = %5.3f kpc\n", 
      pc_length/1000.);
    fprintf(outfile, "                             = %10.4e cm\n", cm_length);
    fprintf(outfile, "                 time unit = %5.3f Myr\n", Myr_time);
    fprintf(outfile, "                           = %10.4e s\n", s_time);
    fprintf(outfile, "                 velocities = %5.1f km/s\n", 
      tmp = (1.e-5 * cm_length / s_time) );
    fprintf(outfile, "                            = %5.1f pc/Myr\n", 
      (pc_length / Myr_time) );
    fprintf(outfile, "                 G = %10.4e cm^3/(g sec^2)\n", G_cgs);
    fprintf(outfile, "                   = %10.4e pc^3/(M_sol Myr^2)\n", G_pMM);
    fprintf(outfile, "                   = %10.4e\n", G_LMT);
    fprintf(outfile, "   ---------------------------------------------"
      "----------------------------\n");
    fprintf(outfile, "   Potential:  oblate Schwarzschild, roughly E%d"
      "-type elliptical;\n", En);
    fprintf(outfile, "                 oblateness parameters:  c1 = %6.5f\n",
      2.*c1half);
    fprintf(outfile, "                                         c2 = %6.5f\n",
      c2);
    fprintf(outfile, "                 core radius  = %4.1f\n", Rcore);
    fprintf(outfile, "                              = %4.1f kpc\n", 
      Rcore * pc_length / 1000.);
    fprintf(outfile, "                 core density = %6.5f\n", rho0);
    fprintf(outfile, "                              = %6.5f M_sol/pc^3\n",
      rho0 * Msol_mass / powri(pc_length,3) );
    fprintf(outfile, "                 potential at origin = %7.4f\n", phi0);
    fprintf(outfile, "                                     = -(%5.1f km/s)^2\n",
      sqrt(fabs(phi0)) * tmp);
    fprintf(outfile, "   ---------------------------------------------"
      "----------------------------\n");
    fprintf(outfile, "   Gas disk:  exponential;\n");
    fprintf(outfile, "              central surface density = %7.5f\n", mu0);
    fprintf(outfile, "                                      = %10.4e "
      "M_sol/pc^2\n", mu0*Msol_mass/SQR(pc_length) );
    fprintf(outfile, "                                      = %10.4e "
      "g/cm^2\n", mu0*g_mass/SQR(cm_length) );
    fprintf(outfile, "              exponential scale length = %5.2f\n", bB);
    fprintf(outfile, "                                       = %5.2f kpc\n",
      bB*pc_length*1.e-3 );
    fprintf(outfile, "              mass = %6.4f\n", Mdisk);
    fprintf(outfile, "                   = %10.4e M_sol\n", Mdisk*Msol_mass);
    fprintf(outfile, "              outer radius = %5.2f\n", Rdisk);
    fprintf(outfile, "                           = %5.2f kpc\n", 
      Rdisk*pc_length*1.e-3 );
    fprintf(outfile, "              inner radius (\"flat\" part) = %5.2f\n",
      Rholm);
    fprintf(outfile, "                                         = %5.2f kpc\n", 
      Rholm*pc_length*1.e-3 );
    fprintf(outfile, "              dispersion velocity = %5.4f  [ = "
      "sqrt( kT/m_cloud ) ]\n", v_rms);
    fprintf(outfile, "                                  = %6.3f km/s\n",
      v_rms*cm_length*1.e-5/s_time );
    fprintf(outfile, "              dynamic viscosity = %10.4e\n", vis_dyn);
    fprintf(outfile, "                                = %10.4e g/cm-s\n",
      vis_dyn*g_mass/(cm_length*s_time) );
    fprintf(outfile, "              warp inclination = %4.1f deg (of "
      "outer disk w.r.t.\n                                         "
      "  inner section)\n", Inclination);
    fprintf(outfile, "   ---------------------------------------------"
      "----------------------------\n\n\n");

  } /* end function print_phys_vars() */





/****************************/
/*  Function write_edges()  */     /* for standard triangulation plots */
/****************************/

void write_edges( void )
  {
    int      i, total_bytes;
    double   *edges, *p, t1;
    FILE     *edgfile;


/*--------------------------------------------------------------------------
    Allocate buffer in which to copy coordinates, open output file, and 
    write karr[] and rarr[].
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    sprintf(filename, "d%03df%03dv%05.1f", d_version, f_version, Time);
    file_type = TRI3D;

    p = edges = (double *) malloc((unsigned) (6*numE)*sizeof(double));
    if (!edges) {
        printf("error:  couldn't allocate memory to write [ %s ]\n", filename);
        errorlevel = 10;
        return;
    }
    if ((edgfile = fopen(filename,"ab")) == NULL) {
        printf("error:  couldn't open file [ %s ]\n", filename);
        errorlevel = 10;
        free(edges);
        return;
    }

    total_bytes = fwrite(karr, sizeof(int), 256, edgfile) * sizeof(int);
    total_bytes += fwrite(rarr, sizeof(double), 256, edgfile) * sizeof(double);

/*--------------------------------------------------------------------------
    Copy endpoints of edges to temporary array, write to file, and clean
    up.  [This was formerly done in VAX VMS format with a FORTRAN routine,
    VAXWR.]
  --------------------------------------------------------------------------*/

    for (i = 0;  i < numE;  ++i) {
        *p++ = E[i].pV[0]->r[0];
        *p++ = E[i].pV[0]->r[1];
        *p++ = E[i].pV[0]->r[2];
        *p++ = E[i].pV[1]->r[0];
        *p++ = E[i].pV[1]->r[1];
        *p++ = E[i].pV[1]->r[2];
    }

    total_bytes += fwrite(edges, sizeof(double), 6*numE, edgfile) *
                                 sizeof(double);
    fclose(edgfile);
    free(edges);

    fprintf(outfile,"  write_edges():  %d edges (%d bytes) written to [ %s ]\n",
      numE, total_bytes, filename);

    write_time += SECOND() - t1;

} /* end function write_edges() */





/**************************/
/*  Function write_tri()  */     /* for rendered surface plots */
/**************************/

void write_tri( void )
  {
    int      i, j, k, *q, *tri, total_bytes;
    double   t1, *p, *verts;
    FILE     *surffile;


/*--------------------------------------------------------------------------
    Allocate buffers in which to copy coordinates and "pointers," open out-
    put file, and write karr[] and rarr[].  Filename is "d###f###s###.#".
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    sprintf(filename, "d%03df%03ds%05.1f", d_version, f_version, Time);
    file_type = SURF;

    p = verts = (double *) malloc((unsigned) (3*numV)*sizeof(double));
    q = tri = (int *) malloc((unsigned) (3*numT)*sizeof(int));
    if (!verts || !tri) {
        printf("error:  couldn't allocate memory to write [ %s ]\n", filename);
        errorlevel = 10;
        if (verts)
            free(verts);
        else if (tri)
            free(tri);
        return;
    }
    if ((surffile = fopen(filename,"ab")) == NULL) {
        printf("error:  couldn't open file [ %s ]\n", filename);
        errorlevel = 10;
        free(verts);
        free(tri);
        return;
    }

    total_bytes = fwrite(karr, sizeof(int), 256, surffile) * sizeof(int);
    total_bytes += fwrite(rarr, sizeof(double), 256, surffile) * sizeof(double);

/*--------------------------------------------------------------------------
    Copy vertex coordinates into verts[] array and triangle "pointers" to 
    vertices into tri[].  The arrays will be picked up by a C program but 
    will be passed to (FORTRAN) DISSPLA routines, so prepare for that now.
  --------------------------------------------------------------------------*/

    for (i = 0;  i < 3;  ++i)
        for (k = 0;  k < numV;  ++k)
            *p++ = V[k].r[i];

    for (k = 0;  k < numT;  ++k)
        for (j = 0;  j < 3;  ++j)
            *q++ = T[k].pV[j] - V + 1;   /* adjust for FORTRAN indexing */

    total_bytes += fwrite(verts, sizeof(double), 3*numV, surffile) *
                                 sizeof(double);
    total_bytes += fwrite(tri, sizeof(int), 3*numT, surffile) *
                               sizeof(int);
    fclose(surffile);
    free(verts);
    free(tri);

    fprintf(outfile, "  write_tri():  %d triangles (%d bytes) written to "
      "[ %s ]\n", numT, total_bytes, filename);

    write_time += SECOND() - t1;

} /* end function write_tri() */





/***************************/
/*  Function write_data()  */     /* for checkpointing and restarting the run */
/***************************/

void write_data( void )
  {
    long     kbytes, rbytes, pbytes, Ebytes, Tbytes, Vbytes;
    double   t1;
    void     *pETV[]={ (void *)E, (void *)T, (void *)V };
    FILE     *savefile;
/*  struct edge       *pE = E;
    struct triangle   *pT = T;
    struct vertex     *pV = V;  */



/*--------------------------------------------------------------------------
    Open output file and write triangulation data.
  --------------------------------------------------------------------------*/

#ifdef MSDOS
    sprintf(filename, "d%03dt000.0", d_version);
#else /* Y-MP */
    t1 = SECOND();
    sprintf(filename, "d%03df%03dt%05.1f", d_version, f_version, Time);
#endif
    file_type = FD_DATA;

    if ((savefile = fopen(filename,"wb")) == NULL) {
        printf("error:  couldn't open save file [ %s ]\n", filename);
        exit(1);
    }
    if (!f_version)
        fprintf(outfile, "\nWriting triangulation data, you betcha.\n");

    kbytes = fwrite(karr, sizeof(int), 256, savefile) * sizeof(int);
    rbytes = fwrite(rarr, sizeof(double), 256, savefile) * sizeof(double);

    pbytes = fwrite(pETV, sizeof(void *), 3, savefile) *
                          sizeof(void *);

    Ebytes = fwrite(E, sizeof(struct edge), numE, savefile) *
                       sizeof(struct edge);
    Tbytes = fwrite(T, sizeof(struct triangle), numT, savefile) *
                       sizeof(struct triangle);
    Vbytes = fwrite(V, sizeof(struct vertex), numV, savefile) *
                       sizeof(struct vertex);

    fclose(savefile);

    if (!f_version) {
        fprintf(outfile, "  Size of data elements (bytes):\n");
        fprintf(outfile, "  karr[]:  %6d      E[]:%8d\n", kbytes, Ebytes);
        fprintf(outfile, "  rarr[]:  %6d      T[]:%8d\n", rbytes, Tbytes);
        fprintf(outfile, "  pointers:%6d      V[]:%8d\n", pbytes, Vbytes);
        printf("\nTotal bytes written to [ %s ]:  %ld\n", filename,
          (kbytes+rbytes+pbytes+Ebytes+Tbytes+Vbytes) );
    }
    fprintf(outfile, "\nTotal bytes written to [ %s ]:  %ld\n", filename,
      (kbytes+rbytes+pbytes+Ebytes+Tbytes+Vbytes) );

    write_time += SECOND() - t1;

  } /* end function write_data() */
