/*---------------------------------------------------------------------------

  TTTTTTTT EEEEEEE  SSSSSS   SSSSSS  EEEEEEE LL         AA   TTTTTTTT EEEEEEE 
     TT    EE      SS     ~ SS     ~ EE      LL        AAAA     TT    EE      
     TT    EEEEEE   SSSSSS   SSSSSS  EEEEEE  LL       AA  AA    TT    EEEEEE  
     TT    EE            SS       SS EE      LL      AA    AA   TT    EE      
     TT    EE      SS    SS SS    SS EE      LL      AAAAAAAA   TT    EE      
     TT    EEEEEEE  SSSSSS   SSSSSS  EEEEEEE LLLLLLL AA    AA   TT    EEEEEEE 

  ---------------------------------------------------------------------------

  tesselate.c

  This file contains the main tesselation routines for delaunay.c.  It is
  designed to be compiled on its own, since the code has worked flawlessly
  for months.  It must be recompiled for EGA plotting (or if the DIAG flag
  is set, and then only if there is a problem with the tesselation).

  ---------------------------------------------------------------------------

     Version:  Y-MP/PC                               Greg Roelofs
     Source:   tesselate.c on eagle and oddjob       23 February 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:   scc -c [-DDIAG] tesselate.c
     on PC:     cl -c -AS -G2 -FPi87 [-DDIAG] [-DEGA] tesselate.c

     DIAG:  print diagnostics to standard output (see global variable Eflag)
     EGA:   plot tesselation as it is generated

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotpts.c:     MS-DOS graphics routines for points load
     plotload.c:    MS-DOS graphics routines for tesselation

     p###.c:        stand-alone program to create and plot points only

     f###.c:        main evolution code (evolver)
     update###.c:   physics (update_points, used only by f###.c)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      most grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

   Version history:

     d230.c:  code last modified (read:  bug fixed) sometime around here
     d270.c:  tesselation routines split out into separate file (this file)
     d280.c:  tesselate() itself moved here from d280.c; using header file
     d285.c:  timing info added
     d290.c:  changed name of header file to fd.h
     d292.c:  compound for-loop-exit-condition bugs fixed (comma --> &&)
     d310.c:  Vindex --> numDirV; Pindex --> numPt; num_points --> numV

  ---------------------------------------------------------------------------*/




#define LOADER
#include "fd.h"       /* includes, defines, typedefs, global prototypes */





/**********************************/
/*  Local Prototypes and Globals  */
/**********************************/

static void             initialize_lists       ( void );
static void             circumcircle           ( struct DVERTEX * );
static struct POINT    *pClosestPt             ( void );
static struct DVERTEX  *pDeletedDirV           ( struct POINT * );
static void             check_neighbors        ( struct DVERTEX * );
static void             find_contig_points     ( void );
static void             mark_old_contiguities  ( void );
static void             add_new_vertices       ( void );
static void             modify_points          ( void );
static struct POINT   **request_heap_space     ( int );

static int   Pindex;





/**************************/
/*  Function tesselate()  */
/**************************/

void tesselate( void )
{
    struct DVERTEX   *pSomeDelDirV;
    double           t1;
#ifdef DIAG
    int              j;
    struct POINT     **pTmpHeap;
#elif defined(EGA)
    int              key, automatic=FALSE;
#endif



/*--------------------------------------------------------------------------
    Initialize stuff.
  --------------------------------------------------------------------------*/

    t1 = SECOND();

    printf("  %d points created; beginning tesselation.\n", numPt);

    initialize_lists();
#ifdef EGA
    getch();
    initialize_plot();
#endif /* EGA */

/*--------------------------------------------------------------------------
    Add points to structure, one by one; start with i=4 since i=0,1,2 are
    three "fake" convex hull points, and i=3 (first "real" point) was added
    by hand.  [Note 1990.6.13:  i=3 is not presently a "real" point (will
    be removed at end), but treat it as such for purposes of this program.]
    [Note 1991.1.8:  last note was only for versions with hole in middle.]
  --------------------------------------------------------------------------*/

    for (Pindex = 4;  Pindex < numPt;  ++Pindex) {
        pNewPt = &Pt[ Pindex ];

#ifdef DIAG /*...............................................................*/
        if (Eflag) {
            printf("\n\n\n-------------- Heap dump ---------------\n");
            printf("--- (regular, before starting Pt[%d]) ---\n", Pindex);
            for ( j = 0  ;  j < Pindex  ;  j++ )
                if (Pt[j].pHeap)
                    printf("  Pt[%d].pHeap -> Heap[%d]\n", j,
                      (Pt[j].pHeap-&Heap[0]));
                else
                    printf("  Pt[%d].pHeap ->    0\n", j);
            pTmpHeap = &Heap[0];
            j = 0;
            while (pTmpHeap < pBottomHeap) {
                if (*pTmpHeap)
                  printf("   Heap[%d] -> Pt[%d]\n", j, (*pTmpHeap-&Pt[0]));
                else
                  printf("   Heap[%d] ->   0\n", j);
                ++pTmpHeap;
                ++j;
            }
            printf("------------ End Heap Dump -------------\n");

            printf("\nBeginning Pt[%d]\n", Pindex);
            printf("  position:  %f,    %f\n", Pt[Pindex].x, Pt[Pindex].y);
        } /* end if (Eflag) */
#elif defined(EGA) /*........................................................*/
        plot_tiles(1);
#endif /*....................................................................*/


    /*----------------------------------------------------------------------
        Step 1:  find a vertex that will be deleted by NewPt (= global ptr)
      ----------------------------------------------------------------------*/
        pSomeDelDirV = pDeletedDirV( pClosestPt( /* pNewPt */ ) );

    /*----------------------------------------------------------------------
        Step 2:  put list of vertices-to-be-deleted in pDelDirV[] array
      ----------------------------------------------------------------------*/
        num_DelDirV = 0;
        check_neighbors( pSomeDelDirV );
        max_DelDirV = MAX(num_DelDirV, max_DelDirV);

#ifdef DIAG
        if (Eflag) {
            printf("num_DelDirV = %d\n", num_DelDirV);
            printf("max_DelDirV = %d\n", max_DelDirV);
        }
#endif /* DIAG */

    /*----------------------------------------------------------------------
        Step 3:  put list of points contiguous to NewPt in pContigPt[]
      ----------------------------------------------------------------------*/
        find_contig_points();
        max_ContigPt = MAX(num_ContigPt, max_ContigPt);

#ifdef DIAG
        if (Eflag) {
            printf("num_ContigPt = %d\n", num_ContigPt);
            printf("max_ContigPt = %d\n", max_ContigPt);
        }
#endif /* DIAG */

    /*----------------------------------------------------------------------
        Step 4:  zero any contiguities that will be removed by NewPt
      ----------------------------------------------------------------------*/
        mark_old_contiguities();

    /*----------------------------------------------------------------------
        Steps 5, 6:  update lists to reflect presence of NewPt
      ----------------------------------------------------------------------*/
        add_new_vertices();
        modify_points();

#ifdef EGA /*................................................................*/
        if (!automatic) {
            key = getch();
            if (tolower(key) == 'g')   /* go continuously */
              automatic = TRUE;
        }
        plot_tiles(0);
        if (!automatic) {
            key = getch();
            if (tolower(key) == 'g')   /* go continuously */
              automatic = TRUE;
        }
#endif /* EGA ...............................................................*/

    } /* end for-loop (Pindex:  points) */

#ifdef EGA /*................................................................*/
        plot_tiles(1);   /* if this depends on Pindex, change to numPt */
        _settextcolor(9);
        _settextposition(5, 1);
        _outtext("Done.  To\ncontinue,\npress ESC.\n");

        while (key != 27)   /* wait for ESC key */
          key = getch();

/*      _setvideomode( _DEFAULTMODE );   /* reset to text mode before exit
                                          * (don't need to:  triangles yet) */
#endif /*...................................................................*/

    if (hole > 0.)
        --numV;    /* correct for (removed) central point */

    fprintf(outfile, "\nTesselation complete, you betcha.\n");
    fprintf(outfile, "  number of (real) points loaded:        %7d\n",
      numV);
    fprintf(outfile, "  number of Dirichlet vertices created:  %7d\n", numDirV);
    fprintf(outfile, "  most deleted vertices:  %4d\n", max_DelDirV);
    fprintf(outfile, "  most contiguous points: %4d\n", max_ContigPt);
    fprintf(outfile, "  heap compacted %d times (size:  %d bytes)\n",
      num_compactions, HEAP_SIZE*sizeof( struct POINT * ) );

    tessel_time = SECOND() - t1;

} /* end function tesselate() */





/*********************************/
/*  Function initialize_lists()  */
/*********************************/

static void initialize_lists( void )
  {
    int   i;



/*--------------------------------------------------------------------------
    The initial setup consists of 3 "fake" points which are placed at a
    large enough distance that they will always form the complete convex
    hull of the tesselation, and 1 "real" point within the experimental
    volume.  Thus the real point possesses a complete tile (with three
    vertices), whereas the fake ones lie in open tiles:  the third vertex
    is at infinity.  All 3 of the initial vertices therefore share the
    real point (Pt[3]) as one of their forming points.
  --------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------
    Initialize forming points for each of 3 initial vertices.
  --------------------------------------------------------------------------*/

    DirV[0].pFormPt[0] = &Pt[3];
    DirV[0].pFormPt[1] = &Pt[1];
    DirV[0].pFormPt[2] = &Pt[2];

    DirV[1].pFormPt[0] = &Pt[3];
    DirV[1].pFormPt[1] = &Pt[2];
    DirV[1].pFormPt[2] = &Pt[0];

    DirV[2].pFormPt[0] = &Pt[3];
    DirV[2].pFormPt[1] = &Pt[0];
    DirV[2].pFormPt[2] = &Pt[1];

/*--------------------------------------------------------------------------
    Initialize neighboring vertices for each of 3 initial vertices.
  --------------------------------------------------------------------------*/

    DirV[0].pNbrDirV[0] = &DirV[1];
    DirV[0].pNbrDirV[1] = &DirV[2];
    DirV[0].pNbrDirV[2] = NULL;

    DirV[1].pNbrDirV[0] = &DirV[2];
    DirV[1].pNbrDirV[1] = &DirV[0];
    DirV[1].pNbrDirV[2] = NULL;

    DirV[2].pNbrDirV[0] = &DirV[0];
    DirV[2].pNbrDirV[1] = &DirV[1];
    DirV[2].pNbrDirV[2] = NULL;

/*--------------------------------------------------------------------------
    Initialize radius of circumcircle and position of its center (which
    is also position of vertex) for each of 3 initial vertices.
  --------------------------------------------------------------------------*/

    circumcircle( &DirV[0] );
    circumcircle( &DirV[1] );
    circumcircle( &DirV[2] );

/*--------------------------------------------------------------------------
    Initialize contiguity counters and heap pointers.
  --------------------------------------------------------------------------*/

    for ( i = 0  ;  i < 4  ;  i++ )
      {
        Pt[i].num_contig = 3;
        Pt[i].pHeap = &Heap[i*4];
        Heap[i*4] = &Pt[i];     /* pointer back to corresponding point, so
                                   heap pointers may be revised as heap gets
                                   shuffled */
      }

/*--------------------------------------------------------------------------
    Initialize heap and adjacent vertices for "fake" points.
  --------------------------------------------------------------------------*/

    Heap[1] = &Pt[1];
    Heap[2] = &Pt[2];
    Heap[3] = &Pt[3];
    Pt[0].pDirV = &DirV[1];     /* an adjacent vertex (not unique) */

    Heap[5] = &Pt[2];
    Heap[6] = &Pt[3];
    Heap[7] = &Pt[0];
    Pt[1].pDirV = &DirV[2];

    Heap[9] = &Pt[3];
    Heap[10] = &Pt[0];
    Heap[11] = &Pt[1];
    Pt[2].pDirV = &DirV[0];

/*--------------------------------------------------------------------------
    Initialize heap and adjacent vertex for first "real" point.
  --------------------------------------------------------------------------*/

    Heap[13] = &Pt[0];
    Heap[14] = &Pt[1];
    Heap[15] = &Pt[2];
    Pt[3].pDirV = &DirV[0];

/*--------------------------------------------------------------------------
    Initialize assorted scalars/pointers.
  --------------------------------------------------------------------------*/

    numDirV = 3;
    heap_avail = HEAP_SIZE - 16;
    pBottomHeap = &Heap[16];

  } /* end function initialize_lists() */





/*****************************/
/*  Function circumcircle()  */
/*****************************/

static void circumcircle( struct DVERTEX *pDirV )
  {
    int      i;
    double   x[3], y[3], m01, m12, m01_minus_m12;



    for ( i = 0  ;  i < 3  ;  i++ )
      {
        x[i] = pDirV->pFormPt[i]->x;
        y[i] = pDirV->pFormPt[i]->y;
      }

    m01 = (x[0] - x[1]) / (y[1] - y[0]);
    m12 = (x[1] - x[2]) / (y[2] - y[1]);

/*--------------------------------------------------------------------------
    If (x[1], y[1]) is one of the "fake" points, m01 and m12 are nearly
    equal and their difference may get rounded to zero.  To avoid a divide-
    by-zero error below, calculate difference "analytically".
  --------------------------------------------------------------------------*/
    m01_minus_m12 = ( x[0]*(y[2]-y[1]) + x[1]*(y[0]-y[2]) + x[2]*(y[1]-y[0]) )
                                 / ( (y[2]-y[1]) * (y[1]-y[0]) );


    pDirV->x = .5*( (y[2]-y[0]) - m12*(x[1]+x[2]) + m01*(x[0]+x[1]) ) /
                                                                m01_minus_m12;
    pDirV->y = .5*( m01*((y[2]-y[0]) - m12*(x[2]-x[0])) / m01_minus_m12  +
                                                                y[0] + y[1] );

    pDirV->rsqr = SQR(pDirV->x - x[0]) + SQR(pDirV->y - y[0]);

  } /* end function circumcircle() */





/***************************/
/*  Function pClosestPt()  */
/***************************/

static struct POINT *pClosestPt( void )
{
    int            i;
    double         dsqr, min_dsqr, cur_dsqr;
    struct POINT   *pMinPt, *pCurPt, **pHeap;


    min_dsqr = SQR(pNewPt->x - Pt[3].x) + SQR(pNewPt->y - Pt[3].y);
    pMinPt = &Pt[3];

    do {
        cur_dsqr = min_dsqr;
        pCurPt = pMinPt;

        min_dsqr = TRULY_HORRENDOUSLY_BIG;
        pMinPt = NULL;

        pHeap = pCurPt->pHeap + 1;   /* first element points back at CurPt */

/*--------------------------------------------------------------------------
        Loop to find contiguous point (MinPt) which is closest to NewPt
        ==> will become new CurPt.
  --------------------------------------------------------------------------*/

        for (i = 0;  i < pCurPt->num_contig;  ++i) {
            dsqr = SQR(pNewPt->x - (*pHeap)->x) + SQR(pNewPt->y - (*pHeap)->y);

            if (dsqr < min_dsqr) {
                min_dsqr = dsqr;
                pMinPt = *pHeap;
            }
#ifdef DIAG
            if (Eflag && (pMinPt == NULL))
                printf("error:  dsqr is pretty darned big\n");
#endif /* DIAG */

            ++pHeap;                     /* point to next contiguous point */

        } /* end for-loop (i:  contiguous points) */

    } while (min_dsqr < cur_dsqr);

/*--------------------------------------------------------------------------
    Now have found point (CurPt) in whose tile NewPt lies:  at least one of
    bounding vertices will be deleted.
  --------------------------------------------------------------------------*/

    return pCurPt;

} /* end function pClosestPt() */





/*****************************/
/*  Function pDeletedDirV()  */
/*****************************/

static struct DVERTEX *pDeletedDirV( struct POINT *pNearestPt )
{
    int              i, j;
    double           dsqr, min_dsqr, cur_dsqr;
    struct DVERTEX   *pCurDirV, *pLastDirV;



    pLastDirV = NULL;
    pCurDirV = pNearestPt->pDirV;

/*..........................................................................*/
#if defined(DIAG)
    if (Eflag)
      {
	printf(" NewPt lies within tile of Pt[%d]\n", (pNearestPt-&Pt[0]));
	printf(" DirV[%d] is vertex on edge of that tile\n", pCurDirV-DirV);
      }
#endif
/*..........................................................................*/

/*  calculate distance from current vertex to new point                     */
    dsqr = SQR(pCurDirV->x - pNewPt->x) + SQR(pCurDirV->y - pNewPt->y);

    while ( dsqr >= pCurDirV->rsqr )
      {
/*      loop through 3 neighboring vertices to current one                  */
        for ( i = 0  ;  i < 3  ;  i++ )
          {
/*          skip vertex we just looked at (LastDirV)                        */
            if ( pCurDirV->pNbrDirV[i] != pLastDirV )
              {
/*              one of remaining neighbors will share NearestPt as a        */
/*              forming point                                               */
                for ( j = 0  ;  j < 3  ;  j++ )
                  if ( pCurDirV->pNbrDirV[i]->pFormPt[j] == pNearestPt )
                    goto FoundNextDirV;
              }

          } /* end for-loop (i) */

        printf("Neighboring vertex not found\n");

FoundNextDirV:
        pLastDirV = pCurDirV;
        pCurDirV = pCurDirV->pNbrDirV[i];

        dsqr = SQR(pCurDirV->x - pNewPt->x) + SQR(pCurDirV->y - pNewPt->y);

      } /* end while */

/*--------------------------------------------------------------------------
    Have found a vertex on NearestPt's tile which is closer to NewPt than
    to NearestPt, and which therefore will be deleted.
  --------------------------------------------------------------------------*/

    return( pCurDirV );

} /* end function pDeletedDirV() */





/********************************/
/*  Function check_neighbors()  */
/********************************/

static void check_neighbors( struct DVERTEX *pCurDelDirV )
{
    int             i;
    double          dsqr;



    for ( i = 0  ;  i < num_DelDirV  ;  i++ )
      {
        if ( pCurDelDirV == pDelDirV[i] ) return;        /* already in list */
      }

    pDelDirV[ num_DelDirV++ ] = pCurDelDirV;            /* add to list      */

/*..........................................................................*/
#if defined(DIAG)
    if ( Eflag  &&  (num_DelDirV > 24) )
      {
        printf("error:  too many vertices-to-be-deleted found.\n");
        return;
      }
#endif
/*..........................................................................*/

    for ( i = 0  ;  i < 3  ;  i++ )
      {
        if ( pCurDelDirV->pNbrDirV[i] != NULL )
          {
/*          calculate distance from neighbor vertex to new point            */
            dsqr = SQR( pCurDelDirV->pNbrDirV[i]->x - pNewPt->x ) +
                   SQR( pCurDelDirV->pNbrDirV[i]->y - pNewPt->y );

            if ( dsqr < pCurDelDirV->pNbrDirV[i]->rsqr )
              check_neighbors( pCurDelDirV->pNbrDirV[i] );
          }

      } /* end for-loop (i) */

} /* end function check_neighbors() */





/***********************************/
/*  Function find_contig_points()  */
/***********************************/

static void find_contig_points( void )
{
    int             i, j, k;



    num_ContigPt = 0;

/*  loop through the vertices deleted by NewPt                             */
    for ( i = 0  ;  i < num_DelDirV  ;  i++ )
      {
/*      loop through their forming points                                  */
        for ( j = 0  ;  j < 3  ;  j++ )
          {
            for ( k = 0  ;  k < num_ContigPt  ;  k++ )
              if ( pContigPt[k] == pDelDirV[i]->pFormPt[j] )
                goto NextFormPt;

            pContigPt[ num_ContigPt++ ] = pDelDirV[i]->pFormPt[j];

NextFormPt: ;

          } /* end for-loop (j) */

      } /* end for-loop (i) */

/*  copy first contiguous point to end of list, too ("cyclic", sort of)    */
    pContigPt[ num_ContigPt ] = pContigPt[0];

} /* end function find_contig_points() */





/**************************************/
/*  Function mark_old_contiguities()  */
/**************************************/

static void mark_old_contiguities( void )
{
    int             i, j, k, m, num_SharFormPt;
    struct POINT    *pTmpPt[3], **pHeap;



/*  can't remove any contiguities with only one deleted vertex              */
    if ( num_DelDirV > 1 )
      {
/*      loop through deleted vertices                                       */
        for ( i = 0  ;  i < num_DelDirV-1  ;  i++ )
          {
/*          loop through all the *other* deleted vertices                   */
            for ( j = i+1  ;  j < num_DelDirV  ;  j++ )
              {
                num_SharFormPt = 0;

/*              loop through vertex (i) forming points...                   */
                for ( k = 0  ;  k < 3  ;  k++ )
/*                ...and compare to forming points of vertex (j)            */
                  for ( m = 0  ;  m < 3  ;  m++ )
                    if ( pDelDirV[i]->pFormPt[k] ==
                                                   pDelDirV[j]->pFormPt[m] )
                      pTmpPt[ num_SharFormPt++ ] = pDelDirV[i]->pFormPt[k];

/*              deleted vertices i, j share a pair of forming points ==>    */
/*              those two points no longer contiguous                       */
                if ( num_SharFormPt == 2 )
                  {
/*..........................................................................*/
#if defined(DIAG)
                    if (Eflag)
                      printf("mark_old_contig:  zeroing old contiguity\n");
#endif
/*..........................................................................*/

                /*----------------------------------------------------------
                    Clear second point from first's contiguity list.
                  ----------------------------------------------------------*/

                    k = 0;
                    pHeap = pTmpPt[0]->pHeap + 1;    /* skip return pointer */

                    while ( (*pHeap != pTmpPt[1]) &&
                            (k++ < pTmpPt[0]->num_contig) )  ++pHeap;

                    if ( k > pTmpPt[0]->num_contig )
                      printf("heap corrupt:  can't find contiguous point.\n");
                    else
                      *pHeap = NULL;

                /*----------------------------------------------------------
                    Clear first point from second one's contiguity list.
                  ----------------------------------------------------------*/

                    k = 0;
                    pHeap = pTmpPt[1]->pHeap + 1;    /* skip return pointer */

                    while ( (*pHeap != pTmpPt[0]) &&
                            (k++ < pTmpPt[1]->num_contig) )  ++pHeap;

                    if ( k > pTmpPt[1]->num_contig )
                      printf("heap corrupt:  can't find contiguous point.\n");
                    else
                      *pHeap = NULL;

                  } /* end if */

              } /* end for-loop (j) */

          } /* end for-loop (i) */

      } /* end if (more than one deleted vertex) */

} /* end function mark_old_contiguities() */




/*********************************/
/*  Function add_new_vertices()  */
/*********************************/

static void add_new_vertices( void )
{
    int             i, j, k, m;
    int             num_Penum, num_NULLs, deleted, found;
    int             num_NewDirV, test_count, contig, diff;
    int             num_NULL_nbrs, num_shared_points;
    struct POINT    **pHeap, *pPtA, *pPtB;



/*--------------------------------------------------------------------------
    First find "penumbral" vertices, i.e. those existing vertices adjacent
    to the deleted set, but not themselves deleted (deleted ones = "umbra").
    Number of penumbral vertices is less than or equal to number of contigu-
    ous points.
  --------------------------------------------------------------------------*/

#ifdef DIAG
    if (Eflag) {
        for (i = 0;  i < num_ContigPt;  ++i)
            printf("     pContigPt[%d] -> Pt[%d]\n", i, (pContigPt[i]-&Pt[0]));
        for (i = 0;  i < num_DelDirV;  ++i)
            printf("     pDelDirV[%d] -> DirV[%d]\n",i, (pDelDirV[i]-&DirV[0]));
    }
#endif

    num_Penum = 0;
    num_NULLs = 0;

/*  loop through deleted vertices...                                        */
    for ( i = 0  ;  i < num_DelDirV  ;  i++ )
      {
/*      loop through each of their neighboring vertices...                  */
        for ( j = 0  ;  j < 3  ;  j++ )
          {
            deleted = FALSE;

/*          and check if neighbor is in deleted list, too                   */
            for ( k = 0  ;  !deleted && k < num_DelDirV  ;  k++ )
              if ( pDelDirV[i]->pNbrDirV[j] == pDelDirV[k] )
                deleted = TRUE;

            /*--------------------------------------------------------------*/
            if ( !deleted )        /* (always, for a NULL penumbral vertex) */
              {
                found = FALSE;

            /*  not deleted:  must be penumbral vertex, so check if already */
            /*  in list and add if not (perhaps not really necessary...);   */
            /*  if vertex is NULL, may get "false find" (can have up to 2   */
            /*  NULL vertices in penumbral set for Pt[4+], and 3 for Pt[3]) */
                for ( k = 0  ;  !found && k < num_Penum  ;  k++ )
                  if ( pDelDirV[i]->pNbrDirV[j] == pPenumDirV[k] )
                    found = TRUE;

                /*----------------------------------------------------------*/
                if ( !found )
                  {
                /*  not found:  add to list                                 */
                    pPenumDirV[ num_Penum ] = pDelDirV[i]->pNbrDirV[j];

                /*  zero any neighbors of penumbral vertex which are in     */
                /*  deleted list, unless penumbral vertex itself is NULL    */
                    if ( pPenumDirV[num_Penum] )
                      {
                        for ( k = 0  ;  k < 3  ;  k++ )
                          if ( pPenumDirV[num_Penum]->pNbrDirV[k] ==
                                                                 pDelDirV[i] )
                            pPenumDirV[num_Penum]->pNbrDirV[k] = NULL;
                      }
                    else
                      ++num_NULLs;

                    ++num_Penum;
                  }
                else  /* found */
                  {
                /*  was found, i.e. penumbral vertex already in list, but   */
                /*  need to zero references to *another* deleted vertex...  */
                    if ( pDelDirV[i]->pNbrDirV[j] )
                      {
                        for ( k = 0  ;  k < 3  ;  k++ )
                          if ( pDelDirV[i]->pNbrDirV[j]->pNbrDirV[k] ==
                                                                  pDelDirV[i] )
                            pDelDirV[i]->pNbrDirV[j]->pNbrDirV[k] = NULL;
                      }
                /*  ...unless the "found" vertex was another (necessarily   */
                /*  distinct) NULL penumbral vertex, in which case should   */
                /*  at update counters and list of penumbral vertices:      */
                    else  /* deleted vertex's neighbor NULL */
                      {
                        ++num_NULLs;
                        pPenumDirV[ num_Penum++ ] = NULL;
                      }

                  } /* end if (penumbral vertex not already in penum. list) */
                /*----------------------------------------------------------*/


              } /* end if (vertex not in deleted list:  penumbral) */
            /*--------------------------------------------------------------*/

          } /* end for-loop (j:  neighbor vertices) */

      } /* end for-loop (i:  deleted vertices) */


#ifdef DIAG /*...............................................................*/
    if (Eflag) {
        if (num_Penum != num_ContigPt)
            printf("note:  number of penumbral "
              "vertices (%d) != number of contiguous points (%d)\n", num_Penum,
              num_ContigPt);
        for (i = 0;  i < num_Penum;  ++i)
            if (pPenumDirV[i])
                printf("       pPenumDirV[%d] -> DirV[%d]\n", i,
                  (pPenumDirV[i]-DirV));
            else
                printf("       pPenumDirV[%d] ->  NULL\n", i);
        printf("       num_NULLs = %d\n", num_NULLs);
    }
#endif /* DIAG ..............................................................*/


/*--------------------------------------------------------------------------
    Now define the new vertices by their forming points:  NewPt plus adja-
    cent pairs of the cyclically-ordered points contiguous to NewPt.  The
    NewDirV list is now a separate entity, no longer part of pDelDirV[].
    (Was overwriting pDelDirV before had finished using it).
  --------------------------------------------------------------------------*/

    for ( i = 0  ;  i < num_ContigPt  ;  i++ )
        pNewDirV[i] = &DirV[ numDirV++ ];

    if ( numDirV >= MAX_DirV )
        printf("error:  vertex overflow in array DirV[]:  Pt[%d]\n", Pindex);

/*--------------------------------------------------------------------------
    take care of single vertex case separately so can flag errors below
  --------------------------------------------------------------------------*/
    if ( num_DelDirV == 1 )
      {
#ifdef DIAG
        if (Eflag && (num_ContigPt != 3))
            printf("error:  1 deleted vertex but not 3 contiguous points\n");
#endif /* DIAG */

/*      will be three new vertices; note that pContigPt[num_ContigPt] was   */
/*      set equal to pContigPt[0] to avoid necessity of a mod operation (%) */
        for ( i = 0  ;  i < num_ContigPt  ;  i++ )
          {
            pNewDirV[i]->pFormPt[0] = pContigPt[ i ];
            pNewDirV[i]->pFormPt[1] = pContigPt[ i+1 ];
            pNewDirV[i]->pFormPt[2] = pNewPt;

            pContigPt[i]->pDirV = pNewDirV[i];

            circumcircle( pNewDirV[i] );
          }
      }
/*--------------------------------------------------------------------------
    have multiple deleted vertices
  --------------------------------------------------------------------------*/
    else
      {
        num_NewDirV = 0;

/*      loop through deleted vertices, using contiguity of their forming    */
/*      points as guide to creating new vertices                            */
        for ( i = 0  ;  i < num_DelDirV  ;  i++ )
          {
            test_count = 0;
#ifdef DIAG
            if (Eflag)
                printf("  looking at deleted vertex DirV[%d]\n",
                  (pDelDirV[i]-&DirV[0]));
#endif /* DIAG */

/*          loop through three possible contiguous pairs of forming points  */
            for ( j = 0  ;  j < 3  ;  j++ )
              {
                pPtA = pDelDirV[i]->pFormPt[  j      ];
                pPtB = pDelDirV[i]->pFormPt[ (j+1)%3 ];
                contig = FALSE;

/*              loop through point B's contiguity list looking for A        */
                pHeap = pPtB->pHeap + 1;
                for ( k = 0  ;  !contig && k < pPtB->num_contig  ;  k++ )
                  if ( *pHeap++ == pPtA )
                    contig = TRUE;

                if ( contig )
                  {
                    pNewDirV[num_NewDirV]->pFormPt[0] = pPtA;
                    pNewDirV[num_NewDirV]->pFormPt[1] = pPtB;
                    pNewDirV[num_NewDirV]->pFormPt[2] = pNewPt;

/*                  resetting edge-vertex pointer of both A and B is        */
/*                  overkill (redundant) but ensures existent vertex        */
                    pPtA->pDirV = pNewDirV[ num_NewDirV ];
                    pPtB->pDirV = pNewDirV[ num_NewDirV ];

                    circumcircle( pNewDirV[num_NewDirV] );

                    ++num_NewDirV;
                    ++test_count;
                  }
              } /* end for-loop (j:  forming points) */

            if ( test_count > 2 )  printf(
              "error:  too many new vertices from one deleted one (%d)\n",
              test_count);

          } /* end for-loop (i:  deleted vertices) */

        if ( num_NewDirV != num_ContigPt )
          printf("error:  wrong number of new vertices (%d)\n", num_NewDirV);

      } /* end if (multiple deleted vertices) */


/*--------------------------------------------------------------------------
    No longer need deleted vertex information, so copy as many new vertices
    as can over the deleted ones and free up space at end of DirV[] array.
    Last new vertex (num_ContigPt-1) goes over first deleted one, and so on.
  --------------------------------------------------------------------------*/

    for ( i = 0, j = num_ContigPt-1  ;  i < num_DelDirV  ;  i++, j-- )
      {
        pDelDirV[i]->x    = pNewDirV[j]->x   ;
        pDelDirV[i]->y    = pNewDirV[j]->y   ;
        pDelDirV[i]->rsqr = pNewDirV[j]->rsqr;

        for ( k = 0  ;  k < 3  ;  k++ )
          pDelDirV[i]->pFormPt[k] = pNewDirV[j]->pFormPt[k];

        pNewDirV[j] = pDelDirV[i];       /* now pNewDirV points to its data */
      }

/*--------------------------------------------------------------------------
    See if any of the contiguous points' edge-vertex pointers point at one
    of the moved vertices; if so, reset them appropriately (numDirV indi-
    cates new end-of-list).
  --------------------------------------------------------------------------*/

    numDirV -= num_DelDirV;

    for ( i = 0  ;  i < num_ContigPt  ;  i++ )
      if ( (diff = pContigPt[i]->pDirV - &DirV[numDirV]) >= 0 )
        pContigPt[i]->pDirV = pDelDirV[ num_DelDirV-diff-1 ];


/*..........................................................................*/
#if defined(DIAG)
    if (Eflag)
      for ( i = 0  ;  i < num_ContigPt  ;  i++ )
        {
          printf("     pNewDirV[%d] -> DirV[%d]\n", i, (pNewDirV[i]-&DirV[0]));
          for ( j = 0  ;  j < 3  ;  j++ )
            printf("        .pFormPt[%d] -> Pt[%d]\n", j,
                                           (pNewDirV[i]->pFormPt[j] - &Pt[0]));
        }
#endif
/*..........................................................................*/

/*--------------------------------------------------------------------------
    For each new vertex, find its "hard" neighboring vertex:  one of the
    penumbral vertices (unique).  First two forming points of each new
    vertex are also forming points for the desired penumbral vertex.
  --------------------------------------------------------------------------*/

    num_NULL_nbrs = 0;

/*  loop through new vertices                                               */
    for ( i = 0  ;  i < num_ContigPt  ;  i++ )
      {
        found = FALSE;

/*      loop through penumbral vertices                                     */
        for ( j = 0  ;  !found && j < num_Penum  ;  j++ )
          {
/*          exclude NULL penumbral vertices                                 */
            if ( pPenumDirV[j] )
              {
                num_shared_points = 0;

/*              loop through forming points of current penumbral vertex     */
                for ( k = 0  ;  k < 3  ;  k++ )
                  if ( (pPenumDirV[j]->pFormPt[k] == pNewDirV[i]->pFormPt[0])
                    || (pPenumDirV[j]->pFormPt[k] == pNewDirV[i]->pFormPt[1]) )
                    ++num_shared_points;

            /*--------------------------------------------------------------
                new vertex is neighbor to a real, non-NULL penumbral vertex
              --------------------------------------------------------------*/
                if ( num_shared_points == 2 )
                  {
                    found = TRUE;

/*                  connect new vertex to penumbral one...                  */
                    pNewDirV[i]->pNbrDirV[2] = pPenumDirV[j];

/*                  ...and penumbral vertex back to new one                 */
                    for ( k = 0  ;  k < 3 && pPenumDirV[j]->pNbrDirV[k] != NULL
                                                                     ;  k++ );
                    if ( k > 2 ) printf(
                      "error:  can't find NULL neighbor for penum. vertex\n");

                    pPenumDirV[j]->pNbrDirV[k] = pNewDirV[i];

/*..........................................................................*/
#if defined(DIAG)
                    if (Eflag)
                      printf("     DirV[%d] not neighbor to NULL vertex\n",
                                                   (pNewDirV[i] - &DirV[0]) );
#endif
/*..........................................................................*/
                  }

              } /* end if (non-NULL penumbral vertex) */

          } /* end for-loop (j:  penumbral vertices) */


    /*----------------------------------------------------------------------
        new vertex must be neighbor to NULL vertex (up to two are possible)
      ----------------------------------------------------------------------*/
        if ( !found )
          {
            pNewDirV[i]->pNbrDirV[2] = NULL;
            ++num_NULL_nbrs;

/*..........................................................................*/
#if defined(DIAG)
            if (Eflag)
              printf("     DirV[%d] IS neighbor to NULL vertex\n",
                                                   (pNewDirV[i] - &DirV[0]) );
#endif
/*..........................................................................*/
          }

      } /* end for-loop (i:  new vertices) */


    if ( num_NULL_nbrs != num_NULLs ) printf(
      "error:  number NULL-neighbors (%d) != number NULL penum verts (%d)\n",
                                                    num_NULL_nbrs, num_NULLs);

/*--------------------------------------------------------------------------
    Loop through new vertices, adding two "easy" neighbors (adjacent new
    vertices) to each of them.  Unfortunately, since the contiguous points
    and hence the new vertices are no longer ordered cyclically, the "easy"
    neighbors are no longer all that easy.
  --------------------------------------------------------------------------*/

/*  loop over all of new vertices                                           */
    for ( i = 0  ;  i < num_ContigPt-1  ;  i++ )
      {
/*      loop over their first two forming points                            */
        for ( j = 0  ;  j < 2  ;  j++ )
          {
/*          loop over remaining new vertices (farther down in list)         */
            for ( k = i+1  ;  k < num_ContigPt  ;  k++ )
/*            loop over first two forming points of remaining vertices      */
              for ( m = 0  ;  m < 2  ;  m++ )
                if ( pNewDirV[i]->pFormPt[j] == pNewDirV[k]->pFormPt[m] )
                  {
                    pNewDirV[i]->pNbrDirV[j] = pNewDirV[k];
                    pNewDirV[k]->pNbrDirV[m] = pNewDirV[i];
                    goto Next_j;
                  }

Next_j:     ;                         /* as soon as have a match, come here */

          } /* end for-loop (j:  forming points) */

      } /* end for-loop (i:  new vertices) */


/*..........................................................................*/
#if defined(DIAG)
    if (Eflag)
      for ( i = 0  ;  i < num_ContigPt  ;  i++ )
        for ( j = 0  ;  j < 3  ;  j++ )
          if ( pNewDirV[i]->pNbrDirV[j] )
            printf("      DirV[%d].pNbrDirV[%d] -> DirV[%d]\n",
                  (pNewDirV[i] - DirV), j, (pNewDirV[i]->pNbrDirV[j] - DirV));
          else
            printf("      DirV[%d].pNbrDirV[%d] ->  NULL\n",
                                                     (pNewDirV[i] - DirV), j);
#endif
/*..........................................................................*/

} /* end function add_new_vertices() */





/******************************/
/*  Function modify_points()  */
/******************************/

static void modify_points( void )
{
    int             i, j, updated, num_NULLs;
    struct POINT    **pHeap, **pNewHeap;



/*--------------------------------------------------------------------------
    First complete data for NewPt:  num_contig, pHeap, pDirV, and actual
    heap data (contiguity list).
  --------------------------------------------------------------------------*/

    pNewPt->num_contig = num_ContigPt;             /* ~ size of heap alloc. */

    pHeap = request_heap_space( num_ContigPt+1 );
    if ( pHeap )
      pNewPt->pHeap = pHeap;                       /* set pointer to heap   */
    else
      {
        printf("error:  heap overflow (NewPt:  Pt[%d]).\n", (pNewPt-&Pt[0]));
        exit(2);
      }

    *pHeap++ = pNewPt;                             /* set return pointer    */
    for ( i = 0  ;  i < num_ContigPt  ;  i++ )
      *pHeap++ = pContigPt[i];                     /* set contiguous points */

    pNewPt->pDirV = pNewDirV[0];                   /* could use ANY new one */

/*--------------------------------------------------------------------------
    Now update contiguous points to reflect contiguity to NewPt.
  --------------------------------------------------------------------------*/

/*  loop through contiguous points                                          */
    for ( i = 0  ;  i < num_ContigPt  ;  i++ )
      {
        updated = FALSE;
        num_NULLs = 0;
        pHeap = pContigPt[i]->pHeap + 1;

/*      loop through contiguity list of current contiguous point to see how */
/*      many points were deleted (NULLs); if find one, replace with NewPt   */
        for ( j = 0  ;  j < pContigPt[i]->num_contig  ;  j++, pHeap++ )
          if ( *pHeap == NULL )
            {
              ++num_NULLs;
              if ( !updated )
                {
                  *pHeap = pNewPt;
                  updated = TRUE;
                }
            }


    /*----------------------------------------------------------------------
        No NULLs found, i.e. didn't lose any contiguities when added NewPt
      ----------------------------------------------------------------------*/
        if ( num_NULLs == 0 )
          {
/*          check if there is a hole *after* the contiguity list            */
            if ( *pHeap == NULL )
              {
                *pHeap = pNewPt;                   /* fill hole with NewPt  */
                ++pContigPt[i]->num_contig;        /* ...and update counter */
              }

/*          no hole:  copy whole list to end of heap and zero old entries   */
            else
              {
                pNewHeap = request_heap_space( pContigPt[i]->num_contig + 2 );

/*              following stmt MUST be executed AFTER request_heap_space    */
/*              called, else may point to wrong area of heap (if crunched)  */
                pHeap = pContigPt[i]->pHeap;

                if ( pNewHeap )
                  pContigPt[i]->pHeap = pNewHeap;
                else
                  {
                    printf("error:  heap overflow (contig points of Pt[%d])\n",
                                                              (pNewPt-&Pt[0]));
                    exit(2);
                  }

                for ( j = 0  ;  j <= pContigPt[i]->num_contig  ;  j++ )
                  {
                    *pNewHeap++ = *pHeap;          /* copy old contiguities */
                    *pHeap++ = NULL;               /* ...and zero them      */
                  }
                *pNewHeap = pNewPt;                /* add new contiguity... */
                ++pContigPt[i]->num_contig;        /* ...and update counter */

                max_ContigPt = MAX(pContigPt[i]->num_contig, max_ContigPt);
              }

          } /* end if (num_NULLs == 0) */

    /*----------------------------------------------------------------------
        2 or more NULLs found, so will have to compress contiguity list
      ----------------------------------------------------------------------*/
        else if ( num_NULLs > 1 )
          {
            --num_NULLs;                           /* filled one with NewPt */

            pHeap = pContigPt[i]->pHeap + 1;
            pNewHeap = pHeap;

            for ( j = 0  ;  j < pContigPt[i]->num_contig  ;  j++, pHeap++ )
              if ( *pHeap )
                *pNewHeap++ = *pHeap;              /* copy all but NULLs    */

            for ( j = 0  ;  j < num_NULLs  ;  j++ )
              *pNewHeap++ = NULL;                  /* NULLs at end of list  */

            pContigPt[i]->num_contig -= num_NULLs; /* update counter approp */
          }

      } /* end for-loop (i:  contiguous points) */

} /* end function modify_points() */





/***********************************/
/*  Function request_heap_space()  */
/***********************************/

static struct POINT **request_heap_space( int amount_requested )
{
    int             i, num_contig;
    struct POINT    **pHeap, **pNewHeap;
/*..........................................................................*/
#if defined(DIAG)
    int             j;
    struct POINT    **pTmpHeap;
#endif
/*..........................................................................*/



/*--------------------------------------------------------------------------
    First check if too little heap space free, and compact if necessary.
  --------------------------------------------------------------------------*/

    if ( amount_requested >= heap_avail )
      {
/*..........................................................................*/
#if defined(DIAG)
        if (Eflag)
          {
        printf("-------------- Heap dump ---------------\n");
        printf("------- (just before crunching) --------\n");
        for ( j = 0  ;  j <= Pindex  ;  j++ )
          if ( Pt[j].pHeap )
            printf("  Pt[%d].pHeap -> Heap[%d]\n", j, (Pt[j].pHeap-&Heap[0]));
          else
            printf("  Pt[%d].pHeap ->    0\n", j);
        pTmpHeap = &Heap[0];
        j = 0;
        while ( pTmpHeap < pBottomHeap )
          {
            if ( *pTmpHeap )
              printf("   Heap[%d] -> Pt[%d]\n", j, (*pTmpHeap-&Pt[0]));
            else
              printf("   Heap[%d] ->   0\n", j);
            ++pTmpHeap;
            ++j;
          }
        printf("------------ End Heap Dump -------------\n");
          } /* end if (Eflag) */
#endif
/*..........................................................................*/


        ++num_compactions;          /* bookkeeping:  keep track of how many */

        pHeap = &Heap[0];
        pNewHeap = pHeap;

        while ( *pHeap == NULL )  ++pHeap;

/*      main compaction loop:  pHeap points to a return pointer at start of  */
/*      each loop iteration                                                  */
        while ( pHeap < pBottomHeap )
          {
            num_contig = (*pHeap)->num_contig;
            (*pHeap)->pHeap = pNewHeap;

            for ( i = 0  ;  i <= num_contig  ;  i++ )
              *pNewHeap++ = *pHeap++ ;

/*          skip NULLs to start of next contiguity list                     */
            while ( (*pHeap == NULL) && (pHeap < pBottomHeap) )  ++pHeap;
          }

        heap_avail += (pBottomHeap - pNewHeap);
        pBottomHeap = pNewHeap;

/*..........................................................................*/
#if defined(DIAG)
        if (Eflag)
          {
        printf("     * HEAP CRUNCHED *   (%d)\n", num_compactions);

        printf("-------------- Heap dump ---------------\n");
        printf("------- (just after crunching) ---------\n");
        for ( j = 0  ;  j <= Pindex  ;  j++ )
          if ( Pt[j].pHeap )
            printf("  Pt[%d].pHeap -> Heap[%d]\n", j, (Pt[j].pHeap-&Heap[0]));
          else
            printf("  Pt[%d].pHeap ->    0\n", j);
        pTmpHeap = &Heap[0];
        j = 0;
        while ( pTmpHeap < pBottomHeap )
          {
            if ( *pTmpHeap )
              printf("   Heap[%d] -> Pt[%d]\n", j, (*pTmpHeap-&Pt[0]));
            else
              printf("   Heap[%d] ->   0\n", j);
            ++pTmpHeap;
            ++j;
          }
        printf("------------ End Heap Dump -------------\n");
          } /* end if (Eflag) */
#endif
/*..........................................................................*/


      } /* end if (not enough heap space) */

/*--------------------------------------------------------------------------
    Now check again; if enough room, allocate requested space, else return
    NULL pointer to indicate heap overflow.
  --------------------------------------------------------------------------*/

    if ( amount_requested < heap_avail )
      {
        pHeap = pBottomHeap;
        pBottomHeap += amount_requested;
        heap_avail  -= amount_requested;
      }
    else
        pHeap = NULL;    /* even compacted, not enough heap space: overflow */

    return( pHeap );

} /* end function request_heap_space() */
