/*---------------------------------------------------------------------------

  tplot.c

  Plot the initial points load, tesselation, and/or triangulation (from 
  d###.c); or one or more views of an evolved, triangulated disk (from 
  either d###.c or f###.c); or an evolved, ringed disk (from rings.c).
  The program uses DISSPLA, and the output is a PostScript file.

  ---------------------------------------------------------------------------

  Allowed plot types:

     type 0:  points plot [d###tes#### -> d###pts####.ps]
     type 1:  tesselation plot [d###tes#### -> d###tes####.ps]
     type 2:  (unwarped) triangulation plot [d###tri#### -> d###tri####.ps]
     type 3:  three-view evolved triangulation plot
                [d###f###v###.# -> t##########.ps]
     type 4:  custom-view evolved triangulation plot (default)
                [d###f###v###.# -> v##########.ps]
     type 5:  custom-view evolved rings plot
                [d###f###r###.# -> r##########.ps]
     type 6:  ring azimuths plot
                [d###f###.karr -> A##########.ps]
     type 7:  ring inclinations plot
                [d###f###.karr -> I##########.ps]
     type 8:  ring radii plot
                [d###f###.karr -> R##########.ps]

  ---------------------------------------------------------------------------

  To compile:    scc -c tplot.c
                 segldr tplot.o -l$disspla -s -o tplot

  ---------------------------------------------------------------------------

     Version:  Y-MP/DISSPLA/PostScript               Greg Roelofs
     Source:   tplot.c on eagle                      28 February 1992

  ---------------------------------------------------------------------------

     1.0:  original program (based on render.c v1.0 and tplot.for v3.0)
     1.1:  added various ring-parameters vs. time plots

  ---------------------------------------------------------------------------*/



/***************************/
/* Includes, Defines, etc. */
/***************************/

#include "fd.h"          /* all sorts of great stuff */
#undef DOT               /* macro name conflicts with DISSPLA routine */
#include "disspla.h"     /* DISSPLA prototypes, fortran.h */
#include <ctype.h>       /* isdigit() */
#include <string.h>      /* strlen() */
#include <time.h>        /* time(), localtime(), strftime() */
#include <sys/types.h>   /* rename() */
#include <sys/stat.h>    /* rename() */

#define VERSION   "1.19"
#define OWNER     "Greg Roelofs, U of C AAC 135"

#define GET_FLOAT(x)   {c1=*(argv[0]+1); if(isdigit(c1)||(c1=='-')||(c1=='.')) \
    {x=atof(argv[0]+1); *(argv[0]+1)='\0';} else {fprintf(stderr, \
    "tplot:  option [-%c] requires an angle\n",c); ++error;} }
/*
 * macro GET_FLOAT( double x )
 * {
 *     c1 = *(argv[0] + 1);            [get next char w/o changing ptr]
 *     if (isdigit(c1) || (c1 == '-') || (c1 == '.')) {
 *         theta = atof(argv[0]+1);    [ptr to digit string]
 *         *(argv[0]+1) = '\0';
 *     } else {
 *         fprintf(stderr, 
 *           "tplot:  option [-%c] requires an angle\n", c);
 *         ++error;
 *     }
 * }
 */

typedef double   six_vector[6];

static void   usage           ( void );
static void   init_hardware   ( void );
static void   init_page       ( void );
static void   PlotInitTess    ( void );
static void   PlotWarpTri3    ( void );
static void   PlotWarpTri     ( void );
static void   PlotRings       ( void );
static void   PlotRingParams  ( void );




/******************************/
/* Global and Local Variables */
/******************************/

void           (*make_plot[])()={PlotInitTess, PlotInitTess, PlotInitTess,
                   PlotWarpTri3, PlotWarpTri, PlotRings, PlotRingParams,
                   PlotRingParams, PlotRingParams};
char           datafn[80], plotfn[80], title[80], xtitle[80], ytitle[80];
int            custom_axes=FALSE, print_angles=TRUE, check_range=FALSE;
int            first_ring=0, last_ring=9999;
int            karr[256], type=4, numDirE, num_edges, numsteps;
double         rarr[256], def_axmax=15., theta=75., phi=-60.;
double         axmin, axinc, axmax, xmin, xinc, xmax, ymin, yinc, ymax;
double         *Pt_x, *Pt_y, *time_axis, *ampl;
four_vector    *edg4;
six_vector     *edg6;
FILE           *infile;
struct stat    statbuf;

/* numerical constants for DISSPLA calls (FORTRAN:  by address) */
const int      im1=-1, im3=-3, im102=-102, im103=-103, im104=-104, im111=-111;
const int      i0=0, i1=1, i2=2, i3=3, i4=4, i5=5, i6=6, i7=7, i8=8, i9=9;
const int      i100=100, i500000=500000;
const double   d0=0., dp001=.001, dp003=.003, dp006=.006, dp01=.01, dp05=.05;
const double   dp1=.1, dp2=.2, dp25=.25, dp7=.7, d1=1., d1p5=1.5, d1p75=1.75;
const double   d2=2., d3=3., d4=4., d4p5=4.5, d5p5=5.5, d6p332=6.332;
const double   d6p366=6.366, d8=8., d8p45=8.45, d8p5=8.5, d8p54=8.54;
const double   d8p63=8.63, d8p72=8.72, d8p81=8.81, d9=9., d10=10., d11=11.;
const double   d15=15., d30=30., d45=45., d90=90., d100=100.;





/********************/
/* Function usage() */
/********************/

static void usage( void )
{
    fprintf(stderr,
"   Usage:  tplot [-#] [-i#.#] [-x#.#] [-t#.#] [-p#.#] [-r#-#] [-o] file\n"
"     -#      plot type:\n"
"               0  points load\n"
"               1  initial tesselation\n"
"               2  initial triangulation\n"
"               3  evolved triangulation, 3 orthogonal views\n"
"               4  evolved triangulation, custom view (default)\n"
"               5  evolved rings, custom view [not yet implemented]\n"
"               6  ring azimuths\n"
"               7  ring inclinations\n"
"               8  ring radii\n"
"     -i#.#   axis label increment\n"
"     -x#.#   maximum axis value\n"
"     -t#.#   polar viewing angle (theta)\n"
"     -p#.#   azimuthal viewing angle (phi)\n"
"     -r#-#   range of rings to plot\n"
"     -o      print \"out-of-range\" messages\n"
"     file    data file from which to read\n");

} /* end function usage() */





/****************/
/* Main program */
/****************/

main( int argc, char *argv[] )
  {
    static char   *abbrev[]={"pts", "tes", "tri"};  /* first 3 plot types */
    static char   abbr[]="tvrAIR";                  /* remaining plot types */
    static char   *descrip[]={
                      "\nPlot type 0:  initial points load\n",
                      "\nPlot type 1:  initial tesselation\n",
                      "\nPlot type 2:  initial (unwarped) triangulation\n",
                      "\nPlot type 3:  three-view evolved triangulation\n",
                      "\nPlot type 4:  custom-view evolved triangulation\n",
                      "\nPlot type 5:  custom-view evolved rings\n",
                      "\nPlot type 6:  ring azimuths\n",
                      "\nPlot type 7:  ring inclinations\n",
                      "\nPlot type 8:  ring radii\n"
                  };
    char     c, c1, *p, *pf=datafn;
/*    int      error=0, num_types=sizeof(make_plot)/sizeof((*)()); */
    int      error=0, num_types=sizeof(make_plot)/sizeof(&PlotInitTess);
    int      user_angles=FALSE, user_axinc=FALSE, user_axmax=FALSE;
    double   t1, dt;


/*---------------------------------------------------------------------------
    Parse command line for '-' options.
  ---------------------------------------------------------------------------*/

    t1 = SECOND();

    if (argc <= 1)
      --argc;
    else
      while ((--argc > 0) && ((*++argv)[0] == '-'))
        while (c = *++(argv[0]))   /* argv was incr => argv[0] not prognam */
          if (isdigit(c)) {
              type = atoi(argv[0]);
              *(argv[0]+1) = '\0';   /* overwrite digit, terminate option */
              if (type >= num_types) {
                  fprintf(stderr, "tplot:  invalid plot type [-%d]\n", type);
                    ++error;
              }
          } else
              switch (c) {
                case 'i':      /* axis label increment */
                    user_axinc = TRUE;
                    GET_FLOAT(axinc)
                    break;

                case 'o':      /* don't suppress out-of-range messages (r,s?) */
                    check_range = TRUE;
                    break;

                case 'p':      /* azimuthal viewing angle (phi) */
                    user_angles |= 2;
                    GET_FLOAT(phi)
                    break;

                case 'r':      /* rings to include */
                    if (p = strchr(argv[0]+1, '-')) {
                        *p = 0;
                        first_ring = atoi(argv[0]+1);
                        if ((last_ring = atoi(p+1)) == 0)
                            last_ring = 9999;
                    } else
                        first_ring = last_ring = atoi(argv[0]+1);
                    *(argv[0]+1) = 0;   /* terminate option */
                    break;

                case 't':      /* polar viewing angle (theta) */
                    user_angles |= 1;
                    GET_FLOAT(theta)
		    break;

                case 'x':      /* maximum axis value */
                    user_axmax = TRUE;
                    GET_FLOAT(axmax)
                    break;

                default:
                    fprintf(stderr, "tplot:  invalid option [-%c]\n", c);
                    ++error;
                    *(argv[0]+1) = '\0';   /* terminate this arg; try next */
                    break;
              }  /* end switch, if (isdigit), while, while */

/*---------------------------------------------------------------------------
    If any invalid options were encountered, print syntax synopsis.  Any 
    argument remaining (i.e., not preceded by '-') must be the name of the 
    data file to be opened.  Otherwise, if no arguments left, print synopsis
    and exit.
  ---------------------------------------------------------------------------*/

    if (argc > 0)
        while (*pf++ = *argv[0]++);   /* copy argument into datafn[] */
    else
        ++error;   /* no filename:  oops (or maybe just want usage) */

    if (error) {
        usage();
        exit(1);   /* ho ho ho, gotta go */
    }

/*---------------------------------------------------------------------------
    Set default plot values, unless user specified one or more, and print
    quick summary of what we are about to do.
  ---------------------------------------------------------------------------*/

    if (user_angles)
        if ((type != 4) && (type != 5))
            fprintf(stderr, "note:  viewing angles ignored for plot type %d\n",
              type);

    if (!user_axmax)
        axmax = def_axmax;

    axmin = -axmax;

    if (!user_axinc)
        axinc = axmax / 5.;

    printf(descrip[type]);

/*---------------------------------------------------------------------------
    Read data and set up plotfile name.  Then initialize DISSPLA stuff, make
    the plot, and free memory malloc'd in read_data().
  ---------------------------------------------------------------------------*/

    read_data();
    if (type < 3)     /* d###{pts,tes,tri}####.ps */
        sprintf(plotfn, "d%03d%s%04d.ps", d_version? d_version:p_version,
          abbrev[type], numV);
    else              /* {t,v,r,A,I,R}##########.ps */
        sprintf(plotfn, "%c%03d%03d%04.0f.ps", abbr[type-3], d_version,
          f_version, 10.*Time);

    init_hardware();
    make_plot[type]();    /* pointer to one of plotting routines */
    DONEPL();

    switch (type) {
        case 1:
            free(edg4);
            /* FALL THROUGH */
        case 0:
            free(Pt_x);
            free(Pt_y);
            break;
        case 2:
            free(edg4);
            break;
        case 3:
        case 4:
            free(edg6);
            break;
        case 6:
        case 7:
        case 8:
            free(time_axis);
            free(ampl);
            break;
        default:
            break;
    }

/*--------------------------------------------------------------------------
    Rename file according to program version numbers and current time (or 
    number of points, in case of initial load).  Renaming prevents appending
    to an existing file; if exists, get error message *after* finished, but
    tplot.ps still exists and may be renamed manually.
  --------------------------------------------------------------------------*/

    if (!stat("tplot.ps", &statbuf))
        printf("\nTotal number of bytes written:  %ld\n", statbuf.st_size);

    if (!stat(plotfn, &statbuf))
        fprintf(stderr, "\ntplot:  [ %s ] exists; not renaming [ tplot.ps ]\n",
          plotfn);
    else {
        printf("\nRenaming output file to [ %s ]\n", plotfn);
        if (rename("tplot.ps", plotfn))
            fprintf(stderr, "tplot:  could not rename [ tplot.ps ] to [ %s ]\n",
              plotfn);
    }

/*---------------------------------------------------------------------------
    All done:  calculate CPU time and exit.
  ---------------------------------------------------------------------------*/

    printf("\nProgram completed, possibly successfully.  Have an "
      "excruciatingly nice day.\n");

    dt = SECOND() - t1;
    printf("(program CPU time: %8.4f sec)\n\n", dt);

} /* end function main() */





/************************/
/* Function read_data() */
/************************/

static void read_data( void )
{
    int           i, j, total_bytes;
    double        **p, *ring_data;


/*---------------------------------------------------------------------------
    Open data file for input and read karr[] and rarr[].
  ---------------------------------------------------------------------------*/

    printf("\nOpening data file [ %s ] for input.\n", datafn);
    if ((infile = fopen(datafn,"rb")) == NULL) {
        fprintf(stderr, "tplot:  couldn't open file [ %s ]\n", datafn);
        exit(1);
    }

    total_bytes  = fread(karr, sizeof(int), 256, infile) * sizeof(int);
    total_bytes += fread(rarr, sizeof(double), 256, infile) * sizeof(double);

/*---------------------------------------------------------------------------
    Depending on plot type, allocate one or more arrays and read points and/
    or edges (tesselation or triangulation).
  ---------------------------------------------------------------------------*/

    switch (type) {
        case 0:
        case 1:
            Pt_x = (double *) malloc((unsigned) numV*sizeof(double));
            Pt_y = (double *) malloc((unsigned) numV*sizeof(double));
            if (!Pt_x || !Pt_y) {
                fprintf(stderr,"tplot:  couldn't allocate memory for points\n");
                exit(10);
            }
            total_bytes += fread(Pt_x, sizeof(double), numV, infile) *
                                       sizeof(double);
            total_bytes += fread(Pt_y, sizeof(double), numV, infile) *
                                       sizeof(double);
            printf("  number of vertices:         %5d\n", numV);
            if (!type)  /* type 0 just has points */
                break;

            if (stat(datafn, &statbuf)) {
                fprintf(stderr, "tplot:  stat(%s) failed\n", datafn);
                exit(1);
            }
            num_edges = numDirE = (statbuf.st_size - 256*sizeof(int) - 
                              (256+2*numV)*sizeof(double)) / (4*sizeof(double));
            edg4 = (four_vector *) malloc((unsigned)(4*numDirE)*sizeof(double));
            if (!edg4) {
                fprintf(stderr, "tplot:  couldn't allocate memory for edges\n");
                exit(10);
            }
            total_bytes += fread(edg4, sizeof(double), 4*numDirE, infile) *
                                       sizeof(double);
            printf("  number of Dirichlet edges:  %5d\n", numDirE);
            break;

        case 2:
            num_edges = numE;
            edg4 = (four_vector *) malloc((unsigned) (4*numE)*sizeof(double));
            if (!edg4) {
                fprintf(stderr, "tplot:  couldn't allocate memory for edges\n");
                exit(10);
            }
            total_bytes += fread(edg4, sizeof(double), 4*numE, infile) *
                                       sizeof(double);
            printf("  number of triangle edges:   %5d\n", numE);
            break;

        case 3:
            edg6 = (six_vector *) malloc((unsigned) (6*numE)*sizeof(double));
            if (!edg6) {
                fprintf(stderr, "tplot:  couldn't allocate memory for edges\n");
                exit(10);
            }
            total_bytes += fread(edg6, sizeof(double), 6*numE, infile) *
                                       sizeof(double);
            printf("  number of triangle edges:   %5d\n", numE);
            break;

        case 4:
            edg6 = (six_vector *) malloc((unsigned) (6*numE)*sizeof(double));
            if (!edg6) {
                fprintf(stderr, "tplot:  couldn't allocate memory for edges\n");
                exit(10);
            }
            total_bytes += fread(edg6, sizeof(double), 6*numE, infile) *
                                       sizeof(double);
            printf("  viewing theta:            %7.1f\n", theta);
            printf("  viewing phi:              %7.1f\n", phi);
            printf("  number of triangle edges:   %5d\n", numE);
            break;

        case 5:
            fprintf(stderr, "tplot:  rings plot not yet implemented\n");
            exit(99);
            break;

        case 6:
        case 7:
        case 8:
            if (stat(datafn, &statbuf)) {
                fprintf(stderr, "tplot:  stat(%s) failed\n", datafn);
                exit(1);
            }
            numsteps = statbuf.st_size / (256*(sizeof(int) + sizeof(double)));
            time_axis = (double *) malloc((unsigned) numsteps*sizeof(double));
            ampl = (double *)
                   malloc((unsigned) ((numRings+1)*numsteps)*sizeof(double));
            p = (double **) malloc((unsigned) (numRings+1)*sizeof(double *));
            if (!time_axis || !ampl || !p) {
                fprintf(stderr, "tplot:  couldn't allocate memory for rings\n");
                exit(10);
            }
            if (type == 6)
                ring_data = pRing_R + 3*(numRings+1);
            else if (type == 7)    /* 2*(numRings+1) -> ring_v */
                ring_data = pRing_R + (numRings+1);
            else /* type == 8 */
                ring_data = pRing_R;
            /* initial p[] pointer array */
            for (j = 0;  j <= numRings;  ++j)
                p[j] = ampl + j*numsteps;
            /* copy info from first karr/rarr pair (read at top of routine) */
            *time_axis = Time;
            for (j = 0;  j <= numRings;  ++j)
                *p[j]++ = ring_data[j];
            /* loop through steps, reading karr/rarr and storing info */
            for (i = 1;  i < numsteps;  ++i) {
                total_bytes += fread(karr, sizeof(int), 256, infile) *
                                           sizeof(int);
                total_bytes += fread(rarr, sizeof(double), 256, infile) *
                                           sizeof(double);
                time_axis[i] = Time;
                for (j = 0;  j <= numRings;  ++j)
                    *p[j]++ = ring_data[j];
            }
            free(p);
            printf("  number of rings:            %5d\n", numRings);
            printf("  number of time steps:       %5d\n", numsteps);
            break;

#if 0
        case N:   /* hypothetical render/surface plot */
            total_bytes += fread(Vx, sizeof(double), numV, infile) * 
                                     sizeof(double);
            total_bytes += fread(Vy, sizeof(double), numV, infile) * 
                                     sizeof(double);
            total_bytes += fread(Vz, sizeof(double), numV, infile) * 
                                     sizeof(double);

            total_bytes += fread(tri, sizeof(int), 3*numT, infile) * 
                                      sizeof(int);
#endif /* 0 */

    } /* end switch (type) */

    fclose(infile);
    printf("  step number:                %5d\n", iteration);
    printf("  time (Myr):              %8.1f\n", Time*Myr_time);
    printf("Total bytes read from file: %7d\n\n", total_bytes);

} /* end function read_data() */





/****************************/
/* Function init_hardware() */
/****************************/

static void init_hardware( void )
{
    int   iarr[16]={0.,0.,0.};


/*---------------------------------------------------------------------------
    Open output file for plot:  use IOMGR to intercept PostScript on way to
    printer, and save in file instead.  This version appends to the end of 
    an existing file of the same name rather than overwriting it (either way,
    the rename() call at the end keeps the files straight).  Also signal use
    of hardware fonts and shading.
  ---------------------------------------------------------------------------*/

    IOMGR( &i5, &im102);
/*  IOMGR( &i2, &im104);         overwrite old file */
    IOMGR( "tplot.ps", &im103);  /* this string works...why? */
    IOMGR( &iarr, &im111);       /* ??? */
    PSCRPT( &d0, &d0, &dp003);   /* default page height/width, .003" lines */
    /* use hardware (PS) characters for all alphabets through Hebrew */
/*  HWCHAR( &d100, &d100, &d90, &i1, _cptofcd("HEBREW",6));  */
    HWCHAR( &d100, &d100, &d90, &i1, "HEBREW");   /* this works:  why?? */
#ifdef RENDER
    HWSHD();   /* used for rendering; probably doesn't hurt other plots */
#endif

/*---------------------------------------------------------------------------
   [Allocate DISSPLA memory.  The BCOMON call is only necessary for render-
    ing, but apparently we don't need a "COMMON" set aside (on either the 
    VAX or Y-MP).  Amount of memory requested is a wild guess.]
  ---------------------------------------------------------------------------*/

#ifdef RENDER
    BCOMON( &i500000);
    SYSBUF();   /* dynamically allocate memory */
#endif

} /* end function init_hardware() */





/************************/
/* Function init_page() */
/************************/

static void init_page( void )
{
    char     *owner=OWNER;
    int      i;
    double   y=d8p63;
    time_t   CurTim, *pCurTim=&CurTim;


/*---------------------------------------------------------------------------
    Initialize layout and plotting region.
  ---------------------------------------------------------------------------*/

/*  RESET( "ALL");  */
    PAGE( &d8p5, &d11);
    NOBRDR();

    SHDCHR( &d90, &i1, &dp001, &i1);
/*  SWISSM();  */  /* Helvetica */
    SCMPLX();      /* Times Roman */

#ifdef RENDER
    AREA2D( &d8, &d10);
    VOLM3D( &d30, &d30, &d30);   /* from -15 to +15, all three axes */
    VIEW( &x, &y, &z);           /* view from user-defined position */
    if (axes) {
        INTAXS();
        X3NAME( _cptofcd("X",1), &i1);
        Y3NAME( _cptofcd("Y",1), &i1);
        Z3NAME( _cptofcd("Z",1), &i1);
    }
    GRAF3D( &axmin,&axinc,&axmax, &axmin,&axinc,&axmax, &axmin,&axinc,&axmax);

#else /* !RENDER */
    AREA2D( &d6p366, &d6p332);
    INTAXS();
    if (*xtitle)
        XNAME( _cptofcd(xtitle,strlen(xtitle)), &i100);
    else
        XNAME( _cptofcd(" ",1), &i100);
    if (*ytitle)
        YNAME( _cptofcd(ytitle,strlen(ytitle)), &i100);
    else
        YNAME( _cptofcd(" ",1), &i100);
    if (!custom_axes) {
        xmin = ymin = axmin;
        xinc = yinc = axinc;
        xmax = ymax = axmax;
    }
    GRAF( &xmin,&xinc,&xmax, &ymin,&yinc,&ymax);
    XNONUM();
    XGRAXS( &xmin,&xinc,&xmax, &d6p366, _cptofcd(" ",1), &im1, &d0, &d6p332);
    YNONUM();
    YGRAXS( &ymin,&yinc,&ymax, &d6p332, _cptofcd(" ",1), &im1, &d6p366, &d0);
    THKFRM( &dp01);
    FRAME();
#endif /* ?RENDER */

/*---------------------------------------------------------------------------
    Print vital info across top of page.
  ---------------------------------------------------------------------------*/

    HEIGHT( &dp05);   /* very small text */

    /* left title block */
    MESSAG( _cptofcd(owner,strlen(owner)), &i100, &d0, &d8p81);
    time( pCurTim);
    strftime( title, 80, "Date:  %Y %b %d (%a) %H:%M %Z", localtime(pCurTim));
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d0, &d8p72);
    y = d8p63;
    if (f_version) {
        sprintf( title, "Evolver version:  FLOPPY %d.%02d", f_version/100,
          f_version%100);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d0, &y);
        y -= .09;
    }
    if (d_version) {
        sprintf( title, "Loader version:   DISC %d.%02d", d_version/100,
          d_version%100);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d0, &y);
        y -= .09;
    }
    if (p_version >= 310) {
        sprintf( title, "Points version:   POINTS %d.%02d", p_version/100,
          p_version%100);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d0, &y);
        y -= .09;
    }

    /* left middle title block */
    sprintf( title, "Plotting program:  TPLOT %s", VERSION);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d1p5, &d8p81);
    sprintf( title, "Plot type:  %d", type);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d1p5, &d8p72);
    sprintf( title, "Data file:  %s", datafn);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d1p5, &d8p63);
    sprintf( title, "Plot file:  %s", plotfn);
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d1p5, &d8p54);

    /* right middle title block */
    sprintf( title, "Program time:  %d Myr  (%d units)",
      (int)(Time*Myr_time+.5), (int)(Time+.5) );
    MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p81);
    if (type <= 5) {
        sprintf( title, "Number of vertices:  %d", numV);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p72);
        if (type == 1) {
            sprintf( title, "Number of Dirichlet edges:  %d", numDirE);
            MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p63);
        } else if (type >= 2) {
            sprintf( title, "Number of edges:    %d", numE);
            MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p63);
            sprintf( title, "Number of triangles:  %d", numT);
            MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p54);
        }
    } else if (type <= 8) {
        sprintf( title, "Number of rings:  %d", numRings);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d3, &d8p72);
    }

    /* right title block */
    if (print_angles) {
        sprintf( title, "Viewing angles:  %.0f, %.0f", theta, phi);
        MESSAG( _cptofcd(title,strlen(title)), &i100, &d4p5, &d8p81);
    }

    HEIGHT( &dp2);   /* normal-size text (?) */
    if (!check_range)
        NOCHEK();    /* suppress out-of-range messages */

} /* end function init_page() */





/*****************************/
/*  Function PlotInitTess()  */
/*****************************/

static void PlotInitTess( void )
{
    int   i;


/*---------------------------------------------------------------------------
    Initialize DISSPLA page and write header info and title.
  ---------------------------------------------------------------------------*/

    theta = 0.;
    phi = -90.;
    init_page();

    switch (type) {
        case 0:
            sprintf( title, "Initial Points Load");
            break;
        case 1:
            sprintf( title, "Initial Tesselation");
            break;
        case 2:
            sprintf( title, "Initial (Unwarped) Triangulation");
            break;
        default:
            fprintf(stderr, "tplot:  wrong routine for this plot type (%d)\n",
              type);
            return;
            break;
    }
    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i1);

/*---------------------------------------------------------------------------
    POINTS PLOT (TYPE 0 & 1)
  ---------------------------------------------------------------------------*/

    if (type < 2) {
        SCLPIC( &dp1);
        MARKER( &i2);
        CURVE( Pt_x, Pt_y, &numV, &im1);
    }

/*---------------------------------------------------------------------------
    (UNWARPED) TESSELATION OR TRIANGULATION PLOT (TYPE 1 & 2)
  ---------------------------------------------------------------------------*/

    if (type)
        for (i = 0;  i < num_edges;  ++i)
            RLVEC( &edg4[i][0], &edg4[i][1], &edg4[i][2], &edg4[i][3], &i0);

    ENDPL( &i0);

} /* end function PlotInitTess() */





/***************************/
/* Function PlotWarpTri3() */
/***************************/

static void PlotWarpTri3( void )
{
    int    i;
    char   *subtitle[3]={"Top view (down Z-axis)", "Front view (down X-axis)",
                         "Side view (down Y-axis)"};


/*---------------------------------------------------------------------------
    TOP-VIEW WARPED TRIANGULATION PLOT (TYPE 3a)
  ---------------------------------------------------------------------------*/

    theta = 0.;
    phi = 0.;
    strcpy( xtitle, "Y-axis --->");
    strcpy( ytitle, "<--- X-axis");
    xmin = axmin;
    xinc = axinc;
    xmax = axmax;
    ymin = -axmin;
    yinc = -axinc;
    ymax = -axmax;
    custom_axes = TRUE;
    init_page();

    if (Time > 0.0)
        sprintf( title, "Evolved, Warped Galactic Gas Disk, %d Myr",
          (int)(Time*Myr_time+.5) );
    else
        sprintf( title, "Warped Galactic Gas Disk, 0 Myr");
    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i2);
    HEADIN( _cptofcd(subtitle[0],strlen(subtitle[0])), &i100, &dp7, &i2);

    for (i = 0;  i < numE;  ++i)
        RLVEC( &edg6[i][1], &edg6[i][0], &edg6[i][4], &edg6[i][3], &i0);

    ENDPL( &i0);

/*---------------------------------------------------------------------------
    FRONT-VIEW WARPED TRIANGULATION PLOT (TYPE 3b)
  ---------------------------------------------------------------------------*/

    theta = 90.;
    phi = 0.;
    strcpy( xtitle, "Y-axis --->");
    strcpy( ytitle, "Z-axis --->");
    custom_axes = FALSE;
    init_page();

    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i2);
    HEADIN( _cptofcd(subtitle[1],strlen(subtitle[1])), &i100, &dp7, &i2);

    for (i = 0;  i < numE;  ++i)
        RLVEC( &edg6[i][1], &edg6[i][2], &edg6[i][4], &edg6[i][5], &i0);

    ENDPL( &i0);

/*---------------------------------------------------------------------------
    SIDE-VIEW WARPED TRIANGULATION PLOT (TYPE 3c)
  ---------------------------------------------------------------------------*/

    theta = 90.;
    phi = 90.;
    strcpy( xtitle, "<--- X-axis");
    strcpy( ytitle, "Z-axis --->");
    xmin = -axmin;
    xinc = -axinc;
    xmax = -axmax;
    ymin = axmin;
    yinc = axinc;
    ymax = axmax;
    custom_axes = TRUE;
    init_page();

    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i2);
    HEADIN( _cptofcd(subtitle[2],strlen(subtitle[2])), &i100, &dp7, &i2);

    for (i = 0;  i < numE;  ++i)
        RLVEC( &edg6[i][0], &edg6[i][2], &edg6[i][3], &edg6[i][5], &i0);

    ENDPL( &i0);

} /* end function PlotWarpTri3() */





/**************************/
/* Function PlotWarpTri() */
/**************************/

static void PlotWarpTri( void )
{
    int      i;
    double   sinth, costh, sinph, cosph;
    double   x1, y1, x2, y2;


/*---------------------------------------------------------------------------
    Initialize DISSPLA page and write header info and title.
  ---------------------------------------------------------------------------*/

    init_page();

    if (Time > 0.0)
        sprintf( title, "Evolved, Warped Galactic Gas Disk, %d Myr",
          (int)(Time*Myr_time+.5) );
    else
        sprintf( title, "Warped Galactic Gas Disk, 0 Myr");
    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i2);
    sprintf( title, "Orthographic view");
    HEADIN( _cptofcd(title,strlen(title)), &i100, &dp7, &i2);

/*---------------------------------------------------------------------------
    CUSTOM-VIEW WARPED TRIANGULATION PLOT (TYPE 4)
  ---------------------------------------------------------------------------*/

    sinth = sin(theta*PI180);
    costh = cos(theta*PI180);
    sinph = sin(phi*PI180);
    cosph = cos(phi*PI180);

    for (i = 0;  i < numE;  ++i) {      /* rotate viewpoint [notes 1990.10.9] */
        x1 = -edg6[i][0]*sinph       + edg6[i][1]*cosph;
        y1 = -edg6[i][0]*cosph*costh - edg6[i][1]*sinph*costh +
              edg6[i][2]*sinth;
        x2 = -edg6[i][3]*sinph       + edg6[i][4]*cosph;
        y2 = -edg6[i][3]*cosph*costh - edg6[i][4]*sinph*costh +
              edg6[i][5]*sinth;

        RLVEC( &x1, &y1, &x2, &y2, &i0);
    }

    ENDPL( &i0);

} /* end function PlotWarpTri() */





/************************/
/* Function PlotRings() */
/************************/

static void PlotRings( void )
{

/*---------------------------------------------------------------------------
    Initialize DISSPLA page and write header info and title.
  ---------------------------------------------------------------------------*/

    init_page();

    sprintf( title, "Evolved, Warped Galactic Gas Disk");
    HEADIN( _cptofcd(title,strlen(title)), &i100, &d1, &i2);
    sprintf( title, "Orthographic view");
    HEADIN( _cptofcd(title,strlen(title)), &i100, &dp7, &i2);

/*---------------------------------------------------------------------------
    CUSTOM-VIEW WARPED RINGS PLOT (TYPE 5)
  ---------------------------------------------------------------------------*/

#if 0
      call pRead(attrib, points, numV)

      i_max = 1
      cur_attrib = 0

C     *** loop over rings
      do 290 n = 1,attrib(numV)
        numpts = 0
        cur_attrib = cur_attrib+1

C       *** store all the points from the current ring
  250   if (attrib(i_max+1) .eq. cur_attrib) then
          i_max = i_max+1
          numpts = numpts+1
C         *** rotate viewpoint [see notes 1990.10.9]
          ring(numpts,1) = -points(1,i_max)*sinph + 
     +      points(2,i_max)*cosph
          ring(numpts,2) = -points(1,i_max)*cosph*costh -
     +      points(2,i_max)*sinph*costh + points(3,i_max)*sinth
c         *** for multi-view plot:
c         do 270 j = 1,3
c 270       ring(numpts,j) = points(j,i_max)
          if (i_max .lt. numV) goto 250
        endif

C       *** close the curve
        numpts = numpts+1
        do 280 j = 1,3
  280     ring(numpts,j) = ring(1,j)

C       *** draw the curve
        call CURVE(ring(1,1), ring(1,2), numpts, 0)
  290   continue

#endif /* 0 */

    ENDPL( &i0);

} /* end function PlotRings() */





/*****************************/
/* Function PlotRingParams() */
/*****************************/

static void PlotRingParams( void )
{
    char     Title[80];
    int      j;
    double   y;


/*---------------------------------------------------------------------------
    Initialize DISSPLA page and write header info and title.
  ---------------------------------------------------------------------------*/

    switch (type) {
        case 6:
            ymax = 200.;
            ymin = -200.;
            strcpy( ytitle, "Azimuth angle (degrees)");
            sprintf( Title, "Evolution of Ring Azimuths");
            break;
        case 7:
            ymax = 50.;
            ymin = 0.;
            strcpy( ytitle, "Inclination angle (degrees)");
            sprintf( Title, "Evolution of Ring Inclinations");
            break;
        case 8:
            ymax = (int) (Rdisk*1.2 + .5);
            ymin = 0.;
            strcpy( ytitle, "Ring radius (program units)");
            sprintf( Title, "Evolution of Ring Radii");
            break;
        default:
            fprintf(stderr, "tplot:  wrong routine for this plot type (%d)\n",
              type);
            return;
            break;
    }
    print_angles = FALSE;
    strcpy( xtitle, "Time (program units)");
    xmin = 0;
    xmax = Time;
    xinc = (xmax-xmin) / 8.;
    yinc = (ymax-ymin) / 10.;
    custom_axes = TRUE;
    init_page();
    if (type == 6) {
        DOT();
        y = 180.;
        RLVEC( &xmin, &y, &xmax, &y, &i0);
        y = -180.;
        RLVEC( &xmin, &y, &xmax, &y, &i0);
        RESET( _cptofcd("DOT",3));
    }

    if (first_ring < 0)
        first_ring = 0;
    if (last_ring > numRings)
        last_ring = numRings;

    HEADIN( _cptofcd(Title,strlen(Title)), &i100, &d1, &i2);
    if (last_ring == first_ring)
        sprintf( title, "Ring %d", first_ring);
    else if (last_ring == first_ring+1)
        sprintf( title, "Rings %d and %d", first_ring, last_ring);
    else
        sprintf( title, "Rings %d to %d", first_ring, last_ring);
    HEADIN( _cptofcd(title,strlen(title)), &i100, &dp7, &i2);

/*---------------------------------------------------------------------------
    AZIMUTH/INCLINATION/RADII PLOT (TYPE 6, 7 & 8)
  ---------------------------------------------------------------------------*/

    for (j = first_ring;  j <= last_ring;  ++j)
        CURVE( time_axis, &ampl[j*numsteps], &numsteps, &i0);

    ENDPL( &i0);

} /* end function PlotRingParams() */





/*************************/
/* Function PlotInclin() */
/*************************/

#if 0
static void PlotInclin( void )
{
    ENDPL( &i0);
} /* end function PlotInclin() */
#endif /* 0 */





#if 0
    MESSAG( _cptofcd("Program time:",13), &i100, &d1p5, &d8p81);
    NEWCLR( _cptofcd("BACK",4));   /* invis. */
    MESSAG( _cptofcd("m",1), &i100, _cptofcd("ABUT",4), _cptofcd("ABUT",4));
    NEWCLR( _cptofcd("FORE",4));
    i = (int) (Time + .5);
    INTNO( &i, _cptofcd("ABUT",4), _cptofcd("ABUT",4));
    MESSAG( _cptofcd(" units",6), &i100, _cptofcd("ABUT",4),_cptofcd("ABUT",4));
    NEWCLR( _cptofcd("BACK",4));
    MESSAG( _cptofcd("Program time:m",14), &i100, &d1p5, &d8p72);  /* invis. */
    NEWCLR( _cptofcd("FORE",4));
    i = (int) (Time*Myr_time + .5);
    INTNO( &i, _cptofcd("ABUT",4), _cptofcd("ABUT",4));
/*  REALNO(phi,+2,'ABUT','ABUT')  */
    MESSAG( _cptofcd(" Myr",4), &i100, _cptofcd("ABUT",4), _cptofcd("ABUT",4));
#endif /* 0 */
