/*---------------------------------------------------------------------------

            UU    UU  PPPPPP    DDDDDD       AA   TTTTTTTT EEEEEEEE 
            UU    UU  PP    PP  DD    DD    AAAA     TT    EE       
            UU    UU  PP    PP  DD    DD   AA  AA    TT    EEEEEEE  
            UU    UU  PPPPPP    DD    DD  AA    AA   TT    EE       
            UU    UU  PP        DD    DD  AAAAAAAA   TT    EE       
             UUUUUU   PP        DDDDDD    AA    AA   TT    EEEEEEEE 

  ---------------------------------------------------------------------------

  update.c

  This file contains update_points(), the actual physical evolution code.
  It compiles into a stand-alone object file.

  ---------------------------------------------------------------------------

     Version:  Y-MP                                  Greg Roelofs
     Source:   update332.c on reynolds               25 May 1992

  ---------------------------------------------------------------------------

  To compile:

     on Y-MP:  scc -c update332.c
     on PC:    cl -c -AS -G2 -FPi87 update332.c

  ---------------------------------------------------------------------------

  Associated files:

     fd.h:          main header file:  common macros, prototypes, etc.
     sh###.c:       routines shared by both loader and evolver
     grav.f:        gravity code (used by update###.c and d###.c)

     p###.c:        stand-alone program to create and plot points only

     d###.c:        main tesselation/triangulation code; loader
     tesselate.c:   tesselation routines called by d###.c
     plotpts.c:     MS-DOS graphics routines for points load
     plotload.c:    MS-DOS graphics routines for tesselation

     f###.c:        main evolution code (evolver)
  -> update###.c:   physics (update_points, used only by f###.c)
     unrolled.h:    include file for update###.c (unrolled inner loops)
     jacobi.c:      ring-fitting code (used only by update###.c)
     repair.c:      most grid-maintenance routines (used only by f###.c)

     update###d.c, repaird.o, f###d:  debugging versions

  ---------------------------------------------------------------------------

   Version history:

      h280.c:  physics code first split out into separately compiled source
                file; removed all grid-repair code
      h285.c:  moved gravitational-force calculation into separate (FORTRAN)
                routine, GRAV()
      h286.c:  fixed bug in viscosity code
      h287.c:  unrolled inner loops for better vectorization
      f290.c:  added new boundary-conditions code (including ring-fitting)
      f291.c:  changed definition of n_phantom[] to simpler one
      f292.c:  no changes to update_points()
      f300.c:  slight change to dbgfile output (less verbose); removed sec-
                ond order effect in pressure calculation (pressure forces
                moved to edge loop from triangle loop)
      f310.c:  modified to reflect new definition of numRings
      f330.c:  added simple local self-gravity (restoring force in disk)
      f331.c:  changed local self-gravity disk profile from infinitely thin
                sheet to isothermal (sech^2 z) sheet
      f332.c:  reduced disk scale height to 120 pc; added jacobi timing info

  ---------------------------------------------------------------------------*/



#include "fd.h"

#define n_phantom     s[1]      /* temporary storage */
#define r_hat         s[0]
#define theta_hat     s[2]
#define r_proj        s[0]
#define orig          s[0]
#define almost_one    x
#define almost_zero   y

int neighbors[MAX_V];    /* number of neighboring triangles */
double T_area[MAX_T];    /* triangle area */
double scaleht2 = .128;  /* 240pc = scale height times two */



/******************************/
/*  Function update_points()  */
/******************************/

void update_points( void )
{
    int      i, j, j1, j2, jmax, k, m, numpts, nrot;
    double   t1, t2, dt, j_t1, mass_over_3, max_z, min_z;
    double   mu_phantom, pres_phantom, area_phantom_half;
    double   min_area=9.e9, angle, x, y, z, r;
    double   numer, testval, costhet, norm;
    double   blk[2], force[3], alt;
    double   rot_inertia[3][3], eigenval[3];
    vector   shr[2], etahat, eigenvec[3];

    static int      i8=8;
    static double   d11_18=11./18., d7_36=7./36.;



/*---------------------------------------------------------------------------
    First clear physical variables.
  ---------------------------------------------------------------------------*/

    GravE = 0.;
    KinE = 0.;
    for (i = 0;  i < 3;  ++i) {
        LinMom[i] = 0.;
        AngMom[i] = 0.;
    }

/*===========================================================================
    Now loop through triangles and store coordinates of the centroid of 
    each sub-triangle.  Then ship array off to FORTRAN routine GRAV() to 
    calculate gravitational potential and acceleration at each position.
    [The force due to gravity is calculated approximately, by assuming the
    mass of each trisection is concentrated at its center-of-mass position.]
  ===========================================================================*/

    t1 = SECOND();
    j_t1 = jacobi_time;

    for (k = 0;  k < numT;  ++k) {      /* triangles */
#ifdef ROLLED
        for (j = 0;  j < 3;  ++j)       /* sub-triangles */
            for (i = 0;  i < 3;  ++i)   /* cartesian coordinates */
                cm_subT[k][j][i] = d11_18 * T[k].pV[j]->r[i]  +
                   d7_36 * (T[k].pV[(j+1)%3]->r[i] + T[k].pV[(j+2)%3]->r[i]);
#else /* !ROLLED */
#  define LOOP1
#  include "unrolled.h"
#endif /* ?ROLLED */
    } /* end k-loop (triangles) */

    numpts = 3*numT + num_phantom;

    /* loop over "phantom" external wedgies:  approximate positions as
       radial extrapolation of outermost ring points */
    for (k = ring_last[numRings-1]+1, j = 0;  j < num_phantom;  ++j, ++k)
        for (i = 0;  i < 3;  ++i)
            cm_phantom[j][i] = (r_cm_phantom / Rdisk) * V[k].r[i];

    t2 = SECOND();
    GRAV(&numpts);
    grav_time += SECOND() - t2;

/*===========================================================================
    Now loop through triangles again and add up gravitational forces on ver-
    tices.  Gravity will tend to flatten disk; gas pressure will produce a
    slight tendency toward buckling it.
  ===========================================================================*/

    for (k = 0;  k < numT;  ++k) {

    /*-----------------------------------------------------------------------
        Start by looping over vertices of current triangle and calculating
        vectors of sides of the triangle.  Then use vectors to calculate the
        area (density) and pressure.  [Get pressure from ideal gas law equa-
        tion (mass/area = density...); area of triangle equals half magnitude
        of cross-product of 2 sides.]
      -----------------------------------------------------------------------*/
        mass_over_3 = T[k].mass / 3.;
#ifdef ROLLED
        for (j = 0;  j < 3;  ++j)         /* vertices */
            for (i = 0;  i < 3;  ++i)     /* cartesian coordinates */
                s[j][i] = T[k].pV[j]->r[i] - T[k].pV[(j+1)%3]->r[i];
#else /* !ROLLED */
        s[0][0] = T[k].pV[0]->r[0] - T[k].pV[1]->r[0];
        s[0][1] = T[k].pV[0]->r[1] - T[k].pV[1]->r[1];
        s[0][2] = T[k].pV[0]->r[2] - T[k].pV[1]->r[2];

        s[1][0] = T[k].pV[1]->r[0] - T[k].pV[2]->r[0];
        s[1][1] = T[k].pV[1]->r[1] - T[k].pV[2]->r[1];
        s[1][2] = T[k].pV[1]->r[2] - T[k].pV[2]->r[2];

        s[2][0] = T[k].pV[2]->r[0] - T[k].pV[0]->r[0];
        s[2][1] = T[k].pV[2]->r[1] - T[k].pV[0]->r[1];
        s[2][2] = T[k].pV[2]->r[2] - T[k].pV[0]->r[2];
#endif /* ?ROLLED */

        T_area[k] = sqrt(SQR( XCROSS(s[1],s[2]) ) +
                    SQR( YCROSS(s[1],s[2]) ) +
                    SQR( ZCROSS(s[1],s[2]) )) / 2.;
        min_area = MIN(T_area[k], min_area);      /* [notes, 1990.10.14] */
        T[k].pressure = kT_m * T[k].mass / T_area[k];

        for (j = 0;  j < 3;  ++j) {   /* loop over three vertices */
            j1 = (j+1)%3;
            j2 = (j+2)%3;

            /* increment neighbor-triangle counter for each vertex */
            ++neighbors[T[k].pV[j]-V];

        /*-------------------------------------------------------------------
            Loop over coordinates again:  add gravitational forces to sums.
          -------------------------------------------------------------------*/
#ifdef ROLLED
            for (i = 0;  i < 3;  ++i)
                T[k].pV[j]->F[i] += mass_over_3 * grav_accel[k][j][i];
#else /* !ROLLED */
            T[k].pV[j]->F[0] += mass_over_3 * grav_accel[k][j][0];
            T[k].pV[j]->F[1] += mass_over_3 * grav_accel[k][j][1];
            T[k].pV[j]->F[2] += mass_over_3 * grav_accel[k][j][2];
#endif /* ?ROLLED */

        /*-------------------------------------------------------------------
            Finally (for triangle loop, anyway), calculate quantities which
            should be conserved:  energy, momentum, angular momentum.
          -------------------------------------------------------------------*/
            GravE += mass_over_3 * grav_potent[k][j][3];
            KinE  += T[k].mass * ( DOT(T[k].pV[j]->v, T[k].pV[j]->v)  +
                                   DOT(T[k].pV[j]->v, T[k].pV[j1]->v) );
/*          IntE  +=  */

            /* "cheat" by using j as cartesian index (linear momentum ONLY) */
            LinMom[j] += mass_over_3 *
                      (T[k].pV[0]->v[j] + T[k].pV[1]->v[j] + T[k].pV[2]->v[j]);

            AngMom[0] += T[k].mass * ( XCROSS(T[k].pV[j]->r, T[k].pV[j]->v)*2.
                                     + XCROSS(T[k].pV[j]->r, T[k].pV[j1]->v)
                                     + XCROSS(T[k].pV[j]->r, T[k].pV[j2]->v) );
            AngMom[1] += T[k].mass * ( YCROSS(T[k].pV[j]->r, T[k].pV[j]->v)*2.
                                     + YCROSS(T[k].pV[j]->r, T[k].pV[j1]->v)
                                     + YCROSS(T[k].pV[j]->r, T[k].pV[j2]->v) );
            AngMom[2] += T[k].mass * ( ZCROSS(T[k].pV[j]->r, T[k].pV[j]->v)*2.
                                     + ZCROSS(T[k].pV[j]->r, T[k].pV[j1]->v)
                                     + ZCROSS(T[k].pV[j]->r, T[k].pV[j2]->v) );

        }  /* end j-loop (vertices) */

    }  /* end k-loop (triangles) */

/*---------------------------------------------------------------------------
    Loop over vertices on outermost ring and add in "phantom" gravitational
    contributions.  Also correct for neighbor-triangle count of hull verts.
  ---------------------------------------------------------------------------*/

    /* loop over "phantom" external wedgies */
    for (k = ring_last[numRings-1]+1, j = 0;  j < num_phantom;  ++j, ++k) {
#ifdef ROLLED
        for (i = 0;  i < 3;  ++i)
            V[k].F[i] += mass_phantom * accel_phantom[j][i];
#else /* !ROLLED */
        V[k].F[0] += mass_phantom * accel_phantom[j][0];
        V[k].F[1] += mass_phantom * accel_phantom[j][1];
        V[k].F[2] += mass_phantom * accel_phantom[j][2];
#endif /* ?ROLLED */
        GravE += mass_phantom * potent_phantom[j][3];
        --neighbors[k];
    }

/*---------------------------------------------------------------------------
    Debugging:  double-check that edge-normals really are normal to edges.
    (Even unrolled, outer loop won't vectorize because of fprintf() call.)
  ---------------------------------------------------------------------------*/

#if 0
    for (k = 0;  k < numE;  ++k) {
        for (i = 0;  i < 3;  ++i)
            s[0][i] = E[k].pV[0]->r[i] - E[k].pV[1]->r[i];   /* temp vec */
        for (m = 0;  m < 2;  ++m)
            if (E[k].pT[m]) {
                testval = DOT(E[k].n[m], s[0]);
                if (fabs(testval) > 1.e-10)
                    fprintf(outfile, "    ERROR:  E[%d] normal not normal:\t"
                      "%13.4e\n", k, testval);
            }
    } /* end k-loop (edges) */
#endif

/*---------------------------------------------------------------------------
    Multiply conservation quantities by constants and print them.
  ---------------------------------------------------------------------------*/

    KinE /= 12.;
    TotE = KinE + GravE + IntE;
    for (i = 0;  i < 3;  ++i)
        AngMom[i] /= 12.;

    fprintf(outfile, "  KinE %11.4e   GravE %11.4e              TotE %11.4e\n",
      KinE, GravE, TotE);
    fprintf(outfile, "  LinMom %11.4e %11.4e %11.4e   TotLinMom %11.4e\n",
      LinMom[0], LinMom[1], LinMom[2], NORM(LinMom) );
    fprintf(outfile, "  AngMom %11.4e %11.4e %11.4e   TotAngMom %11.4e\n",
      AngMom[0], AngMom[1], AngMom[2], NORM(AngMom) );

    fprintf(outfile, "  Minimum triangle area = %7.5f\n", min_area);

/*===========================================================================
    Loop through edges and add in pressure forces on vertices.  [Hull edges:
    notes 1991.06.24-27, 1991.12.18; regular edges:  notes 1992.02.14].  In
    two dimensions pressure is a force per length, so the force on each edge
    is proportional to its length, and the force on each of the edge's ver-
    tices is half of that.  Edge normals point inward from the edges (or out-
    ward from the opposing vertices), so the force is negative to the normal;
    the normals are updated at the bottom of swap_edges() (in sh###.c).
  ===========================================================================*/

/*---------------------------------------------------------------------------
    "Phantom" pressure is constant for all hull edges.  Compensate for the
    stretching/shrinking of the disk:  the 2D mass density changes as the
    square of the factor by which the outermost "real" ring has grown.
  ---------------------------------------------------------------------------*/
    mu_phantom = mu_phantom_orig * SQR(Rdisk/ring_R[numRings]);
    pres_phantom = kT_m * mu_phantom;
    area_phantom_half = .5 * mass_phantom / mu_phantom;

    for (k = 0;  k < numE;  ++k) {

        if (E[k].status & E_ON_HULL) {
            *force = .5 * E[k].len * (pres_phantom - E[k].pT[0]->pressure);
#ifdef ROLLED
            for (i = 0;  i < 3;  ++i) {
                E[k].pV[0]->F[i] += (*force) * E[k].n[0][i];
                E[k].pV[1]->F[i] += (*force) * E[k].n[0][i];
            }
#else /* !ROLLED */
                E[k].pV[0]->F[0] += (*force) * E[k].n[0][0];
                E[k].pV[0]->F[1] += (*force) * E[k].n[0][1];
                E[k].pV[0]->F[2] += (*force) * E[k].n[0][2];

                E[k].pV[1]->F[0] += (*force) * E[k].n[0][0];
                E[k].pV[1]->F[1] += (*force) * E[k].n[0][1];
                E[k].pV[1]->F[2] += (*force) * E[k].n[0][2];
#endif /* ?ROLLED */
        } else {
            int Tindex, Vindex, phTindex=0, phVindex=0;
            double ht_avg=0., mass_sum=0., mu_avg=0., area_sum=0.;
            double sqrt3_2=.5*sqrt(3), area[6];
      /* double scaleht2 = .5*(E[k].pT[0]->scaleht2 + E[k].pT[1]->scaleht2); */
            struct triangle *pT[6], phantomT[4];   /* should never need > 2 */
            struct vertex *pV[8], phantomV[4];     /* ditto (rarely > 1) */

            *force = .25 * E[k].len *
              (E[k].pT[1]->pressure - E[k].pT[0]->pressure);
#ifdef ROLLED
            for (i = 0;  i < 3;  ++i) {
                E[k].pV[0]->F[i] += (*force) * (E[k].n[0][i] - E[k].n[1][i]);
                E[k].pV[1]->F[i] += (*force) * (E[k].n[0][i] - E[k].n[1][i]);
            }
#else /* !ROLLED */
                E[k].pV[0]->F[0] += (*force) * (E[k].n[0][0] - E[k].n[1][0]);
                E[k].pV[0]->F[1] += (*force) * (E[k].n[0][1] - E[k].n[1][1]);
                E[k].pV[0]->F[2] += (*force) * (E[k].n[0][2] - E[k].n[1][2]);

                E[k].pV[1]->F[0] += (*force) * (E[k].n[0][0] - E[k].n[1][0]);
                E[k].pV[1]->F[1] += (*force) * (E[k].n[0][1] - E[k].n[1][1]);
                E[k].pV[1]->F[2] += (*force) * (E[k].n[0][2] - E[k].n[1][2]);
#endif /* ?ROLLED */

        /*-------------------------------------------------------------------
            Find the surrounding triangles and vertices, or make up some
            if we're near the hull.  [Notes 1992.5.20-23]
          -------------------------------------------------------------------*/
            pV[0] = E[k].pV[0];   pT[0] = E[k].pT[0];
            pV[1] = E[k].pV[1];   pT[1] = E[k].pT[1];
            Vindex = Tindex = 2;
            for (m = 0;  m < 2;  ++m) {                 /* tri's adj to edge */
                for (j = 0;  j < 3;  ++j) {             /* edges of triangle */
                    if (pT[m]->pE[j] == &E[k])              /* original edge */
                        pV[Vindex++] = pT[m]->pV[j];
                    else {                              /* not original edge */
                        pT[Tindex] = (pT[m]->pE[j]->pT[0] == pT[m])?
                          pT[m]->pE[j]->pT[1] : pT[m]->pE[j]->pT[0];
                        if (pT[Tindex] == NULL) {          /* phantom wedgie */
                            pT[Tindex] = &phantomT[phTindex];
                            pV[Vindex++] = &phantomV[phVindex];
                            phantomT[phTindex].mass = .5 * mass_phantom;
                            area[Tindex++] = area_phantom_half;
                            for (i = 0;  i < 3;  ++i)      /* equilateral */
                                phantomV[phVindex].r[i] = -sqrt3_2 *
                                  pT[m]->pE[j]->len * pT[m]->pE[j]->n[0][i];
                            phantomV[phVindex++].mass = mass_phantom;
                        } else {                            /* true triangle */
                            for (j2 = 0;  j2 < 3;  ++j2)
                                if (pT[Tindex]->pE[j2] == pT[m]->pE[j])
                                    break;          /* GRR: check for error? */
                            area[Tindex] = T_area[pT[Tindex]-T];
                            pV[Vindex++] = pT[Tindex++]->pV[j2];
                        } /* end if (phantom wedgie or true triangle) */
                    } /* end if (edge == E[k]) */
                } /* end j-loop (edges of triangle adjacent to E[k]) */
            } /* end m-loop (triangles adjacent to E[k]) */

        /*-------------------------------------------------------------------
            Now have the 8 necessary vertices and 6 triangles:  use the
            inertia tensor of the vertices to find the local plane.
          -------------------------------------------------------------------*/
            for (i = 0;  i < 3;  ++i)
                orig[i] = .5 * (E[k].pV[0]->r[i] + E[k].pV[1]->r[i]);
            for (m = 0;  m < 9;  ++m)
                ((double *)rot_inertia)[m] = 0.;

            /* loop over vertices and add contributions to tensor */
            for (m = 0;  m < Vindex;  ++m) {
                double dx = pV[m]->r[0] - orig[0];
                double dy = pV[m]->r[1] - orig[1];
                double dz = pV[m]->r[2] - orig[2];

                rot_inertia[0][0] += pV[m]->mass * (SQR(dy) + SQR(dz));
                rot_inertia[0][1] -= pV[m]->mass * dx * dy;
                rot_inertia[0][2] -= pV[m]->mass * dx * dz;
                rot_inertia[1][1] += pV[m]->mass * (SQR(dx) + SQR(dz));
                rot_inertia[1][2] -= pV[m]->mass * dy * dz;
                rot_inertia[2][2] += pV[m]->mass * (SQR(dx) + SQR(dy));
            }
            /* tensor is symmetric... */
            rot_inertia[1][0] = rot_inertia[0][1];
            rot_inertia[2][0] = rot_inertia[0][2];
            rot_inertia[2][1] = rot_inertia[1][2];

            /* find eigenvalues, eigenvectors */
            grr_jacobi((double **)rot_inertia, 3, eigenval,
              (double **)eigenvec, &nrot);
#ifdef UDEBUG
            fprintf(dbgfile, "   number of JACOBI rotations:  %d\n", nrot);
#endif /* UDEBUG */
            if (eigenval[1] > eigenval[0])
                if (eigenval[2] > eigenval[1])  jmax = 2;
                else  jmax = 1;
            else
                if (eigenval[2] > eigenval[0])  jmax = 2;
                else  jmax = 0;

        /*-------------------------------------------------------------------
            Now have local normal to disk ("local z-axis"); decompose
            position vectors to find average height (average weighted by
            mass; could do plain average, too).  Then find new origin
            and average density of surrounding triangles.
          -------------------------------------------------------------------*/
            for (m = 0;  m < Vindex;  ++m) {
                mass_sum += pV[m]->mass;
                ht_avg += pV[m]->mass * (
                          (pV[m]->r[0] - orig[0]) * eigenvec[jmax][0] +
                          (pV[m]->r[1] - orig[1]) * eigenvec[jmax][1] +
                          (pV[m]->r[2] - orig[2]) * eigenvec[jmax][2] );
            }
            ht_avg /= mass_sum;
            for (i = 0;  i < 3;  ++i)
                orig[i] += ht_avg * eigenvec[jmax][i];

            for (m = 0;  m < Tindex;  ++m) {
                mu_avg += pT[m]->mass;
                area_sum += area[m];
            }
            mu_avg /= area_sum;

        /*-------------------------------------------------------------------
            Loop over two vertices on edge:  calculate height and restoring
            force on each vertex.  Divide force by neighbors[], since that
            many edges will contribute overall.  [Notes 1992.5.23]
          -------------------------------------------------------------------*/
            for (m = 0;  m < 2;  ++m) {
                double zeta;

                zeta = (E[k].pV[m]->r[0] - orig[0]) * eigenvec[jmax][0] +
                       (E[k].pV[m]->r[1] - orig[1]) * eigenvec[jmax][1] +
                       (E[k].pV[m]->r[2] - orig[2]) * eigenvec[jmax][2];

                *force = -tanh(zeta/scaleht2) *
                         TWOPI * G_LMT * mu_avg * E[k].pV[m]->mass / 
                                 neighbors[E[k].pV[m]-V];

                E[k].pV[m]->F[0] += (*force) * eigenvec[jmax][0];
                E[k].pV[m]->F[1] += (*force) * eigenvec[jmax][1];
                E[k].pV[m]->F[2] += (*force) * eigenvec[jmax][2];

            } /* end m-loop (vertices on E[k]) */

        } /* end if (hull edge) */

    } /* end k-loop (edges) */

/*===========================================================================
    Loop through edges and add in viscous forces on vertices.  [Hull edges:
    notes 1991.06.27, 1991.08.07-11]
  ===========================================================================*/

    for (k = 0;  k < numE;  ++k) {

        if (E[k].status & E_ON_HULL) {

            /* calc phantom unit edge-normal:  s[1] (same plane as real
             * edge-normal).  n_phantom is automatically normalized */
            n_phantom[0] = -E[k].n[0][0];
            n_phantom[1] = -E[k].n[0][1];
            n_phantom[2] = -E[k].n[0][2];

            /* define "average" edge unit-normal */
            etahat[0] = n_phantom[0] - E[k].n[0][0];
            etahat[1] = n_phantom[1] - E[k].n[0][1];
            etahat[2] = n_phantom[2] - E[k].n[0][2];
            norm = NORM(etahat);
            etahat[0] /= norm;
            etahat[1] /= norm;
            etahat[2] /= norm;

            shr[0][0] = shr[1][0] = 0.;   /* shr[1], blk[1]:  for "phantoms" */
            shr[0][1] = shr[1][1] = 0.;
            shr[0][2] = shr[1][2] = 0.;
            blk[0] = blk[1] = 0.;

            /* do triangle adjacent to edge:  loop over vertices (or edges) */
/* GRR: UNROLL this */
            for (j = 0;  j < 3;  ++j) {
                /* calculate altitude to current vertex */
                alt = DOT( E[k].pT[0]->n[j], E[k].pT[0]->pV[j]->r )  -
                      DOT( E[k].pT[0]->n[j], E[k].pT[0]->pV[(j+1)%3]->r );

                if (alt < .001) {
                    fprintf(outfile, "    WARNING:  alt = %13.4e  E[%d]  "
                      "T[%d]\n", alt, k, E[k].pT[0]-T );
                    fprintf(outfile, "    (not skipping viscous-force "
                      "calculation, but probably should)\n");
                }

                /* loop over cartesian coords and calculate terms under
                 * sum in shear expression for real cell (vector; see GRR 
                 * viscosity notes, eq. "11")  [E.n, pV->r bugs fixed, 
                 * 1991.08.08] */
                for (i = 0;  i < 3;  ++i)
                    shr[0][i] += ( DOT(etahat,E[k].pT[0]->n[j]) *
                                   E[k].pT[0]->pV[j]->v[i] +
                                   E[k].pT[0]->n[j][i] *
                                   DOT(etahat,E[k].pT[0]->pV[j]->v) ) / alt;

                /* do same for bulk expression (p. 6), except scalar */
                blk[0] += DOT(E[k].pT[0]->n[j], E[k].pT[0]->pV[j]->v) / alt;
             /* blk[1] = 0.; */    /* for circular ring-fit, div v[] is 
                                    * identically 0 (see notes, 1991.08.11) */

            } /* end j-loop (vertices) */
/* GRR: UNROLL this */

            /* calc radial unit vector (at midpoint of edge) */
            r_hat[0] = .5 * (E[k].pV[0]->r[0] + E[k].pV[1]->r[0]);
            r_hat[1] = .5 * (E[k].pV[0]->r[1] + E[k].pV[1]->r[1]);
            r_hat[2] = .5 * (E[k].pV[0]->r[2] + E[k].pV[1]->r[2]);
            norm = NORM(r_hat);
            r_hat[0] /= norm;
            r_hat[1] /= norm;
            r_hat[2] /= norm;
            
            /* calc azimuthal unit vector (also at midpoint of edge) */
            theta_hat[0] = XCROSS(ring_last_n, r_hat);
            theta_hat[1] = YCROSS(ring_last_n, r_hat);
            theta_hat[2] = ZCROSS(ring_last_n, r_hat);

            almost_one  = DOT(etahat, r_hat);
            almost_zero = DOT(etahat, theta_hat);

/* GRR UNROLL: */
            /* do "phantom" triangle/wedgie/whatever viscous force */
            for (i = 0;  i < 3;  ++i) {
                shr[1][i] = dvdr_phantom * 
                            (almost_one*theta_hat[i] + almost_zero*r_hat[i]);

                force[i] = (E[k].len/2.) * ( vis_dyn *
                  (scaleht2_phantom*shr[1][i] - E[k].pT[0]->scaleht2*shr[0][i])
                  +  etahat[i] * gamma_dyn * (-E[k].pT[0]->scaleht2*blk[0]) );

                /* add force component to each of two vertices:  each gets half
                 * of total viscous force on edge (hence "len/2" in above calc)
                 */
                E[k].pV[0]->F[i] += force[i];
                E[k].pV[1]->F[i] += force[i];
            }
/* GRR UNROLL^ */

#ifdef UDEBUG
            fprintf(dbgfile,
              "   E[%d] (hull):  almost_one = %e, almost_zero = %e\n",
              k, almost_one, almost_zero);
            fprintf(dbgfile,
              "   extra viscous force = (%10.2e, %10.2e, %10.2e)\n",
              force[0], force[1], force[2]);
#endif /* UDEBUG */

        } else if (!(E[k].status & E_SKIP_VISC)) {  /* OK to do reg. visc. */

        /*-------------------------------------------------------------------
            Do viscous force on regular edge.  [GRR Viscosity Notes, 1990]
          -------------------------------------------------------------------*/
            /* define "average" edge unit-normal */
            etahat[0] = E[k].n[1][0] - E[k].n[0][0];
            etahat[1] = E[k].n[1][1] - E[k].n[0][1];
            etahat[2] = E[k].n[1][2] - E[k].n[0][2];
            norm = NORM(etahat);
            etahat[0] /= norm;
            etahat[1] /= norm;
            etahat[2] /= norm;

            shr[0][0] = shr[1][0] = 0.;
            shr[0][1] = shr[1][1] = 0.;
            shr[0][2] = shr[1][2] = 0.;
            blk[0] = blk[1] = 0.;

#ifdef ROLLED
            /* loop over two triangles adjacent to edge */
            for (m = 0;  m < 2;  ++m)
                /* loop over vertices (or edges) of current triangle */
                for (j = 0;  j < 3;  ++j) {
                    /* calculate altitude to current vertex */
                    alt = DOT( E[k].pT[m]->n[j], E[k].pT[m]->pV[j]->r )  -
                          DOT( E[k].pT[m]->n[j], E[k].pT[m]->pV[(j+1)%3]->r );

                    if (alt < .001) {
                        fprintf(outfile, "    WARNING:  alt = %13.4e  E[%d]  "
                          "T[%d]\n", alt, k, E[k].pT[m]-T );
                        fprintf(outfile, "    (not skipping viscous-force "
                          "calculation, but probably should)\n");
                    }

                    /* loop over cartesian coords and calculate terms under
                     * sum in shear expression (vector; see GRR viscosity
                     * notes, eq. "11")  [E.n, pV->r bugs fixed, 1991.08.08] */
                    for (i = 0;  i < 3;  ++i)
                        shr[m][i] += ( DOT(etahat,E[k].pT[m]->n[j]) *
                                       E[k].pT[m]->pV[j]->v[i] +
                                       E[k].pT[m]->n[j][i] *
                                       DOT(etahat,E[k].pT[m]->pV[j]->v) ) / alt;

                    /* do same for bulk expression (p. 6), except scalar */
                    blk[m] += DOT(E[k].pT[m]->n[j], E[k].pT[m]->pV[j]->v) /
                                                alt;

                } /* end j-loop (vertices), m-loop (triangles) */
#else /* !ROLLED */
#  define LOOP2
#  include "unrolled.h"
#endif /* ?ROLLED */

            /* loop over cartesian coords */
            for (i = 0;  i < 3;  ++i) {
                force[i] = (E[k].len/2.) * ( vis_dyn *
                  (E[k].pT[1]->scaleht2*shr[1][i] -
                   E[k].pT[0]->scaleht2*shr[0][i])  +  etahat[i] * gamma_dyn *
                  (E[k].pT[1]->scaleht2*blk[1] -
                   E[k].pT[0]->scaleht2*blk[0]) );

                /* add force component to each of two vertices:  each gets half
                 * of total viscous force on edge (hence "len/2" in above calc)
                 */
                E[k].pV[0]->F[i] += force[i];
                E[k].pV[1]->F[i] += force[i];
            }

        } /* end if (OK to do viscosity calc) */
    } /* end k-loop (edges) */

/*===========================================================================
    Now loop through vertices and move them (standard leapfrog approach).
  ===========================================================================*/

    ++iteration;
    Time += delta_t;

    for (k = 0;  k < numV;  ++k)
        /* loop over three coordinates */
        for (i = 0;  i < 3;  ++i) {
            V[k].v[i] += (V[k].F[i] / V[k].mass) * delta_t;
            V[k].r[i] += V[k].v[i] * delta_t;
            V[k].F[i] = 0.;
        }

    fprintf(outfile, "\nStep%6d     Time%9.3f     ----------"
      "-----------------------------\n", iteration, Time);
#ifdef UDEBUG
    fprintf(dbgfile, "\nStep%6d     Time%9.3f     ----------"
      "-----------------------------\n", iteration, Time);
#else /* !UDEBUG */
    fprintf(dbgfile, "Step%6d     Time%9.3f\n", iteration, Time);
#endif /* ?UDEBUG */

/*===========================================================================
    Loop through edges, recalculating length of each and clearing status 
    bits.  Also do some more integrity checks.
  ===========================================================================*/

    for (k = 0;  k < numE;  ++k) {
        E[k].len = DIST(E[k].pV[0]->r, E[k].pV[1]->r);
        E[k].status &= ~(E_UNMODIFIED | E_FORCE_SWAP | E_CHECKED_FOLD |
                         E_SKIP_VISC);

#if TRUE
        /* test for vertex overrun [notes, 1990.10.23] */
        /* (this test can probably be dropped now [notes, 1991.1.6]) */
        numer = ((E[k].pV[1]->v[0] - E[k].pV[0]->v[0]) *
                 (E[k].pV[1]->r[0] - E[k].pV[0]->r[0])) +
                ((E[k].pV[1]->v[1] - E[k].pV[0]->v[1]) *
                 (E[k].pV[1]->r[1] - E[k].pV[0]->r[1])) +
                ((E[k].pV[1]->v[2] - E[k].pV[0]->v[2]) *
                 (E[k].pV[1]->r[2] - E[k].pV[0]->r[2]));

        if ((testval = numer*delta_t/SQR(E[k].len)) <= -1.) {
            x = (E[k].pV[0]->r[0] + E[k].pV[1]->r[0]) / 2.;
            y = (E[k].pV[0]->r[1] + E[k].pV[1]->r[1]) / 2.;
            z = (E[k].pV[0]->r[2] + E[k].pV[1]->r[2]) / 2.;
            fprintf(outfile, "    WARNING:  vertex overrun imminent:  V[%d], "
              "V[%d], x = %5.2f, y = %5.2f, z = %5.2f\n", E[k].pV[0]-V,
              E[k].pV[1]-V, x, y, z);
        }
#endif
    } /* end k-loop (edges) */

    /* use vector routine to find shortest edge [notes, 1990.10.17] */
    k = ISMIN(&numE, &E[0].len, &i8) - 1;   /* adjust FORTRAN indexing */
    x = (E[k].pV[0]->r[0] + E[k].pV[1]->r[0]) / 2.;
    y = (E[k].pV[0]->r[1] + E[k].pV[1]->r[1]) / 2.;
    z = (E[k].pV[0]->r[2] + E[k].pV[1]->r[2]) / 2.;
    fprintf(outfile, "  Minimum edge length = %6.4f:  E[%d], x = %5.2f, "
      "y = %5.2f, z = %5.2f\n", E[k].len, k, x, y, z);

/*===========================================================================
    Finally, calculate rotational inertia tensor of rings of points in order 
    to fit best circle to each and thereby to find inclination and radius of 
    each.
  ===========================================================================*/

    /* loop over rings (skip "ring 0," i.e., central point) */
    for (m = 1;  m <= numRings;  ++m) {
        double   sum_R=0., dot;

        /* initialize work variables */
        numpts = ring_last[m] - ring_last[m-1];
        for (j = 0;  j < 9;  ++j)
            ((double *)rot_inertia)[j] = 0.;

        /* loop over points in ring and add contributions to tensor */
        for (k = ring_last[m];  k > ring_last[m-1];  --k) {
            rot_inertia[0][0] += SQR(V[k].r[1]) + SQR(V[k].r[2]);
            rot_inertia[0][1] -= V[k].r[0] * V[k].r[1];
            rot_inertia[0][2] -= V[k].r[0] * V[k].r[2];
            rot_inertia[1][1] += SQR(V[k].r[0]) + SQR(V[k].r[2]);
            rot_inertia[1][2] -= V[k].r[1] * V[k].r[2];
            rot_inertia[2][2] += SQR(V[k].r[0]) + SQR(V[k].r[1]);
        }
        /* tensor is symmetric... */
        rot_inertia[1][0] = rot_inertia[0][1];
        rot_inertia[2][0] = rot_inertia[0][2];
        rot_inertia[2][1] = rot_inertia[1][2];

        /* find eigenvalues, eigenvectors */
        grr_jacobi((double **)rot_inertia, 3, eigenval, (double **)eigenvec,
          &nrot);

#ifdef UDEBUG
        fprintf(dbgfile, "   number of JACOBI rotations:  %d\n", nrot);
#endif /* UDEBUG */

        if (eigenval[1] > eigenval[0])
            if (eigenval[2] > eigenval[1])  jmax = 2;
            else  jmax = 1;
        else
            if (eigenval[2] > eigenval[0])  jmax = 2;
            else  jmax = 0;

#ifdef UDEBUG
        fprintf(dbgfile,
          "   ring %d eigenvec = (%10.2e, %10.2e, %10.2e)\n",
          m, eigenvec[jmax][0], eigenvec[jmax][1], eigenvec[jmax][2]);
#endif /* UDEBUG */

        ring_incl[m] = acos(MIN(eigenvec[jmax][2], 1.)) / PI180;
        if ((eigenvec[jmax][0] == 0.) && (eigenvec[jmax][1] == 0.))
            ring_azim[m] = 0.;
        else
            ring_azim[m] = atan2(eigenvec[jmax][1], eigenvec[jmax][0]) / PI180;

        /* again loop over points in ring, now calculating projected radii */
        for (j = 0, k = ring_last[m];  j < numpts;  ++j, --k) {
            dot = DOT(V[k].r, eigenvec[jmax]);
            for (i = 0;  i < 3;  ++i)
                r_proj[i] = V[k].r[i] - dot*eigenvec[jmax][i];

            sum_R += NORM(r_proj);
        }
        /* best-fit circle is just average radius of ring (try best ellipse?) */
        ring_R[m] = sum_R / numpts;

    } /* end m-loop (rings) */

/*---------------------------------------------------------------------------
    Store normal to last ring for later use in viscosity; print CPU time.
  ---------------------------------------------------------------------------*/

    x = (eigenvec[jmax][2] < 0.)?  -1. : 1.;   /* sign of z-component */
    for (i = 0;  i < 3;  ++i)
        ring_last_n[i] = x * eigenvec[jmax][i];

    dt = jacobi_time - j_t1;
    fprintf(outfile, "  grr_jacobi() time:           %8.6f\n", dt);
    dt = SECOND() - t1;
    fprintf(outfile, "  update_points() time:        %8.6f\n", dt);
    update_time += dt;

} /* end function update_points() */
